/* 
   NGActiveConnection.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGActiveConnection.m,v 1.10 1999/01/21 13:31:59 helge Exp $

#import "common.h"
#import "NGActiveConnection.h"
#import "NGConnection+private.h"
#import "NGDistantObject.h"
#import "DOCoder.h"
#import "DOProtocol.h"
#import "NSInvocation+DO.h"

NSMapTableKeyCallBacks NGIdentityObjectKeyCallbacks = {
  (unsigned(*)(NSMapTable *, const void *))          __NSHashPointer,
  (BOOL(*)(NSMapTable *, const void *, const void *))__NSComparePointers,
  (void (*)(NSMapTable *, const void *anObject))     __NSRetainObjects,
  (void (*)(NSMapTable *, void *anObject))           __NSReleaseObjects,
  (NSString *(*)(NSMapTable *, const void *))        __NSDescribePointers,
  (const void *)NULL
};

@implementation NGActiveConnection

+ (NGConnection *)connectionWithRegisteredAddress:(id<NGSocketAddress>)_address {
  NGActiveSocket *socket = nil;

  NS_DURING {
    socket = [NGActiveSocket socketInDomain:[_address domain]];
    [socket connectToAddress:_address];
  }
  NS_HANDLER {
    socket = nil;
  }
  NS_ENDHANDLER;
  
  if (socket == nil)
    return nil;
  else {
    id bufferedSocket = [NGBufferedStream filterWithSource:socket];
    //id bufferedSocket = socket;
    return AUTORELEASE([[self alloc] initWithStream:bufferedSocket]);
  }
}

- (id)initWithStream:(id<NGSerializer,NGStream>)_stream {
  if ((self = [super init])){
    self->connection   = RETAIN(_stream);
    self->coder        = [[DOCoder allocWithZone:[self zone]]
                                   initWithStream:_stream connection:self];

    self->vendedObjects = NSCreateMapTable(NGIdentityObjectKeyCallbacks, // id
                                           NSObjectMapValueCallBacks,    // l-proxy
                                           119);
    self->refToProxy = NSCreateMapTable(NSNonOwnedPointerMapKeyCallBacks, // ref
                                        NSNonRetainedObjectMapValueCallBacks,
                                        119);
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  [self shutdown];
  
  [self->coder invalidate];

  if (self->vendedObjects) {
    NSFreeMapTable(self->vendedObjects);
    self->vendedObjects = NULL;
  }
  if (self->refToProxy) {
    NSFreeMapTable(self->refToProxy);
    self->refToProxy = NULL;
  }
  
  RELEASE(self->connection);   self->connection   = nil;
  RELEASE(self->coder);        self->coder        = nil;
  [super dealloc];
}
#endif

// operations

- (NGDistantObject *)vendObject:(id<NSObject>)_localObject {
  // called during coding of a local object
  NGDistantObject *proxy = nil;

  if (_localObject == nil) // nil reference
    return nil;

  // look whether the object is already vended ..
  proxy = NSMapGet(self->vendedObjects, _localObject);
  if (proxy == nil) { // it isn't vended yet ..

    proxy = [[NGDistantObject allocWithZone:[self zone]]
                              initWithLocal:_localObject
                              connection:self];

    // mark object as vended
    NSMapInsert(self->vendedObjects, _localObject, proxy);

    NSLog(@"proxy: vended object 0x%08X<%s>, proxy=0x%08X",
          _localObject, class_get_class_name(*(Class *)_localObject),
          proxy);

    RELEASE(proxy); // is retained in vendedObjects ..
  }

  // return reference
  return proxy;
}
- (void)unvendObject:(id<NSObject>)_localObject {
  // called when remote-proxy is released

  if (_localObject == nil) // nil object
    return;

  NSLog(@"proxy: releasing local object 0x%08X<%s>, was remote-proxy=0x%08X",
        _localObject, class_get_class_name(*(Class *)_localObject),
        NSMapGet(self->vendedObjects, _localObject));
    
  // remove from vended-objects set
  NSMapRemove(self->vendedObjects, _localObject);
  _localObject = nil;
}

- (BOOL)hasProxyForRemoteReference:(unsigned)_reference {
  if (_reference == 0)
    return YES;

  if (NSMapGet(self->refToProxy, (void *)_reference))
    return YES;
  else
    return NO;
}
- (NGDistantObject *)proxyForRemoteReference:(unsigned)_reference {
  return _reference ? NSMapGet(self->refToProxy, (void *)_reference) : nil;
}
- (void)registerProxy:(NGDistantObject *)_proxy {
  if (_proxy->reference)
    NSMapInsert(self->refToProxy, (void *)_proxy->reference, _proxy);
}

// invocation

- (void)releaseProxy:(NGDistantObject *)_proxy {
  // called when a proxy was deallocated

  // ignore serial, is oneway ..
  DOSendRequestHead(self->connection, DOMsg_RELEASE);
  [self->connection serializeInt:_proxy->reference];
  [self->connection flush];

  NSLog(@"proxy: releasing proxy 0x%08X<%s>, reference=0x%08X",
        _proxy, class_get_class_name(*(Class *)_proxy),
        _proxy->reference);
  
  // remove proxy from proxy set
  NSMapRemove(self->refToProxy, (void *)_proxy->reference);
}

- (NSMethodSignature *)signatureForSelector:(SEL)_selector
  ofProxy:(NGDistantObject *)_proxy {

  NSMethodSignature *signature = nil;
  char *csig = NULL;

  csig = DOGetSignature(self->connection, self->connection,
                        _selector, _proxy->reference);
  if (csig) {
    signature = [NSMethodSignature signatureWithObjCTypes:csig];
    NGFree(csig); csig = NULL;
  }
  return signature;
}

- (void)performInvocation:(NSInvocation *)_invocation
  forProxy:(NGDistantObject *)_proxy {

  NSMethodSignature *signature  = nil;
  DOMessageCode     code        = DOMsg_INVOKE;
  unsigned          serial      = 1;
  struct timeval    tv;
  NSTimeInterval    t;

  NSAssert([_proxy isProxy], @"can only invoke methods for proxies ..");

  gettimeofday(&tv, NULL);
  t = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);

  signature = [_invocation methodSignature];
  if (signature == nil)
    signature = [_proxy methodSignatureForSelector:[_invocation selector]];

  NSAssert(_invocation, @"got no invocation ..");

  if (signature == nil) {
    // throw does-not-recognize if no signature could be found ..
    char buffer[512];
    SEL  sel = [_invocation selector];
    
    NSLog(@"do: got no signature for selector '%@' on proxy 0x%08X<%s>",
           NSStringFromSelector(sel), _proxy, *(Class *)_proxy);

    // Don't use initWithFormat: here because it can cause infinite recursion.

    sprintf(buffer,
            "remote object 0x%08X for proxy 0x%08X<%s> does not recognize %s",
            _proxy->reference, (unsigned)_proxy,
            class_get_class_name(*(Class *)_proxy),
            sel ? sel_get_name(sel) : "<null selector>");
    
    THROW([[ObjcRuntimeException alloc]
	      setReason:[NSString stringWithCString:buffer]]);
    return;
  }
  
  NSAssert(signature, @"got no signature ..");

  NSLog(@"do: %@: shall perform '%@'%s for proxy 0x%08X ..",
         self, NSStringFromSelector([_invocation selector]),
         [signature isOneway] ? " oneway" : "", _proxy);

  // send head
  serial = DOSendRequestHead(self->connection, DOMsg_INVOKE);
  [self->coder encodeInvocation:_invocation];
  /*
    DOMarshallInvocation(_invocation, signature, _proxy->reference,
                         self->coder, self->connection);
  */
  [self->connection flush];

  NSLog(@"do: .. send request ..");

  if (![signature isOneway]) {
    // handle reply
    // code = DOReceiveHead(self, &replySerial);
    code = [self dispatchUntilReplyWithSerial:serial];
    NSAssert(code & REPLY, @"got a request, expected reply ..");

    if ((code & VALUE) == DOMsg_INVOKE) {
      // receive body
      int resultLen;

      [self->connection deserializeDataAt:&resultLen
                        ofObjCType:@encode(int)
                        context:nil];

      NSAssert([signature methodReturnLength] == resultLen,
               @"mismatch in invocation result length ..");
    
      if (resultLen > 0) {
        void *result = NGMalloc(resultLen);

        [self->coder decodeValueOfObjCType:[signature methodReturnType]
             at:result];
    
        [_invocation setReturnValue:result];

        NGFree(result); result = NULL;
      }
    }
    else if ((code & VALUE) == DOMsg_THROW) {
      char *className = NULL, *name = NULL, *why = NULL;
      NSString *aName = nil, *aReason = nil;
      id exception = nil;

      [self->connection deserializeDataAt:&className
           ofObjCType:@encode(char *) context:nil];
      [self->connection deserializeDataAt:&name
           ofObjCType:@encode(char *) context:nil];
      [self->connection deserializeDataAt:&why
           ofObjCType:@encode(char *) context:nil];

      aName   = name ? [NSString stringWithCString:name] : nil;
      aReason = why  ? [NSString stringWithCString:why]  : nil;

      if (className) {
        Class c = objc_lookup_class(className);
        exception = [c ? c : [NSException class] allocWithZone:[self zone]];
      }
      else
        exception = [NSException allocWithZone:[self zone]];

      // release decoding resources
      NGFree(className); className = NULL;
      NGFree(name);      name      = NULL;
      NGFree(why);       why       = NULL;
    
      // throw exception
      THROW([exception initWithName:aName reason:aReason userInfo:
                       [NSDictionary dictionaryWithObject:self forKey:@"ORB"]]);
    }
    else {
      NSLog(@"invalid reply code for invocation !");
    }
  }
  else { // oneway
    //NSLog(@"don't wait for reply, was oneway !");
  }

  gettimeofday(&tv, NULL);
  t = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - t;
  NSLog(@"perf: invocation of '%@' took %4.3fs.",
         NSStringFromSelector([_invocation selector]),
         t < 0.0 ? -1.0 : t);
}

- (NGDistantObject *)rootProxy {
  DOMessageCode   code        = 0;
  unsigned        serial      = 1;
  unsigned        replySerial = 0;
  NGDistantObject *proxy      = nil;

  // send request head (root has no body)
  serial = DOSendRequestHead(self->connection, DOMsg_ROOT);
  [self->connection flush];

  // receive response head
  code = DOReceiveHead(self->connection, &replySerial);
  NSAssert(code & REPLY,          @"got a request, expected reply ..");
  NSAssert(serial == replySerial, @"unorded messages ..");

  // receive response body (a proxy)
  proxy = [self->coder decodeObject];

  return proxy;
}

// handling requests

- (void)sendRootMessageReplyWithSerial:(unsigned)_serial {
  DOSendReplyHead(self->connection, DOMsg_ROOT | REPLY, _serial);
  [self->coder encodeRootObject:[self rootObject]];
  [self->connection flush];
}

- (void)sendSignatureReplyWithSerial:(unsigned)_serial
  selector:(SEL)_selector object:(id)_object {
  
  NSMethodSignature *signature = nil;

  NSAssert(![_object isProxy], @"cannot serve signature of proxy ..");

  //NSLog(@"do: getting method signature for selector '%@' object '%@'",
  //      NSStringFromSelector(_selector), _object);
  
  signature = [_object methodSignatureForSelector:_selector];
  
  // send message 
  DOSendReplyHead(self->connection, DOMsg_SIGNATURE | REPLY, _serial);
  {
    NSString   *sig  = [signature objCTypes];
    const char *csig = [sig cString];

    [self->connection serializeDataAt:&csig ofObjCType:@encode(char *) context:nil];
  }
  [self->connection flush];
}

- (void)sendInvocationReplyWithSerial:(unsigned)_serial
  invocation:(NSInvocation *)_inv {

  //NSLog(@"do: invoking '%@' ..", NSStringFromSelector([_inv selector]));

  NS_DURING {
    NSMethodSignature *signature = [_inv methodSignature];
    //if ([signature isOneway]) NSLog(@"ONEWAY: '%s'", [signature types]);
    
    [_inv invoke];

    if (![signature isOneway]) {
      //[self->coder encodeInvocationResult:_inv];

      // send message head
      DOSendReplyHead(self->connection, DOMsg_INVOKE | REPLY, _serial);

      // send message body
      {
        int returnLen = [signature methodReturnLength];

        [self->connection serializeDataAt:&returnLen
                          ofObjCType:@encode(int)
                          context:nil];

        if (returnLen > 0) {
          void *buffer = NGMalloc(returnLen);

          [_inv getReturnValue:buffer];
          [self->coder encodeValueOfObjCType:[signature methodReturnType] at:buffer];
        }
      }
      [self->connection flush];
    }
    else { // oneway
      //NSLog(@"was oneway, sending no return ..");
    }
  }
  NS_HANDLER {
    const char *className = class_get_class_name(*(Class *)localException);
    const char *name      = [[localException name]   cString];
    const char *why       = [[localException reason] cString];

    NGCLogT("do", "exception %s occured during invocation ..", name);

    DOSendReplyHead(self->connection, DOMsg_THROW | REPLY, _serial);
    [self->connection serializeDataAt:&className
                      ofObjCType:@encode(char *) context:nil];
    [self->connection serializeDataAt:&name ofObjCType:@encode(char *) context:nil];
    [self->connection serializeDataAt:&why  ofObjCType:@encode(char *) context:nil];
    [self->connection flush];
  }
  NS_ENDHANDLER;
}

- (DOMessageCode)dispatchUntilReplyWithSerial:(unsigned)_serial {
  DOMessageCode code;
  unsigned      serial;

  do {
    code = DOReceiveHead(self->connection, &serial);
    
    if (code & REPLY) {
      if (serial == _serial) {
        // found required reply
        break;
      }
      else {
        NSLog(@"got reply %i, serial %i .. expected serial %i",
              code & VALUE, serial, _serial);
        break;
      }
    }
    else {
      NSLog(@"dispatching nested message (code=%i) ..",
            code & VALUE);
      [self dispatchMessageWithCode:code serial:serial];
    }
  }
  while (YES);

  return code;
}

- (void)dispatchMessageWithCode:(DOMessageCode)_code serial:(unsigned)_serial {
  BOOL          isReply = _code & REPLY;
  DOMessageCode code    = _code & VALUE;

  NSAssert(isReply == NO, @"no reply accepted ..");
  
  switch (code) {
    case DOMsg_SHUTDOWN:
      [self shutdown];
      break;
    
    case DOMsg_ROOT:
      // reply
      [self sendRootMessageReplyWithSerial:_serial];
      break;

    case DOMsg_SIGNATURE: {
      char *selName = NULL;
      SEL  selector = NULL;
      id   object   = nil;

      [self->connection deserializeDataAt:&selName
                        ofObjCType:@encode(char *)
                        context:nil];
      selector = selName ? sel_get_any_uid(selName) : NULL;
      NGFree(selName); selName = NULL;

      object = (id)[self->connection deserializeInt];
      
      // reply
      [self sendSignatureReplyWithSerial:_serial
            selector:selector object:object];
      break;
    }

    case DOMsg_INVOKE: {
      NSInvocation *invocation = nil;

      invocation = [self->coder decodeObject];
      //invocation = DOUnmarshallInvocation(self->coder, self->connection);
      //NSLog(@"do: got invocation %@", invocation);

      [self sendInvocationReplyWithSerial:_serial
            invocation:invocation];
      break;
    }

    case DOMsg_RELEASE: {
      id reference = (id)[self->connection deserializeInt];

      if (reference) {
        NSLog(@"proxy: release proxy reference 0x%08X ..", reference);
        [self unvendObject:reference];
      }
      else {
        NSLog(@"tried to release invalid proxy reference 0x%08X ..", reference);
      }
      break;
    }

    default:
      NSLog(@"WARNING: Unknown message type %i ..", code);
      break;
  }
}

- (void)handleRequestException:(NGStreamException *)_exception
  passive:(NGPassiveConnection *)_parent {
  
  NSLog(@"%@: %@", self, _exception);
  
  [self shutdown];
  [_parent removeChildConnection:self];
}
- (void)connectionWasTerminated:(NGEndOfStreamException *)_exception
  passive:(NGPassiveConnection *)_parent {
    
  NSLog(@"%@: %@", self, _exception);

  abort();
  [self shutdown];
  [_parent removeChildConnection:self];
}

- (void)handleRequest:(NGPassiveConnection *)_parent {
  //NSLogT(@"do: %@: handle request ..", self);

  NS_DURING {
    DOMessageCode code;
    unsigned int  serial;

    code = DOReceiveHead(self->connection, &serial);
    NSLog(@"do: got message %i serial %i isReply=%s",
          code, serial, code & REPLY ? "yes" : "no");

    if ((code < 1) || (code > 8)) {
      THROW([[NSException alloc]
                          initWithName:@"UnknownMessageCode"
                          reason:@"NGConnection received invalid message code"
                          userInfo:nil]);
    }

    if (code == DOMsg_SHUTDOWN)
      [self shutdown];
    else
      [self dispatchMessageWithCode:code serial:serial];
    
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      [self connectionWasTerminated:(NGEndOfStreamException *)localException
            passive:_parent];
    }
    else if ([localException isKindOfClass:[NGStreamException class]]) {
      [self handleRequestException:(NGStreamException *)localException
            passive:_parent];
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;
}

- (void)shutdown {
  if (self->coder) {
    if ([(id)self->connection isConnected]) {
      NS_DURING {
        DOSendRequestHead(self->connection, DOMsg_SHUTDOWN);
        [self->connection flush];
        NSLog(@"do: send shutdown.", self);
      }
      NS_HANDLER {
        if (![localException isKindOfClass:[NGStreamException class]])
          [localException raise];
      }
      NS_ENDHANDLER;
    }
  
    NSResetMapTable(self->vendedObjects);

    [self->connection close];
    RELEASE(self->connection); self->connection = nil;
    RELEASE(self->coder);      self->coder      = nil;

    NSLog(@"do: connection %@ was shutdown.", self);
  }
}

@end
