/* 
   NGPassiveConnection.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGPassiveConnection.m,v 1.10 1999/01/21 13:31:59 helge Exp $

#import "common.h"
#import "NGPassiveConnection.h"
#import "NGConnection+private.h"
#import "NGActiveConnection.h"

@implementation NGPassiveConnection

static NGPassiveConnection *defaultConnection = nil;
static int connectionBackLog = 5;

+ (NGConnection *)defaultConnection {
  if (defaultConnection == nil) {
    defaultConnection = [[NGPassiveConnection alloc] init];
  }
  return defaultConnection;
}

- (id)init {
  if ((self = [super init])) {
    self->addressToConnection = NSCreateMapTable(NSObjectMapKeyCallBacks,
                                                 NSObjectMapValueCallBacks,
                                                 16);
    self->addressToSocket     = NSCreateMapTable(NSObjectMapKeyCallBacks,
                                                 NSObjectMapValueCallBacks,
                                                 16);
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->port); self->port = nil;

  if (self->addressToConnection) {
    NSFreeMapTable(self->addressToConnection);
    self->addressToConnection = NULL;
  }
  if (self->addressToSocket) {
    NSFreeMapTable(self->addressToSocket);
    self->addressToSocket = NULL;
  }
  [super dealloc];
}
#endif

// registry

- (BOOL)registerAddress:(id<NGSocketAddress>)_address {
  volatile BOOL result = NO;
  
  if (self->port) {
    NSLog(@"do: connection is already registered ..");
    return NO;
  }

  NS_DURING {
    self->port = [[NGPassiveSocket allocWithZone:[self zone]]
                                   initWithDomain:[_address domain]];
    [self->port bindToAddress:_address];
    [self->port listenWithBacklog:connectionBackLog];

    result = YES;
  }
  NS_HANDLER { }
  NS_ENDHANDLER;

#ifdef LIB_FOUNDATION_LIBRARY
  if (result && (self->port != nil)) {
    [[NSNotificationCenter defaultCenter]
                           addObserver:self selector:@selector(acceptConnection:)
                           name:NSFileObjectBecameActiveNotificationName
                           object:self->port];
    
    [[NSRunLoop currentRunLoop]
                addFileObject:self->port
                activities:NSPosixReadableActivity
                forMode:NSDefaultRunLoopMode];

    //NSLog(@"added object and notification ..");
  }
#endif

  return result;
}

// managing active connections

- (id)remoteNameOfConnection:(NGActiveConnection *)_connection {
  NSMapEnumerator en = NSEnumerateMapTable(self->addressToConnection);
  id<NGSocketAddress> key;
  NGConnection        *value;

  if (NSNextMapEnumeratorPair(&en, (void**)&key, (void**)&value))
    return key;
  else
    return nil;
}

- (void)addConnectionForStream:(id<NGActiveSocket>)_stream
  remote:(id<NGSocketAddress>)_remote {
  
  NGActiveConnection *connection = nil;
  id stream = _stream;

  stream = [NGBufferedStream filterWithSource:_stream];
  //[stream setReadBufferSize:0];

  connection = [[NGActiveConnection allocWithZone:[self zone]]
                                    initWithStream:stream];
  [connection setRootObject:[self rootObject]];

  NSLog(@"do: adding connection %@:%@ ..", connection, _remote);

  NSMapInsert(self->addressToConnection, _remote, connection);
  NSMapInsert(self->addressToSocket,     _remote, stream);

#if LIB_FOUNDATION_LIBRARY
    [[NSNotificationCenter defaultCenter]
                           addObserver:self selector:@selector(requestPending:)
                           name:NSFileObjectBecameActiveNotificationName
                           object:stream];
    [[NSRunLoop currentRunLoop]
                addFileObject:stream
                activities:NSPosixReadableActivity
                forMode:NSDefaultRunLoopMode];
#endif
  
  RELEASE(connection); connection = nil;
}

- (void)removeChildConnection:(NGActiveConnection *)_childConnection {
  NSMapEnumerator     knownAs = NSEnumerateMapTable(self->addressToConnection);
  id<NGSocketAddress> name    = nil;
  NGConnection        *value  = nil;
  
  NSLog(@"do: removing connection %@ ..", _childConnection);

  if (NSNextMapEnumeratorPair(&knownAs, (void**)&name, (void**)&value)) {
    id<NGActiveSocket> socket = NSMapGet(self->addressToSocket, name);

#if LIB_FOUNDATION_LIBRARY
    [[NSNotificationCenter defaultCenter]
                           removeObserver:self
                           name:NSFileObjectBecameActiveNotificationName
                           object:socket];
    [[NSRunLoop currentRunLoop]
                removeFileObject:socket
                forMode:NSDefaultRunLoopMode];
#endif

    if ([socket isConnected])
      [socket close];

    NSMapRemove(self->addressToConnection, name);
    NSMapRemove(self->addressToSocket,     name);
  }
}

// running

- (void)handleReceiveException:(NSException *)_exception
  connection:(NGConnection *)_connection {

  NGCLog("%@: catched exception during receive: %@ ..", self, _exception);
  abort();
  //[_connection shutdown];
}

- (void)acceptConnection:(id)_sender {
  id<NGActiveSocket> socket = [self->port accept];

  if (socket) {
    NSLog(@"do: accepted connection %@ ..", socket);
      
    [self addConnectionForStream:socket
          remote:[socket remoteAddress]];
  }
}

- (void)requestPending:(NSNotification *)_sender {
  NGActiveSocket      *socket = [_sender object];
  id<NGSocketAddress> address = [socket remoteAddress];

  NSLog(@"request pending from %@ ..", address);

  if (address) {
    NGActiveConnection *connection = nil;

    if ((connection = NSMapGet(self->addressToConnection, address))) {
      RETAIN(connection);
      
      NS_DURING {
        BOOL (*isConnected)(id, SEL)              = NULL;
        BOOL (*wouldBlock)(id, SEL, NGStreamMode) = NULL;
        int count = 1;

        isConnected = (void*)[socket methodForSelector:@selector(isConnected)];
        wouldBlock  = (void*)[socket methodForSelector:@selector(wouldBlockInMode:)];

        // collect all requests in queue .. (could be multiple oneway messages)
        do {
          //NSLog(@"handle request %i ..", count);
          [connection handleRequest:self];
          count++;

          if (wouldBlock(socket, @selector(wouldBlockInMode:),NGStreamMode_readOnly))
            break;
        }
        while (isConnected(socket, @selector(isConnected)));
      }
      NS_HANDLER {
        [self handleReceiveException:localException
              connection:connection];
      }
      NS_ENDHANDLER;
      
      RELEASE(connection); connection = nil;
    }
  }
}

- (void)run {
  NSAssert(self->port, @"connection is not registered yet ..");

  [self->port listenWithBacklog:5];

  while (YES) {
    NSAutoreleasePool *pool      = nil;
    struct pollfd     *pfdArray  = NULL;
    unsigned          numSockets = 0;
    NSMapTable        *idx2name;

    pool = [NSAutoreleasePool new];

    numSockets = NSCountMapTable(self->addressToSocket) + 1;
    pfdArray   = NGMallocAtomic(sizeof(struct pollfd) * numSockets);
    idx2name   = NSCreateMapTable(NSIntMapKeyCallBacks,
                                  NSObjectMapValueCallBacks,
                                  numSockets);
    
    { // fill array
      NSMapEnumerator     names = NSEnumerateMapTable(self->addressToSocket);
      id<NGSocketAddress> name;
      NGActiveSocket      *socket; // needs to be typed for 'fileDescriptor'
      register int i;

      // active sockets
      for (i = 0;
           NSNextMapEnumeratorPair(&names, (void**)&name, (void**)&socket); i++) {
        pfdArray[i].fd      = [socket fileDescriptor];
        pfdArray[i].events  = POLLRDNORM;
        pfdArray[i].revents = 0;

        NSMapInsert(idx2name, (void*)(i + 1), name);
      }

      // passive socket
      pfdArray[numSockets - 1].fd      = [(id)self->port fileDescriptor];
      pfdArray[numSockets - 1].events  = POLLIN;
      pfdArray[numSockets - 1].revents = 0;
    }

    // run poll
    NS_DURING {
      int result;

      result = poll(pfdArray, numSockets, 6000 * 36 * 1000); // 1 hour

      if (result == -1) {
        NSLog(@"do: error during poll: %s", strerror(errno));
      }
      else if (result == 0) {
        NSLog(@"do: poll timeout ..");
      }
      else { // result is the number of descriptors with events
        NSMutableArray *shutdown = nil;
        NSMutableArray *pending  = nil;
        int i;

        shutdown = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:4];
        pending  = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:4];
        &shutdown;
        &pending;

        // collect connections with events
        for (i = 0; (i < (numSockets - 1)) && (result > 0); i++) {
          short revents = pfdArray[i].revents;

          if (revents) {
            id name = NSMapGet(idx2name, (void *)(i + 1));

            if (name) {
              if (revents & POLLHUP) {
                [shutdown addObject:name];
                result--;
              }
              else if (revents & POLLRDNORM) {
                [pending  addObject:name];
                result--;
              }
            }
            else {
              NSLog(@"got no name for connection index %i ..", i);
            }
          }
        }

        // finish shutdown connections
        if ([shutdown count] > 0) {
          NSEnumerator *names = [shutdown objectEnumerator];
          id           name   = nil;

          while ((name = [names nextObject])) {
            NSString *reason = @"%@: catched exception during shutdown: %@ ..";
            NGActiveConnection *connection =
              NSMapGet(self->addressToConnection, name);

            NS_DURING {
              NSLog(@"do: connection %@ was closed ..", connection);
              [connection shutdown];
            }
            NS_HANDLER {
              NSLog(reason, self, localException);
            }
            NS_ENDHANDLER;
            
            [self removeChildConnection:connection];
          }

          RELEASE(shutdown); shutdown = nil;
        }

        // execute events
        if ([pending count] > 0) {
          NSEnumerator *names = [pending objectEnumerator];
          id           name   = nil;

          while ((name = [names nextObject])) {
            NGActiveConnection *connection =
              NSMapGet(self->addressToConnection, name);
            NGActiveSocket *socket = NSMapGet(self->addressToSocket, name);

            RETAIN(connection);
            RETAIN(socket);
            
            NS_DURING {
              int count = 1;

              do {
                //NSLog(@"handle request %i ..", count);
                [connection handleRequest:self];
                count++;
                
                if ([socket wouldBlockInMode:NGStreamMode_readOnly]) {
                  //NSLog(@"  request %i would block, exiting ..", count);
                  break;
                }
              }
              while ([socket isConnected]);

              /*
              if (![socket isConnected]) {
                NSLog(@"  socket isn't connected anymore..");
                }*/
            }
            NS_HANDLER {
              NGCLog("%@: catched exception during receive: %@ ..",
                     self, localException); 
            }
            NS_ENDHANDLER;

            [connection release]; connection = nil;
            [socket     release]; socket     = nil;
          }

          RELEASE(pending); pending = nil;
        }

        RELEASE(shutdown); shutdown = nil;
        RELEASE(pending);  pending  = nil;

        // check whether a new connection is pending
        if (pfdArray[numSockets - 1].revents)
          [self acceptConnection:self];
      }
    }
    NS_HANDLER {
      NGCLog("%@: catched exception %@ ..", self, localException);
    }
    NS_ENDHANDLER;

    // release resources
    if (idx2name) {
      NSFreeMapTable(idx2name);
      idx2name = NULL;
    }
    if(pfdArray) {
      NGFree(pfdArray);
      pfdArray = NULL;
    }

    [pool release]; pool = nil;
  }
}

@end
