/* 
   NGHashMap.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan@mdlink.de)
           Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGHashMap.m,v 1.7 1999/01/12 12:19:46 helge Exp $

#import "common.h"
#import "NGHashMap.h"
#import "NGMemoryAllocation.h"
#import "NGNil.h"

typedef struct _LList {
  struct _LList *next;
  id            object;
  unsigned int  count;
} LList;

static inline void *initLListElement(id _object, LList* _next) {
  LList *element  = NGMalloc(sizeof(LList));
  RETAIN(_object);
  element->object        = _object;
  element->next          = _next;
  element->count         = 0;
  return element;
}

static inline void checkForAddErrorMessage(id _self, id _object, id _key) {
  if (!_key) {
    InvalidArgumentException *exc = [InvalidArgumentException alloc];
    [exc initWithFormat:@"nil key to be added in HashMap with object %@",
           _object ? _object : @"<nil>"];
    [exc setUserInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                                     _self, @"map",
                                     _key ? _key : @"<nil>",
                                     @"key",
                                     _object ? _object : @"<nil>",
                                     @"object",
                                     nil]];
    [exc raise];
  }
  if (!_object) {
    InvalidArgumentException *exc = [InvalidArgumentException alloc];
    [exc initWithFormat:@"nil object to be added in HashMap for key %@",
           _key ? _key : @"<nil>"];
    [exc setUserInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                                     _self, @"map",
                                     _key ? _key : @"<nil>",
                                     @"key",
                                     _object ? _object : @"<nil>",
                                     @"object",
                                     nil]];
    [exc raise];
  }
}

static inline void checkForRemoveErrorMessage(id _self, id _object, id _key) {
  if (!_object || !_key) {
    InvalidArgumentException *exc = [InvalidArgumentException alloc];
    [exc initWithFormat:@"nil object to be removed in HashMap for key %@",
           _key ? _key : @"<nil>"];
    [exc setUserInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                                     _self, @"map",
                                     _key ? _key : @"<nil>",
                                     @"key",
                                     _object ? _object : @"<nil>",
                                     @"object",
                                     nil]];
    [exc raise];
  }
}

static inline void raiseInvalidArgumentExceptionForNilKey(id _self) {
    InvalidArgumentException *exc = nil;
    exc = [[InvalidArgumentException alloc] initWithFormat:@"key is nil\n"];
    [exc setUserInfo:[NSDictionary dictionaryWithObject:_self forKey:@"map"]];
    [exc raise];
}

@interface _NGHashMapObjectEnumerator : NSEnumerator
{
  NSEnumerator *keys;
  NSEnumerator *elements;
  NGHashMap    *hashMap;
}
- (id)initWithHashMap:(NGHashMap *)_hashMap;
- (id)nextObject;
@end

@interface _NGHashMapObjectForKeyEnumerator : NSEnumerator
{
  LList *element;
  NGHashMap    *map;
}
- (id)initWithHashMap:(NGHashMap *)_hashMap andKey:(id)_key;
- (id)nextObject;
@end

@interface _NGHashMapKeyEnumerator : NSEnumerator
{
  NSMapEnumerator enumerator;
  NGHashMap *map;
}
- (id)initWithHashMap:(NGHashMap *)_hashMap;
- (id)nextObject;
@end

// ************************* NGHashMap *************************

@interface NGHashMap(private)
- (LList *)__structForKey:(id)_key;
- (NSMapEnumerator)__keyEnumerator;
- (void)__removeAllObjectsForKey:(id)_key;
- (void)__removeAllObjects;
@end

@implementation NGHashMap

// final methods

static inline void _removeAllObjectsInList(LList *list) {
  while (list) {
    register LList *element = NULL;
    
    RELEASE(list->object);
    element = list;
    list = list->next;
    NGFree(element);
  }
}

static inline LList *__structForKey(NGHashMap *self, id _key) {
  if (_key == nil) raiseInvalidArgumentExceptionForNilKey(self);
  return (LList *)NSMapGet(self->table, (void *)_key);
}

static inline unsigned __countObjectsForKey(NGHashMap *self, id _key) {
  LList *list = NULL;
  return (list = __structForKey(self, _key)) ? list->count : 0;
}

// methods

+ (id)hashMap {
  return AUTORELEASE([[self alloc] init]);
}

+ (id)hashMapWithHashMap:(NGHashMap *)_hashMap {
  return AUTORELEASE([[self alloc] initWithHashMap:_hashMap]);
}

+ (id)hashMapWithObjects:(NSArray *)_objects forKey:(id)_key {
  return AUTORELEASE([[self alloc] initWithObjects:_objects forKey:_key]);
}

+ (id)hashMapWithDictionary:(NSDictionary *)_dict {
  return AUTORELEASE([[self alloc] initWithDictionary:_dict]);
}

- (id)init {
  return [self initWithCapacity:0];
}

- (id)initWithCapacity:(unsigned int)_size {
  if ((self = [super init])) {
    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
                                     NSNonOwnedPointerMapValueCallBacks, 
                                     _size * 4/3 ,NULL);
  }
  return self;
}

- (id)initWithHashMap:(NGHashMap *)_hashMap {
  NSEnumerator *keys    = nil;
  id            key     = nil;
  LList *list    = NULL;
  LList *newList = NULL;
  LList *oldList = NULL;

  if ((self = [self initWithCapacity:[_hashMap count]])) {
    keys  = [_hashMap keyEnumerator];
    while ((key = [keys nextObject])) {
      list           = [_hashMap __structForKey:key];
      newList        = initLListElement(list->object,NULL);
      newList->count = list->count;
      NSMapInsert(table,key,newList);
      while (list->next) {
        oldList       = newList;
        list          = list->next;
        newList       = initLListElement(list->object,NULL);
        oldList->next = newList;
      }
    }
  }
  return self;
}

- (id)initWithObjects:(NSArray *)_objects forKey:(id)_key {
  LList *root    = NULL;
  LList *element = NULL;
  LList *pred    = NULL;  
  int           count   = 0;
  int           i       = 0; 

  if (( self = [self initWithCapacity:1])) {
    count = [_objects count];
    if (count == 0) 
      return self;

    root = initLListElement([_objects objectAtIndex:0], NULL);
    pred = root;
    for (i = 1; i < count; i++) {
      element    = initLListElement([_objects objectAtIndex:i], NULL);
      pred->next = element;
      pred       = element;
    }
    root->count = i;
    NSMapInsert(table,_key, root);
  }
  return self;
}

- (id)initWithDictionary:(NSDictionary *)_dictionary {
  if (![self isKindOfClass:[NGMutableHashMap class]]) {
    AUTORELEASE(self);
    self = [[NGMutableHashMap allocWithZone:[self zone]]
                              initWithCapacity:[_dictionary count]];
  }
  else
    self = [self initWithCapacity:[_dictionary count]];

  if (self) {
    NSEnumerator *keys = [_dictionary keyEnumerator];
    id key;

    while ((key = [keys nextObject]))
      [(NGMutableHashMap *)self setObject:[_dictionary objectForKey:key] forKey:key];
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  if (table) {
    NSMapEnumerator mapenum;
    id key = nil, value = nil;

    mapenum = [self __keyEnumerator];

    while (NSNextMapEnumeratorPair(&mapenum, (void **)&key, (void **)&value))
      _removeAllObjectsInList((LList *)value);

    NSFreeMapTable(table);
    table = NULL;
  }
  [super dealloc];
}
#endif

// removing

- (void)__removeAllObjectsForKey:(id)_key {
  _removeAllObjectsInList(__structForKey(self, _key));
  NSMapRemove(table, _key);
}

- (void)__removeAllObjects {
  NSEnumerator *keys = nil;
  id           key  = nil;

  keys = [self keyEnumerator];
  while ((key = [keys nextObject]))
    _removeAllObjectsInList(__structForKey(self, key));

  NSResetMapTable(table);
}

- (BOOL)isEqual:(id)anObject {

  if (self == anObject)
    return YES;
  
  if ([anObject isKindOfClass:[NGHashMap class]] == NO)
    return NO;
  
  return [self isEqualToHashMap:anObject];
}

- (BOOL)isEqualToHashMap:(NGHashMap *)_other {
  NSEnumerator *keyEnumerator = nil;
  id            key           = nil;
  LList *selfList      = NULL;
  LList *otherList     = NULL;

  if (_other == self) 
    return YES;

  if ([self count] != [_other count])
    return NO;

  keyEnumerator = [self keyEnumerator];
  while ((key = [keyEnumerator nextObject])) {
    if (__countObjectsForKey(self, key) != [_other countObjectsForKey:key])
      return NO;

    selfList  = __structForKey(self, key);
    otherList = [_other __structForKey:key];
    while (selfList) {
      if (![selfList->object isEqual:otherList->object]) 
        return NO;

      selfList = selfList->next;
      otherList = otherList->next;      
    }
  }
  return YES;
}


- (id)objectForKey:(id)_key {
  LList *list;

  if (!(list = __structForKey(self, _key))) 
    return nil;

  if (list->next) {
    InvalidArgumentException *exc = nil;
    exc = [[InvalidArgumentException alloc]
              initWithFormat:@"more than one element for key %@\n", _key];
    [exc setUserInfo:[NSDictionary dictionaryWithObject:self forKey:@"map"]];
    [exc raise];
  }

  return list->object;
}

- (NSArray *)objectsForKey:(id)_key {
  NSArray         *array      = nil;
  NSEnumerator    *objectEnum = nil;
  id              object      = nil;
  id              *objects    = NULL;
  unsigned int    i           = 0;

  if (!(objectEnum = [self objectEnumeratorForKey:_key]))
    return nil;
  
  objects    = NGMalloc(sizeof(id) * __countObjectsForKey(self, _key));
  while ((object = [objectEnum nextObject]))
    objects[i++] = object;

  array = [NSArray arrayWithObjects:objects count:i]; //return AUTORELEASE object
  NGFree(objects);
  return array;
}

- (id)objectAtIndex:(unsigned int)_index forKey:(id)_key {
  LList *list = NULL;

  if (!(list = __structForKey(self, _key)))
    return nil;

  if (!(_index < list->count)) {
    IndexOutOfRangeException *exc = nil;
    exc = [[IndexOutOfRangeException alloc]
              initForSize:list->count index:_index];
    [exc setUserInfo:[NSDictionary dictionaryWithObject:self forKey:@"object"]];
    [exc raise];
  }

  while (_index--)
    list = list->next;

  return list->object;
}

- (NSArray *)allKeys {
  NSArray      *array   = nil;
  NSEnumerator *keys    = nil;  
  id           *objects = NULL;
  id           object   = nil;
  int          i        = 0;

  objects = NGMalloc(sizeof(id) * [self count]);
  keys    = [self keyEnumerator];
  while ((object = [keys nextObject]))
    objects[i++] = object;

  array = [[NSArray alloc] initWithObjects:objects count:i];
  NGFree(objects);
  return AUTORELEASE(array);
}

- (NSArray *)allObjects {
  NSEnumerator   *keys   = nil;
  id             object  = nil;
  NSMutableArray *mArray = nil;

  mArray = [[NSMutableArray alloc] init];
  keys   = [self keyEnumerator];
  while ((object = [keys nextObject])) 
    [mArray addObjectsFromArray:[self objectsForKey:object]];
  
  return [mArray copy];
}

- (unsigned int)countObjectsForKey:(id)_key {
  return __countObjectsForKey(self, _key);
}

- (NSEnumerator *)keyEnumerator {
  return AUTORELEASE([[_NGHashMapKeyEnumerator alloc] initWithHashMap:self]);
}

- (NSEnumerator *)objectEnumerator {
  return AUTORELEASE([[_NGHashMapObjectEnumerator alloc] initWithHashMap:self]);
}

- (NSEnumerator *)objectEnumeratorForKey:(id)_key {
  if (!_key)
    raiseInvalidArgumentExceptionForNilKey(self);
  
  return AUTORELEASE([[_NGHashMapObjectForKeyEnumerator alloc]
                          initWithHashMap:self andKey:_key]);
}

- (NSDictionary *)asDictionary {
  NSDictionary  *dict    = nil;
  NSEnumerator  *keys    = nil;
  id            key      = nil;
  id            *dicObj  = NULL;
  id            *dicKeys = NULL;
  int           cntKey   = 0;

  keys    = [self keyEnumerator];
  dicObj  = NGMalloc(sizeof(id) * [self count]);
  dicKeys = NGMalloc(sizeof(id) * [self count]);  
  
  while ((key = [keys nextObject])) {
    id            object   = nil;    
    LList  *list    = NULL;
    
    list = __structForKey(self, key);
    if (list->next) {
      id   *objects = NULL;
      int  cntObj   = 0;      
      
      objects = NGMalloc(sizeof(id) * list->count);
      {
        cntObj  = 0;
        while (list) {
          objects[cntObj++] = list->object;
          list = list->next;
        }
        
        if (cntObj == 0)
          object = [NGNil nilObject];
        else if (cntObj == 1)
          object = objects[0];
        else
          object = [NSArray arrayWithObjects:objects count:cntObj];
      }
      NGFree(objects); objects = NULL;
    }
    else 
      object = list->object;

    dicObj[cntKey]    = object;
    dicKeys[cntKey++] = key;
  }
  
  dict = [[NSDictionary alloc]
                        initWithObjects:dicObj
                        forKeys:dicKeys
                        count:cntKey];
  
  NGFree(dicObj);  dicObj  = NULL;
  NGFree(dicKeys); dicKeys = NULL;
  return AUTORELEASE(dict);
}

- (id)propertyList {
  NSDictionary  *dict    = nil;
  NSEnumerator  *keys    = nil;
  id            key      = nil;
  id            *dicObj  = NULL;
  id            *dicKeys = NULL;
  int           cntKey   = 0;

  keys    = [self keyEnumerator];
  dicObj  = NGMalloc(sizeof(id) * [self count]);
  dicKeys = NGMalloc(sizeof(id) * [self count]);  
  
  while ((key = [keys nextObject])) {
    id            object   = nil;    
    LList  *list    = NULL;
    
    list = __structForKey(self, key);
    if (list->next) {
      id   *objects = NULL;
      int  cntObj   = 0;      
      
      objects = NGMalloc(sizeof(id) * list->count);
      {
        cntObj  = 0;
        while (list) {
          objects[cntObj++] = list->object;
          list = list->next;
        }
        object = [NSArray arrayWithObjects:objects count:cntObj];
      }
      NGFree(objects); objects = NULL;
    }
    else 
      object = list->object;

    dicObj[cntKey]    = object;
    dicKeys[cntKey++] = key;
  }
  dict = AUTORELEASE([[NSDictionary alloc] initWithObjects:dicObj forKeys:dicKeys
                                           count:cntKey]);
  NGFree(dicObj);  dicObj  = NULL;
  NGFree(dicKeys); dicKeys = NULL;
  return dict;
}

- (NSString *)description {
  return [[self propertyList] description];
}

- (unsigned int)hash {
  return [self count];
}

- (unsigned int)count {
  return table ? NSCountMapTable(table) : 0;
}

- (id)copyWithZone:(NSZone *)_zone {
  return [[NGHashMap allocWithZone:_zone] initWithHashMap:self];    
}

- (id)mutableCopyWithZone:(NSZone *)_zone {
  return [[NGMutableHashMap allocWithZone:_zone] initWithHashMap:self];  
}

- (NSMapEnumerator)__keyEnumerator {
  return NSEnumerateMapTable(table);
}

- (LList *)__structForKey:(id)_key {
  return __structForKey(self, _key);
}

// NSCoding

- (void)encodeWithCoder:(NSCoder *)_encoder {
  unsigned        keyCount = [self count];
  NSMapEnumerator mapenum  = [self __keyEnumerator];
  id              key      = nil;
  LList           *value   = NULL;
  
  [_encoder encodeValueOfObjCType:@encode(unsigned) at:&keyCount];

  while (NSNextMapEnumeratorPair(&mapenum, (void **)&key, (void **)&value)) {
    unsigned valueCount = value ? value->count : 0;
    unsigned outCount   = 0; // debugging

    [_encoder encodeObject:key];
    [_encoder encodeValueOfObjCType:@encode(unsigned) at:&valueCount];

    while (value) {
      [_encoder encodeObject:value->object];
      value = value->next;
      outCount++;
    }

    NSAssert(valueCount == outCount, @"didn't encode enough value objects");
  }
}

- (id)initWithCoder:(NSCoder *)_decoder {
  NGMutableHashMap *map = [[NGMutableHashMap alloc] init];
  unsigned keyCount;
  unsigned cnt;

  [_decoder decodeValueOfObjCType:@encode(unsigned) at:&keyCount];
  for (cnt = 0; cnt < keyCount; cnt++) {
    unsigned valueCount = 0, cnt2 = 0;
    id       key        = nil;

    key = [_decoder decodeObject];
    [_decoder decodeValueOfObjCType:@encode(unsigned) at:&valueCount];

    for (cnt2 = 0; cnt2 < valueCount; cnt2++) {
      id value = [_decoder decodeObject];
      [map addObject:value forKey:key];
    }
  }

  self = [self initWithHashMap:map];
  RELEASE(map); map = nil;

  return self;
}

@end

// ************************* NGMutableHashMap ******************

@implementation NGMutableHashMap

+ (id)hashMapWithCapacity:(unsigned int)_numItems {
  return AUTORELEASE([[self alloc] initWithCapacity:_numItems]);
}

- (id)init {
  return [self initWithCapacity:0];
}

// inserting objects

- (void)insertObject:(id)_object atIndex:(unsigned int)_index forKey:(id)_key {
  [self insertObjects:&_object count:1 atIndex:_index forKey:_key];
}

- (void)insertObjects:(NSArray *)_objects atIndex:(unsigned int)_index
               forKey:(id)_key {
  id  *objects = NULL;
  int i        = 0;
  int cntI     = 0;
  
  cntI    = [_objects count];
  objects = NGMalloc(sizeof(id) * cntI);
  for (i = 0 ; i < cntI; i++) 
    objects[i] = [_objects objectAtIndex:i];

  [self insertObjects:objects count:cntI atIndex:_index forKey:_key];
  NGFree(objects);
}

- (void)insertObjects:(id*)_objects count:(unsigned int)_count
              atIndex:(unsigned int)_index forKey:(id)_key {
  id            object  = nil;
  LList *root    = NULL;
  LList *element = NULL;
  int    i              = 0;
  
  if (_count == 0)
    return;

  checkForAddErrorMessage(self, _objects[0],_key);
  if ((root = [self __structForKey:_key]) == NULL) {
    if (_index > 0) {
      IndexOutOfRangeException *exc = nil;
      exc = [[IndexOutOfRangeException alloc] initForSize:0 index:_index];
      [exc setUserInfo:[NSDictionary dictionaryWithObject:self forKey:@"map"]];
      [exc raise];
    }

    root        = initLListElement(_objects[0], NULL);
    root->count = _count;
    NSMapInsert(table, _key, root);
  }
  else {
    if (!(_index < root->count)) {
      IndexOutOfRangeException *exc = nil;
      exc = [[IndexOutOfRangeException alloc]
                                       initForSize:root->count index:_index];
      [exc setUserInfo:[NSDictionary dictionaryWithObject:self forKey:@"map"]];
      [exc raise];
    }
    
    root->count += _count;
    if (_index == 0) {
      element         = initLListElement(_objects[0],NULL);
      object          = element->object;
      element->next   = root->next;
      element->object = root->object;      
      root->object    = object;
      root->next      = element;
    }
    else {
      while (--_index)
        root = root->next;

      element       = initLListElement(_objects[0], NULL);
      element->next = root->next;
      root->next    = element;
      root          = root->next;
    }
  }
  for (i = 1; i < _count; i++) {
    checkForAddErrorMessage(self, _objects[i], _key);
    element       = initLListElement(_objects[i], NULL);
    element->next = root->next;
    root->next    = element;
    root          = element;
  }
}

// adding objects

- (void)addObjects:(id*)_objects count:(unsigned int)_count forKey:(id)_key {
  LList *root     = NULL;
  LList *element  = NULL;
  int           i        = 0;

  if (_count == 0)
    return;

  checkForAddErrorMessage(self, _objects[0],_key);
  if ((root = [self __structForKey:_key]) == NULL) {
    root        = initLListElement(_objects[0], NULL);
    root->count = _count;
    NSMapInsert(table, _key, root);
  }
  else {
    root->count += _count;
    while (root->next)
      root = root->next;
    
    element    = initLListElement(_objects[0], NULL);
    root->next = element;
    root       = root->next;
  }
  for (i = 1; i < _count; i++) {
    checkForAddErrorMessage(self, _objects[i], _key);
    element    = initLListElement(_objects[i], NULL);
    root->next = element;
    root       = element;
  }
}

- (void)addObject:(id)_object forKey:(id)_key {
  checkForAddErrorMessage(self, _object,_key);
  [self addObjects:&_object count:1 forKey:_key];  
}

- (void)addObjects:(NSArray *)_objects forKey:(id)_key {
  id  *objects = NULL;
  int i        = 0;
  int cntI     = 0;
  
  cntI    = [_objects count];
  objects = NGMalloc(sizeof(id) * cntI);
  for (i = 0 ; i < cntI; i++) 
    objects[i] = [_objects objectAtIndex:i];

  [self addObjects:objects count:cntI forKey:_key];
  NGFree(objects);
}

// setting objects

- (void)setObject:(id)_object forKey:(id)_key {
  checkForAddErrorMessage(self, _object, _key);
  [self removeAllObjectsForKey:_key];
  [self addObjects:&_object count:1 forKey:_key];
}

- (void)setObjects:(NSArray *)_objects forKey:(id)_key {
  checkForAddErrorMessage(self, _objects, _key);  
  [self removeAllObjectsForKey:_key];
  [self addObjects:_objects forKey:_key];
}

// removing objects

- (void)removeAllObjects {
  [self __removeAllObjects];
}

- (void)removeAllObjectsForKey:(id)_key {
  [self __removeAllObjectsForKey:_key];
}

- (void)removeAllObjects:(id)_object forKey:(id)_key {
  LList  *list    = NULL;
  LList  *root    = NULL;
  LList  *oldList = NULL;  
  unsigned int  cnt      = 0;

  checkForRemoveErrorMessage(self, _object, _key);
  if (!(root = [self __structForKey:_key])) 
    return;

  while ([root->object isEqual:_object]) {
    RELEASE(root->object);    
    if (!root->next) {
      NGFree(root);
      root = NULL;
      NSMapRemove(table,_key);
      break;
    }
    else {
      list         = root->next;
      root->next   = list->next;
      root->object = list->object;
      root->count--;
      NGFree(list);
      list = NULL;
    }
  }
  if (root) {
    list = root;
    while (list->next) {
      oldList = list;    
      list    = list->next;
      if ([list->object isEqual:_object]) {
        cnt++;
        oldList->next = list->next;
        NGFree(list);
        list = oldList;
      }
    }
    root->count -= cnt;
  }
}

- (void)removeAllObjectsForKeys:(NSArray *)_keyArray {
  int index  = 0;
  for (index = [_keyArray count]; index > 0;)
    [self removeAllObjectsForKey:[_keyArray objectAtIndex:--index]];
}

@end

// ************************* Enumerators ******************

@implementation _NGHashMapKeyEnumerator

- (id)initWithHashMap:(NGHashMap *)_hashMap {
  map        = RETAIN(_hashMap);
  enumerator = [_hashMap __keyEnumerator];
  return self;
}
- (void)dealloc {
  RELEASE(map); map = nil;
  [super dealloc];
}

- (id)nextObject {
  id key, value;
  return NSNextMapEnumeratorPair(&enumerator,(void**)&key, (void**)&value) ?
         key : nil;
}

@end

@implementation _NGHashMapObjectEnumerator

- (id)initWithHashMap:(NGHashMap *)_hashMap {
  keys     = RETAIN([_hashMap keyEnumerator]);
  hashMap  = RETAIN(_hashMap);
  elements = nil;
  return self;
}

- (void)dealloc {
  RELEASE(keys);     keys     = nil;
  RELEASE(hashMap);  hashMap  = nil;
  RELEASE(elements); elements = nil;
  [super dealloc];
}

- (id)nextObject {
  id object = [elements nextObject];
  if (object == nil) {
    id key = nil;
    if ((key = [keys nextObject])) {
      ASSIGN(elements, RETAIN([hashMap objectEnumeratorForKey:key]));
      object = [elements nextObject];
    }
  }
  return object;
}

@end

@implementation _NGHashMapObjectForKeyEnumerator

- (id)initWithHashMap:(NGHashMap *)_hashMap andKey:(id)_key {
  element = [_hashMap __structForKey:_key];

  map = RETAIN(_hashMap);
  return self;
}
- (void)dealloc {
  RELEASE(map); map = nil;
  [super dealloc];
}

- (id)nextObject {
  id object;
  if (!(element)) 
    return nil;

  object  = element->object;
  element = element->next;
  return object;
}

@end
