/* 
   NGStack.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStack.m,v 1.10 1998/12/30 08:56:07 helge Exp $

#import "common.h"
#import "NGStack.h"
#import "NGMemoryAllocation.h"

@interface _NGConcreteStackEnumerator : NSEnumerator
{
  NGStack  *stack; // for retain
  id       *trace;
  unsigned toGo;
  BOOL     downWard; // top=>down
}

- (id)initWithStack:(NGStack *)_stack trace:(id *)_ptr count:(int)_size
  topDown:(BOOL)_downWard;

- (id)nextObject;

@end

@implementation NGStack

+ (id)stackWithCapacity:(unsigned int)_capacity {
  return AUTORELEASE([[self alloc] initWithCapacity:_capacity]);
}
+ (id)stack {
  return AUTORELEASE([[self alloc] init]);
}
+ (id)stackWithArray:(NSArray *)_array {
  return AUTORELEASE([[self alloc] initWithArray:_array]);
}

- (id)init {
  return [self initWithCapacity:256];
}
- (id)initWithCapacity:(unsigned int)_capacity {
  self = [super init];
  if (self) {
    stackPointer = 0;
    capacity     = (_capacity > 0) ? _capacity : 16;

    stack = NGMalloc(sizeof(id) * capacity);
  }
  return self;
}
- (id)initWithArray:(NSArray *)_array {
  register unsigned int count = [_array count];

  if ((self = [self initWithCapacity:(count + 1)])) {
    int cnt;

    for (cnt = 0; cnt < count; cnt++)
      [self push:[_array objectAtIndex:cnt]];
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  if (stack) {
    [self clear];
    NGFree(stack);
  }
  [super dealloc];
}
#endif

// sizing

- (void)_increaseStack {
  if (capacity > 256) capacity += 256;
  else capacity *= 2;

  stack = NGRealloc(stack, sizeof(id) * capacity);
}

// state

- (unsigned int)capacity {
  return capacity;
}
- (unsigned int)stackPointer {
  return stackPointer;
}
- (unsigned int)count {
  return stackPointer;
}
- (BOOL)isEmpty {
  return (stackPointer == 0);
}

// operations

- (void)push:(id)_obj {
  stackPointer++;
  if (stackPointer >= capacity) [self _increaseStack];
  stack[stackPointer] = RETAIN(_obj);
}

- (id)pop {
  id obj = stack[stackPointer];
  if (stackPointer <= 0) {
    [[[NGStackException alloc] initWithReason:
              @"tried to pop an object from an empty stack !"] raise];
  }
  stack[stackPointer] = nil;
  stackPointer--;
  AUTORELEASE(obj);
  return obj;
}

- (void)clear {
  int cnt;
  for (cnt = 1; cnt <= stackPointer; cnt++) {
#if !LIB_FOUNDATION_BOEHM_GC
    RELEASE(stack[cnt]);
#endif
    stack[cnt] = nil;
  }
  stackPointer = 0;
}

// elements

- (id)elementAtTop {
  return (stackPointer == 0) ? nil : stack[stackPointer];
}
- (id)elementAtBottom {
  return (stackPointer == 0) ? nil : stack[1];
}

- (NSEnumerator *)topDownEnumerator {
  if (stackPointer == 0)
    return nil;

  return AUTORELEASE([[_NGConcreteStackEnumerator alloc]
                        initWithStack:self trace:&(stack[stackPointer])
                          count:stackPointer topDown:YES]);
}
- (NSEnumerator *)bottomUpEnumerator {
  if (stackPointer == 0)
    return nil;

  return AUTORELEASE([[_NGConcreteStackEnumerator alloc]
                        initWithStack:self trace:&(stack[1])
                          count:stackPointer topDown:NO]);
}

// NSCoding

- (Class)classForCoder {
  return [NGStack class];
}

- (void)encodeWithCoder:(NSCoder *)_encoder {
  int cnt;
  
  [_encoder encodeValueOfObjCType:@encode(unsigned int) at:&capacity];
  [_encoder encodeValueOfObjCType:@encode(unsigned int) at:&stackPointer];

  for (cnt = 1; cnt <= stackPointer; cnt++) {
    id obj = stack[cnt];
    [_encoder encodeObject:obj];
  }
}

- (id)initWithCoder:(NSCoder *)_decoder {
  int tmpCapacity;
  int tmpStackPointer;

  [_decoder decodeValueOfObjCType:@encode(unsigned int) at:&tmpCapacity];
  [_decoder decodeValueOfObjCType:@encode(unsigned int) at:&tmpStackPointer];

  self = [self initWithCapacity:tmpCapacity];
  {
    register int cnt;

    for (cnt = 1; cnt <= tmpStackPointer; cnt++) {
      id obj = [_decoder decodeObject];
      stack[cnt] = [obj retain];
    }
    stackPointer = tmpStackPointer;
  }
  return self;
}

// copying

- (id)copyWithZone:(NSZone *)_zone {
  register NGStack *newStack = nil;
  register int cnt;

  newStack = [[NGStack allocWithZone:(_zone ? _zone : NSDefaultMallocZone())]
                       initWithCapacity:[self stackPointer]];

  for (cnt = 1; cnt <= stackPointer; cnt++)
    [newStack push:stack[cnt]];

  return newStack;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<%@[0x%08X] capacity=%u SP=%u count=%u content=%s>",
                     NSStringFromClass([self class]), (unsigned)self,
                     [self capacity], [self stackPointer], [self count],
                     [[[self toArray] description] cString]];
}

- (NSArray *)toArray {
  register NSMutableArray *array = nil;
  register int cnt;

  array = [[NSMutableArray alloc] initWithCapacity:stackPointer];

  for (cnt = 1; cnt <= stackPointer; cnt++)
    [array addObject:stack[cnt]];

  return AUTORELEASE(array);
}

@end

@implementation _NGConcreteStackEnumerator

- (id)initWithStack:(NGStack *)_stack trace:(id *)_ptr count:(int)_size
  topDown:(BOOL)_downWard {

  stack    = RETAIN(_stack);
  trace    = _ptr;
  toGo     = _size;
  downWard = _downWard;

  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(stack); stack = nil;
  trace = NULL;
  [super dealloc];
}
#endif

- (id)nextObject {
  id result = nil;
  
  if (toGo == 0)
    return nil;

  toGo--;

  result = *trace;
  
  if (downWard) trace--; // top=>bottom (downward)
  else          trace++; // bottom=>top (upward)

  return result;
}

@end

@implementation NGStackException
@end

@implementation NSMutableArray(Stack)

// state

- (unsigned int)stackPointer {
  return ([self count] - 1);
}

- (BOOL)isEmpty {
  return ([self count] == 0) ? YES : NO;
}

// operations

- (void)push:(id)_obj {
  [self addObject:_obj];
}
- (id)pop {
  unsigned lastIdx = ([self count] - 1);

  if (lastIdx >= 0) {
    id element = [self objectAtIndex:lastIdx];
    [self removeObjectAtIndex:lastIdx];
    return element;
  }
  else {
    [[[NGStackException alloc] initWithReason:
              @"tried to pop an object from an empty stack !"] raise];
    return nil;
  }
}

- (void)clear {
  [self removeAllObjects];
}

// elements

- (id)elementAtTop {
  return [self lastObject];
}

- (NSEnumerator *)topDownEnumerator {
  return [self reverseObjectEnumerator];
}
- (NSEnumerator *)bottomUpEnumerator {
  return [self objectEnumerator];
}

@end

void __link_NGExtensions_NGStack() {
  __link_NGExtensions_NGStack();
}
