/* 
   NGTool.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGTool.m,v 1.15 1999/01/29 11:22:33 helge Exp $

#import "common.h"
#import "NGTool.h"

NSString *NGProcessWillExitNotificationName  = @"NGProcessWillExit";
NSString *NGProcessWillAbortNotificationName = @"NGProcessWillAbort";

@interface NGTool(PrivateMethods)
#if !WIN32
- (void)handleSignal:(int)_signal;
#endif
@end

@implementation NGTool

static NGTool *tool = nil;

+ (NGTool *)tool {
  return tool;
}

#if !WIN32
static void toolSignalHandler(int _signal) {
  [tool handleSignal:_signal];
  signal(_signal, toolSignalHandler);
}
#endif

static inline void logFoundationVersion(void) {
}

- (id)initWithArguments:(char **)_argv count:(int)_argc environment:(char **)_env {
  if ((self = [super init])) {
    tool = self;
    
    tzset();
    self->argv = _argv;
    
#if LIB_FOUNDATION_LIBRARY
    [NSProcessInfo initializeWithArguments:_argv count:_argc environment:_env];
#endif

    self->processInfo    = RETAIN([NSProcessInfo processInfo]);
    self->executablePath = [[NSString alloc] initWithCString:_argv[0]];

    logFoundationVersion();
    self->func = NULL;
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->processInfo);
  RELEASE(self->executablePath);
  RELEASE(self->options);
  RELEASE(self->files);
  [super dealloc];
  tool = nil;
}
#endif

// accessors

- (void)setToolFunction:(NGToolFunction)_func {
  self->func = _func;
}

// signals

#if !WIN32
- (void)registerSignalHandlers {
  signal(SIGPIPE, toolSignalHandler);
  signal(SIGCHLD, toolSignalHandler);
  signal(SIGINT,  toolSignalHandler);
}
- (void)handleSignal:(int)_signal {
  switch (_signal) {
    case SIGPIPE:
      NSLog(@"signal: received SIGPIPE");
      break;

    case SIGCHLD:
      NSLog(@"signal: received SIGCHLD");
      break;

    case SIGINT:
      NSLog(@"signal: received SIGINT");
      [self prepareForExit];
      [self exitWithCode:10];
      break;
      
    default:
      NSLog(@"%s: catched signal number %i, ignoring ..",
            __PRETTY_FUNCTION__, _signal);
      break;
  }
}
#endif /* !WIN32 */

// starting

- (void)handleException:(NSException *)_exception {
  [self prepareForAbort];
  {
    fprintf(stderr, "tool catched: %s\n",
            [[_exception description] cString]);
    fflush(stderr);
  }
  [self coreDump];
}

- (void)run {
#if !OPENSTEP
  [NSAutoreleasePool enableDoubleReleaseCheck:YES];
#endif
  
  if (self->func == NULL) {
    [self prepareForAbort];
    fprintf(stderr, "abort: no tool function set ..\n");
    fflush(stderr);
    [self coreDump];
  }
  else {
    NSAutoreleasePool *pool = [NSAutoreleasePool new];

#if !WIN32
    [self registerSignalHandlers];
#endif

    NS_DURING {
      NSArray *args = [[NSProcessInfo processInfo] arguments];

      self->func(self, args, [args count], self->argv);
    }
    NS_HANDLER {
      [self handleException:localException];
    }
    NS_ENDHANDLER;

    RELEASE(pool);
    pool = nil;
  }
}

// exiting & aborting

- (void)exitWithCode:(int)_exitCode {
  exit(_exitCode);
}
- (void)coreDump {
  abort();
}

- (void)prepareForExit {
  [[NSNotificationCenter defaultCenter]
                         postNotificationName:NGProcessWillExitNotificationName
                         object:self];
}
- (void)prepareForAbort {
  [[NSNotificationCenter defaultCenter]
                         postNotificationName:NGProcessWillAbortNotificationName
                         object:self];
}

// misc

- (NSString *)toolName {
  return [processInfo processName];
}

- (NSString *)executablePath {
  return executablePath;
}

// arguments

static inline NGOptionDescription *_lookupOption(NSString *_name,
                                                 NGOptionDescription *_options) {
  NGOptionDescription *track = _options;
  
  while (track->name != nil) {
    if ([track->name isEqualToString:_name])
      return track;
    else if ([track->longName isEqualToString:_name])
      return track;
    
    track++;
  }
  return NULL;
}

static NGOptionDescription *optionDescriptions = NULL;

- (void)parseOptions:(NGOptionDescription *)_optionDescriptions {
  NSArray *args = [[NSProcessInfo processInfo] arguments];
  int     cnt, count = [args count];

  optionDescriptions = _optionDescriptions;

  for (cnt = 1; cnt < count; cnt++) {
    NSString            *arg     = [args objectAtIndex:cnt];
    NGOptionDescription *desc    = _lookupOption(arg, _optionDescriptions);
    IMP                 procFunc = NULL;

    //NSLog(@"processing option %@ ..", arg);

    if (desc == NULL) {
      desc = _lookupOption(@"*", _optionDescriptions);
      if (desc == NULL) {
#if 0
        NSLog(@"could not handle option: %@ (no description available)", arg);
#endif
        continue;
      }
    }

    procFunc = [self methodForSelector:desc->processor];

    if (procFunc == NULL) {
#if 0
      NSLog(@"could not process option: %@ (processor not implemented)",
            desc->name);
#endif
      continue;
    }
    
    if (desc->numberOfArguments == 0)
      procFunc(self, desc->processor, arg, nil, desc);
    else {
      NSMutableArray *values = nil;
      char           *types  = desc->argumentTypes;
      int            cnt2;
      
      values = [NSMutableArray arrayWithCapacity:desc->numberOfArguments + 1];

      for (cnt2 = 0;
           (cnt2 < desc->numberOfArguments) && (cnt < count);
           cnt2++, types++) {
        
        id value = nil;

        cnt++;
        if (cnt >= [args count]) {
          value = nil;
          NSLog(@"missing parameter ..");
        }
        else
          value = [args objectAtIndex:cnt];

        NSAssert(types, @"invalid type field in option ..");
        switch (*types) {
          case NGBoolOption:
            value = [NSNumber numberWithBool:[value boolValue]];
            break;
            
          case NGIntOption:
            value = [NSNumber numberWithInt:[value intValue]];
            break;
            
          default:
            // keep string
            break;
        }

        if (value)
          [values addObject:value];
      }

      procFunc(self, desc->processor, arg, values, desc);
    }
  }

  optionDescriptions = NULL;
}

- (void)registerOption:(NSString *)_option value:(NSArray *)_args
  description:(NGOptionDescription *)_description {

  if (options == nil)
    options = [[NSMutableDictionary alloc] initWithCapacity:16];

  if ([_args count] == 0)
    [options setObject:[NSNumber numberWithBool:YES] forKey:_option];
  else if ([_args count] == 1)
    [options setObject:[_args objectAtIndex:0] forKey:_option];
  else
    [options setObject:_args forKey:_option];
}

- (void)registerFile:(NSString *)_fileName value:(NSArray *)_alwaysNil
  description:(NGOptionDescription *)_description {

  NSAssert([_alwaysNil count] == 0, @"a file option cannot have arguments ..");

  if (files == nil)
    files = [[NSMutableArray alloc] initWithCapacity:16];

  [files addObject:_fileName];
}

- (void)showHelpOption:(NSString *)_option value:(NSArray *)_value
  description:(NGOptionDescription *)_description {

  NGOptionDescription *track = optionDescriptions;
  unsigned maxOptWidth = 0;

  [self prepareForExit];
  
  printf("Usage: %s <options>\n", [[self toolName] cString]);

  // determine max option width
  while (track->name != nil) {
    char *types = track->argumentTypes;
    int optionWidth = 0;

    if (types) {
      while (*types) {
        switch (*types) {
          case NGIntOption:  optionWidth += 6; break;
          case NGBoolOption: optionWidth += 7; break;
          default:           optionWidth += 9; break;
        }
        types++;
      }
    }

    if (optionWidth > maxOptWidth)
      maxOptWidth = optionWidth;

    track++;
  }
  
  track = optionDescriptions;
  
  while (track->name != nil) {
    char *types = track->argumentTypes;
    char *n, *ln;
    int  optionWidth = 0;

    n  = (char *)[track->name cString];
    ln = (char *)[track->longName cString];
    
    printf("  %-4s %-16s ", n ? n : "", ln ? ln : "");

    if (types) {
      while (*types) {
        switch (*types) {
          case NGIntOption:  printf(" <int>");    optionWidth += 6; break;
          case NGBoolOption: printf(" <bool>");   optionWidth += 7; break;
          default:           printf(" <string>"); optionWidth += 9; break;
        }
        types++;
      }
    }

    // indent
    {
      int i;
      for (i = maxOptWidth; i >= optionWidth; i--)
        fputc(' ', stdout);
    }

    printf(" [%s]\n", [track->helpText cString]);
    
    track++;
  }
  
  [self exitWithCode:0];
}

- (id)valueOfOption:(NSString *)_optionName {
  return [options objectForKey:_optionName];
}
- (BOOL)isFlagOptionSet:(NSString *)_optionName {
  return [[options objectForKey:_optionName] boolValue];
}
- (BOOL)isOptionSet:(NSString *)_optionName {
  return ([options objectForKey:_optionName] != nil);
}

- (int)intValueOfOption:(NSString *)_optionName {
  return [[options objectForKey:_optionName] intValue];
}

- (NSArray *)fileArguments {
  return files;
}

@end
