/* 
   NSConcreteAttributedString.m

   Copyright (C) 1998 MDlink online service center
   All rights reserved.

   Author: Jan Reichmann <jan@mdlink.de>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSConcreteAttributedString.m,v 1.5 1998/12/30 08:56:07 helge Exp $

#import "common.h"
#import "NSAttributedString.h"
#import "NSConcreteAttributedString.h"
#import "NSAttributedStringTypeDefinitions.h"
#import "NSShortAttributedStringFunctions.h"
#import "NSLongAttributedStringFunctions.h"

static inline BOOL isNSConcreteShortAString(id object) {
  if ([object conformsToProtocol:@protocol(_NSShortAttributedString)])
    return YES;
  return NO;
}

@interface _NSConcreteShortAttributedString(private) 

- (id)__initWithShortStringStruc:(NSShortAttributedString *)_string;
- (NSShortAttributedString *)__shortStringStruct;

@end

@interface _NSConcreteShortAttributedString(description)

- (NSString *)descriptionWithMatrixLength:(int)_length;
- (NSString *)simpleDescription;
- (NSString *)descriptionWithTags;

@end

/*
  _NSConcreteShortAttributedString

  ShortAttributedString use the NSShortAttributedString struct for storing the
  data.
  NSShortAttributedString->attributeDicts is an array[0..255] of id's.
  It contains all differnt attribute combinations (stored as NSDictionary).
  
  NSShortAttributedString->lastAttributeEntry is the last entry in
  attributeDicts.
  NSShortAttributedString->attributes is a char field with the size of
  [string length] which pointed in attributeDicts.
*/

/****************_NSConcreteShortAttributedString**********/


@implementation _NSConcreteShortAttributedString

- (id)__initWithShortStringStruc:(NSShortAttributedString *)_string {
  if ((self = [super init])) 
    attributedString = _string;

  return self;
}

- (id)initWithString:(NSString *)_string attributes:(NSDictionary *)_attributes {
  NSShortAttributedString *tmpAString = NULL;
  
  tmpAString = NSCreateShortAStringFromString(_string, _attributes, NULL);
  return [self __initWithShortStringStruc:tmpAString];
}

- (id)initWithAttributedString:(NSAttributedString *)_attributedString {
  NSShortAttributedString *tmpAString = NULL;
  
  if (isNSConcreteShortAString(_attributedString)) {
    id<_NSShortAttributedString> anObject = _attributedString;

    tmpAString =
      NSCreateShortAStringWithShortAString([anObject __shortStringStruct], NULL);
  }
  else {
    [[[NSAttributedStringException alloc] initWithFormat:
              @"unsupported NSConcreteAttributedString %@ \n",
              [_attributedString class]]
              raise];
  }
  
  return [self __initWithShortStringStruc:tmpAString];
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  if (attributedString) {
    NSFreeShortAttributedString(attributedString);
    attributedString = NULL;
  }
  [super dealloc];
}
#endif


- (NSShortAttributedString *)__shortStringStruct {
  return attributedString;
}

- (NSString *)string {
  return attributedString->string;
}

- (unsigned int)length {
  return [attributedString->string length];
}

// Retrieving attribute information
- (id)attribute:(NSString *)_attributeName
      atIndex:(unsigned int)_index
      effectiveRange:(NSRange *)_range {
  return [NSShortAllAttributesAtIndex(attributedString, _index, _range, NULL)
                                     objectForKey:_attributeName];
}

- (id)attribute:(NSString *)_attributeName
      atIndex:(unsigned int)_index
      longestEffectiveRange:(NSRange *)_range
      inRange:(NSRange)_rangeLimit {
  return NSShortAttributeAtIndex(attributedString, _attributeName,
                                 _index, _range, &_rangeLimit);
}

- (NSAttributedString *)attributedSubstringFromRange:(NSRange)_range {
  NSAttributedString *astr = nil;

  astr = [[NSAttributedString alloc]
                              __initWithShortStringStruc:
                                NSShortAttributedSubstring(attributedString,
                                                           _range)];
  return AUTORELEASE(astr);
}


- (NSDictionary *)attributesAtIndex:(unsigned int)_index
                  effectiveRange:(NSRange *)_range {
  return NSShortAllAttributesAtIndex(attributedString, _index, _range, NULL);
}
 
- (NSDictionary *)attributesAtIndex:(unsigned int)_index
                  longestEffectiveRange:(NSRange *)_range
                  inRange:(NSRange)_rangeLimit {
  return NSShortAllAttributesAtIndex(attributedString,
                                     _index,
                                     _range,
                                     &_rangeLimit);
}

- (BOOL)isEqualToAttributedString:(NSAttributedString *)_otherString {
  if (isNSConcreteShortAString(_otherString)) {
    id <_NSShortAttributedString> anObject = _otherString;
    
    return NSCompareShortAttributedStrings(attributedString,
                                           [anObject __shortStringStruct]);
  }
  return NO;
}

//  NSCopying

- (id)copyWithZone:(NSZone *)_zone {
  return AUTORELEASE([[_NSConcreteShortAttributedString alloc]
                                 __initWithShortStringStruc:
                                       NSCreateShortAStringWithShortAString(
                                                  attributedString, _zone)]);
}
 
- (id)mutableCopyWithZone:(NSZone *)_zone {
  return AUTORELEASE([[_NSConcreteShortMutableAttributedString alloc]
                                              __initWithShortStringStruc:
                                      NSCreateShortAStringWithShortAString(
                                              attributedString, _zone)]);
}

// Description

- (NSString *)description {
  return [self descriptionWithMatrixLength:100];
}

- (NSString *)descriptionWithMatrixLength:(int)_length{
  NSLongAttributedString *longString   = NULL;
  NSString               *description  = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeWithMatrixLongAttributedString(longString,_length);
  NSFreeLongAttributedString(longString);
  return description;
}

- (NSString *)simpleDescription {
  NSLongAttributedString *longString  = NULL;
  NSString               *description = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeSimpleLongAttributedString(longString);
  NSFreeLongAttributedString(longString);
  return description;
}

- (NSString *)descriptionWithTags {
  NSLongAttributedString *longString  = NULL;
  NSString               *description = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeWithTagsLongAttributedString(longString);
  NSFreeLongAttributedString(longString);
  return description;
}
// Coding

- (void)encodeWithCoder:(NSCoder *)_encoder {
  NSLongAttributedString *longString = NULL;
  longString = NSCreateLongAStringWithShortAString(attributedString);
  NSLongAttributedStringEncodeWithCoder(longString, _encoder);
  NSFreeLongAttributedString(longString);
}

- (id)initWithCoder:(NSCoder *)_decoder {
  NSLongAttributedString  *longString  = NULL;
  NSShortAttributedString *shortString = NULL;
  
  longString  = NSLongAttributedStringDecodeWithCoder(_decoder);
  shortString = NSCreateShortAStringWithLongAString(longString);
  NSFreeLongAttributedString(longString);
  return AUTORELEASE([[_NSConcreteShortAttributedString alloc]
                                                  __initWithShortStringStruc:
                                                               shortString]);
}

@end /* _NSConcreteShortAttributedString */

/****************_NSConcreteShortMutableAttributedString**********/

@interface _NSConcreteShortMutableAttributedString(private) 

- (id)__initWithShortStringStruc:(NSShortAttributedString *)_string;
- (NSShortAttributedString *)__shortStringStruct;

@end

@interface _NSConcreteShortMutableAttributedString(description)

- (NSString *)descriptionWithMatrixLength:(int)_length;
- (NSString *)simpleDescription;
- (NSString *)descriptionWithTags;

@end

@implementation _NSConcreteShortMutableAttributedString

+ (void)initialize {
  static BOOL isInitialized = NO;
  
  if(!isInitialized) {
    isInitialized = YES;
    class_add_behavior(self, [_NSConcreteShortAttributedString class]);
  }
}

- (id)__initWithShortStringStruc:(NSShortAttributedString *)_string {
  if ((self = [super init])) 
    attributedString = _string;

  return self;
}

- (NSShortAttributedString *)__shortStringStruct {
  return attributedString;
}

// Description
- (NSString *)descriptionWithMatrixLength:(int)_length{
  NSLongAttributedString *longString   = NULL;
  NSString               *description  = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeWithMatrixLongAttributedString(longString,_length);
  NSFreeLongAttributedString(longString);
  return description;
}

- (NSString *)simpleDescription {
  NSLongAttributedString *longString  = NULL;
  NSString               *description = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeSimpleLongAttributedString(longString);
  NSFreeLongAttributedString(longString);
  return description;
}

- (NSString *)descriptionWithTags {
  NSLongAttributedString *longString  = NULL;
  NSString               *description = nil;

  longString  = NSCreateLongAStringWithShortAString(attributedString);
  description = NSDescribeWithTagsLongAttributedString(longString);
  NSFreeLongAttributedString(longString);
  return description;
}


- (NSMutableString *)mutableString {
  return [attributedString->string mutableCopy];
}

// Changing characters

-(void)replaceCharactersInRange:(NSRange)aRange withString:(NSString *)aString {
 NSShortReplaceWithString(attributedString, aRange, aString); 
}

- (void)deleteCharactersInRange:(NSRange)aRange {
  NSShortReplaceWithString(attributedString, aRange, @"");
}

// Changing attributes

- (void)setAttributes:(NSDictionary *)attributes range:(NSRange)aRange {
  NSShortSetAttributes(attributedString, aRange, attributes);
}

- (void)addAttribute:(NSString *)name value:(id)value range:(NSRange)aRange {
  NSShortAddAttributes(attributedString, aRange,
                       [NSDictionary dictionaryWithObjects:&value
                                     forKeys:&name
                                     count:1]);  
}

- (void)addAttributes:(NSDictionary *)attributes range:(NSRange)aRange {
  NSShortAddAttributes(attributedString, aRange,attributes);
}
- (void)removeAttribute:(NSString *)name range:(NSRange)aRange {
  NSShortRemoveAttribute(attributedString, aRange, name);
}

// Changing characters and attributes

- (void)appendAttributedString:(NSAttributedString *)_attributedString {
  if (isNSConcreteShortAString(_attributedString)) {
    id <_NSShortAttributedString> anObject = _attributedString;
    
    NSShortReplaceWithAttributedString(attributedString,
                                       NSMakeRange([attributedString->
                                                    string length],
                                                   0),
                                       [anObject __shortStringStruct]);
  }
}
- (void)insertAttributedString:(NSAttributedString *)_attributedString
        atIndex:(unsigned int)index {
  if (isNSConcreteShortAString(_attributedString)) {
    id <_NSShortAttributedString> anObject = _attributedString;
    
    NSShortReplaceWithAttributedString(attributedString, NSMakeRange(index, 0),
                                       [anObject __shortStringStruct]);
  }
}

- (void)replaceCharactersInRange:(NSRange)aRange
        withAttributedString:(NSAttributedString *)_attributedString {
  if (isNSConcreteShortAString(_attributedString)) {
    id <_NSShortAttributedString> anObject = _attributedString;    
    NSShortReplaceWithAttributedString(attributedString, aRange,
                                       [anObject __shortStringStruct]);
  }
}

- (void)setAttributedString:(NSAttributedString *)_attributedString {
  if (isNSConcreteShortAString(_attributedString)) {
    id <_NSShortAttributedString> anObject = _attributedString;
    
    NSShortReplaceWithAttributedString(attributedString,
                                       NSMakeRange(0,
                                                   [attributedString->
                                                    string length]),
                                       [anObject __shortStringStruct]);
  }
}

@end /* _NSConcreteShortMutableAttributedString */

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
