/* 
   NSLongAttributedStringFunctions.m

   Copyright (C) 1998 MDlink online service center
   All rights reserved.

   Author: Jan Reichmann <jan@mdlink.de>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
//$Id: NSLongAttributedStringFunctions.m,v 1.5 1998/12/30 08:56:07 helge Exp $

#import "common.h"
#import "NSLongAttributedStringFunctions.h"

#import "NSException+misc.h"
#import "NGMemoryAllocation.h"
#import "NSAttributedString.h"
#import "NGHashMap.h"

#define COPY 0

#ifndef MAX
#define MAX(x, y) ((x < y) ? y : x)
#endif

static inline void raiseIndexOutOfBoundException(int _index, int _bounds) {
  [[[NSAttributedStringException alloc] initWithFormat:
                                          @"Index %i is out of bounds (idx=%i)",
                                          _index, _bounds]
                                 raise];
}


NSDictionary *NSShortAllAttributesAtIndex(
  NSShortAttributedString *_string,
  unsigned int            _index,
  NSRange                 *_longestEffectiveRange,
  NSRange                 *_rangeLimit);

NSLongAttributedString
*NSCreateLongAStringWithShortAString(NSShortAttributedString *_string) {
  int                    i             = 0;
  int                    strLen        = 0;  
  NSLongAttributedString *longString   = NULL;
  NSLongASListElement    *startElement = NULL;
  NSLongASListElement    *element      = NULL;  
  NSLongASListElement    *oldElement   = NULL;  
  NSDictionary           *dict         = nil;
  NSRange                *range        = NULL;

  longString = NGMalloc(sizeof(NSLongAttributedString));
  
  strLen = [_string->string length];
  for (i = 0; i < strLen;) {
    element = NGMalloc(sizeof(NSLongASListElement));
    if (i == 0)
      startElement = element;
    
    range = NGMalloc(sizeof(NSRange));
    dict = NSShortAllAttributesAtIndex(_string, i, range, NULL);
    #if COPY
    {
      element->dict = RETAIN([dict copy]);
    }
    #else
    {
      element->dict = RETAIN(dict);
    }
    #endif
    element->range  = range;
    if (i == 0) {
      element->prev = NULL;
    }
    else {
      element->prev    = oldElement;
      oldElement->next = element;
    }
    element->next = NULL;
    oldElement = element;
    i = range->location + range->length;
  }
  longString->attributes = startElement;
  #if COPY
  {
    longString->string = RETAIN([_string->string copy]);
  }
  #else
  {
    longString->string = RETAIN(_string->string);
   }
  #endif
  longString->zone = _string->zone;
  return longString;
}

 void NSFreeLongAttributedString(NSLongAttributedString *_string) {
  NSLongASListElement *element    = NULL;
  NSLongASListElement *tmpElement = NULL;  
  RELEASE(_string->string);
  element = _string->attributes;
  while (element) {
    tmpElement = element;
    element = element->next;    
    tmpElement->next = NULL;
    tmpElement->prev = NULL;
    RELEASE(tmpElement->dict);
    NGFree(tmpElement);
    tmpElement = NULL;
  }
  NGFree(_string);
}

NSMutableDictionary *NSLongAllAttributesAtIndex(NSLongAttributedString *_string,
                                                unsigned int _index) {
  NSLongASListElement *element = NULL;

  if (_index > [_string->string length])
    raiseIndexOutOfBoundException(_index, [_string->string length]);
  
  element = _string->attributes;
  while (element) {
    if ((element->range->location <= _index) &&
        (element->range->length + element->range->location > _index))
      return (NSMutableDictionary *)element->dict ;
    element = element->next;
  }
  return NULL;
}

NSString *NSDescribeWithTagsLongAttributedString(NSLongAttributedString *_string) {
  NSMutableString     *mString     = nil;
  NSString            *string      = nil;  
  NSLongASListElement *element     = NULL;
  NSMutableDictionary *newDict     = nil;
  NSMutableDictionary *oldDict     = nil;
  NSMutableDictionary *dict        = nil;  
  NSEnumerator        *dictKeys    = nil;
  id                  value        = nil;
  
  dict        = [[NSMutableDictionary alloc] initWithCapacity:10];
  mString     = [[NSMutableString alloc] initWithCapacity:100];
  element     = _string->attributes;

  while (element) {
    oldDict  = [[NSMutableDictionary alloc] initWithCapacity:10];
    dictKeys = [dict keyEnumerator];
    while ((value = [dictKeys nextObject])) {
      if (!([[dict objectForKey:value] isEqual:[element->dict objectForKey:
                                                       value]])) {
        [oldDict setObject:[dict objectForKey:value] forKey:value];
        [dict removeObjectForKey:value];
      }
    }
    newDict   = [[NSMutableDictionary alloc] initWithCapacity:10];
    dictKeys  = [element->dict keyEnumerator];
    while ((value = [dictKeys nextObject])) {
      if (!([[element->dict objectForKey:value] isEqual:
                                                [dict objectForKey:value]])) {
        [newDict setObject:[element->dict objectForKey:value] forKey:value];
        [dict setObject:[element->dict objectForKey:value] forKey:value];
      }
    }
    // Build ReturnString
    [mString appendString:@"{"];
    dictKeys = [oldDict keyEnumerator];
    while ((value = [dictKeys nextObject])) {
      [mString appendString:[NSString stringWithFormat:@"/(%@ = %@), ",value,
                                      [oldDict objectForKey:value]]];   
    }
    dictKeys = [newDict keyEnumerator];
    while ((value = [dictKeys nextObject])) {
      [mString appendString:[NSString stringWithFormat:@"(%@ = %@), ",value,
                                      [newDict objectForKey:value]]];   
    }
    if ([mString length] > 1)     
      [mString deleteCharactersInRange:NSMakeRange([mString length] - 2, 2)];
    [mString appendString:@"} "];    
    [mString appendString:[_string->string substringWithRange:
                                  *(element->range)]];
    [mString appendString:@" "];    
    element = element->next;
  }
  [mString appendString:@"{"];
  dictKeys = [dict keyEnumerator];
  while ((value = [dictKeys nextObject])) {
    [mString appendString:[NSString stringWithFormat:@"/(%@ = %@), ",value,
                                    [dict objectForKey:value]]];   
  }
  if ([mString length] > 1) 
    [mString deleteCharactersInRange:NSMakeRange([mString length] - 2, 2)];  
  [mString appendString:@"}\n"];
  
  string = [mString copy];
  AUTORELEASE(mString);
  return string;
}

NSString *__NSSimpleDictDescription(NSDictionary *_dict) {
  NSString        *string  = nil;
  NSMutableString *mString = nil;
  NSEnumerator    *keyEnum = nil;
  id              value    = nil;

  keyEnum = [_dict keyEnumerator];
  mString = [[NSMutableString alloc] initWithCapacity:10];
  [mString appendString:@"{"];
  while ((value = [keyEnum nextObject])) {
    [mString appendString:[NSString stringWithFormat:@"(%@ = %@), ", value,
                                    [_dict objectForKey:value]]];
  }
  if ([mString length] > 1) 
    [mString deleteCharactersInRange:NSMakeRange([mString length] - 2, 2)];    
  [mString appendString:@"}"];
  string = [mString copy];
  AUTORELEASE(mString);
  return string;
}

NSString *NSDescribeSimpleLongAttributedString(NSLongAttributedString *_string) {
  NSMutableString     *mString  = nil;
  NSString            *string   = nil;
  int                 maxLength = 0;
  int                 i         = 0;
  NSLongASListElement *element = NULL;

  element = _string->attributes;
  mString = [[NSMutableString alloc] initWithCapacity:10];
  while (element) {
    maxLength = MAX(maxLength, [(string = __NSSimpleDictDescription(element->dict)) length]);
    element = element->next;
  }
  maxLength = maxLength + 2;
  element = _string->attributes;
  while (element) {
    int length = 0;
    length = [(string = __NSSimpleDictDescription(element->dict)) length];
    [mString appendString:string];
    for (i = 0; i < maxLength - length; i++)
      [mString appendString:@" "];
    
    [mString appendString:[_string->string substringWithRange:*element->range]];
    [mString appendString:@"\n"];
    element = element->next;
  }
  string = [mString copy];
  AUTORELEASE(mString);
  return string;
}

 int __NSLongestValue(NSEnumerator *_enumerator) {
  id value     = nil;
  int maxValue = 0;
  
  while ((value = [_enumerator nextObject])) 
    maxValue = MAX(maxValue, [[value stringValue] length]);
  
  return maxValue;
}


BOOL __NSDictIsInRange(NSShortAttributedString *_string,
                       NSRange *_range,
                       char _pos,
                       char *pos_) {
  int  i        = 0;
  int  rangePos = 0;
  BOOL status   = NO;
  rangePos = _range->length + _range->location;
  for (i = 0; i < rangePos; i++) {
    if (_string->attributes[i] == _pos) {
      status = YES;
      pos_[i] = 1;
    }
  }
  return status;
}

NSString *NSDescribeWithMatrixLongAttributedString(NSLongAttributedString *_string,
                                                   int _length) {
  int                 maxKey    = 0;
  int                 maxValue  = 0;
  int                 strLen    = 0;
  int                 count     = 0;
  int                 maxLength = 0;
  int                 i         = 0;  
  NSLongASListElement *element  = NULL;
  NGMutableHashMap    *hashMap  = nil;
  NSMutableString     *mString  = nil;
  NSString            *string   = nil;
  
  strLen = [_string->string length];
  element = _string->attributes;
  while (element) {
    maxKey   = MAX(maxKey  , __NSLongestValue([element->dict keyEnumerator]));
    maxValue = MAX(maxValue, __NSLongestValue([element->dict objectEnumerator]));
    element = element->next;
  }
  maxValue++;
  maxKey++;

  count     = 0;
  maxLength = 0;
  element   = _string->attributes;
  mString   = [[NSMutableString alloc] initWithCapacity:100];  
  while (_length * count < strLen) {
    BOOL nextElementInRange = YES;


    hashMap = [[NGMutableHashMap alloc] initWithCapacity:10];
    while (nextElementInRange && element) {
      if ((element->range->location + element->range->length) >
            (_length * count + _length))
        nextElementInRange = NO;
      {
        NSEnumerator *enumerator = nil;
        id           key         = nil;
        id           value       = nil;
        
        enumerator = [element->dict keyEnumerator];
        while ((key = [enumerator nextObject]))
          if (!( [[hashMap objectsForKey:key] containsObject:
                                              (value = [element->dict objectForKey:
                                                               key])]))
            [hashMap addObject:value forKey:key];

        if (nextElementInRange)
          element = element->next;        
      }
    }
    // build String
    if (count > 0)
      [mString appendString:@"\n"];
    {
      NSEnumerator *keyEnumerator    = nil;
      NSEnumerator *objectEnumerator = nil;      
      id           key               = nil;
      id           object            = nil;
      BOOL         status            = NO;

      keyEnumerator = [hashMap keyEnumerator];
      while ((key = [keyEnumerator nextObject])) {
        [mString appendString:[NSString stringWithFormat:@"%@",key]];
        for (i = 0; i < (maxKey - [key length]); i++)
          [mString appendString:@" "];
        
        objectEnumerator = [hashMap objectEnumeratorForKey:key];
        status = NO;
        while ((object = [objectEnumerator nextObject])) {

          if (!status)
            status = YES;
          else 
            for (i = 0; i < maxKey; i++)
              [mString appendString:@" "];

          [mString appendString:[NSString stringWithFormat:@"%@",object]];
          for (i = 0; i < (maxValue - [object length]); i++)
            [mString appendString:@" "];


          for (i = _length * count; i < _length * count + _length; i++) {
            NSDictionary *dict        = nil;
            
            if (i == [_string->string length])
              break;
            
            dict = NSLongAllAttributesAtIndex(_string, i);
            if ([[dict objectForKey:key] isEqual:object])
              [mString appendString:@"#"];
            else
              [mString appendString:@" "];              
          }
          [mString appendString:@"\n"];          
        }
      }
    }
    for (i = 0; i < maxKey + maxValue; i++)
      [mString appendString:@" "];

    {
      int rangeLength = 0;
      if (_length * count + _length > [_string->string length])
        rangeLength = [_string->string length] - _length * count;
      else
        rangeLength = _length;
      [mString appendString:[_string->string substringWithRange:
                                    NSMakeRange(_length * count,
                                                rangeLength)]];
    }
    [mString appendString:@"\n"];    
    count++;
  }
  string  = [mString copy];
  RELEASE(mString);
  return string;
}

void NSLongAttributedStringEncodeWithCoder(NSLongAttributedString *_string,
                                           NSCoder *_encoder) {
  unsigned int        countList = 0;
  NSLongASListElement *element  = NULL;

  element = _string->attributes;
  countList = 0;
  while (element) {
    countList++;
    element = element->next;
  }
  element = _string->attributes;
  [_encoder encodeValueOfObjCType:@encode(unsigned int) at:&countList];
  while (element) {
    [_encoder encodeObject:element->dict];
    [_encoder encodeValueOfObjCType:@encode(unsigned int)
              at:&element->range->location];
    [_encoder encodeValueOfObjCType:@encode(unsigned int)
              at:&element->range->length];
    element = element->next;
  }
  [_encoder encodeObject:_string->string];
}

NSLongAttributedString *NSLongAttributedStringDecodeWithCoder(NSCoder *_decoder) {
  unsigned int           countList    = 0;
  NSLongASListElement    *element     = NULL;
  NSLongASListElement    *lastElement = NULL;  
  NSLongAttributedString *string      = NULL;

  string = NGMalloc(sizeof(NSLongAttributedString));
  [_decoder decodeValueOfObjCType:@encode(unsigned int) at:&countList];
  lastElement = NULL;
  {
    register int i = 0;
    for (i = 0; i < countList; i++) {
      NSRange *range;
      element = NGMalloc(sizeof(NSLongASListElement));
      range   = NGMalloc(sizeof(NSRange));
      element->range = range;
      if (i == 0)
        string->attributes = element;

      element->dict = RETAIN([_decoder decodeObject]);
      [_decoder decodeValueOfObjCType:@encode(unsigned int) at:
                &element->range->location];
      [_decoder decodeValueOfObjCType:@encode(unsigned int) at:
                &element->range->length];
      element->prev = lastElement;
      element->next = NULL;
      if (lastElement)
        lastElement->next = element;
      
      lastElement = element;
    }
    string->string = RETAIN([_decoder decodeObject]);
    string->zone   = NULL;
  }
  return string;
}

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
