/* 
   NSShortAttributedStringFunctions.m

   Copyright (C) 1998 MDlink online service center
   All rights reserved.

   Author: Jan Reichmann <jan@mdlink.de>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSShortAttributedStringFunctions.m,v 1.5 1998/12/30 08:56:07 helge Exp $

#import "common.h"
#import "NGMemoryAllocation.h"
#import "NSShortAttributedStringFunctions.h"
#import "NSAttributedString.h"

#ifndef MAX
#define MAX(x, y) ((x < y) ? y : x)
#endif
#ifndef MIN
#define MIN(x, y) ((x > y) ? y : x)
#endif

#define NSAttributedString_COPYMODE 0

static inline void raiseToManyAttributeCombinationsException() {
  [[[NSAttributedStringException alloc] initWithReason:
       @"more than 256 attribute combinations"] raise];
}

static inline void raiseRangeOutOfBoundException(int _index, NSRange _range) {
  [[[NSAttributedStringException alloc] initWithReason:
                               [NSString stringWithFormat:@"Range %@ is out "
                                                          @"of StringLength %i ",
                                         NSStringFromRange(_range),
                                         _index]] raise];
}

static inline void raiseIndexOutOfBoundException(int _index, int _int) {
  [[[NSAttributedStringException alloc] initWithReason:
                                 [NSString stringWithFormat:@"Index %i is out "
                                                            @"of bound %i ",
                                           _index,
                                           _int]] raise];
}
static inline void raiseIndexOutOfRangeBoundException(int _index, NSRange _range) {
  [[[NSAttributedStringException alloc] initWithReason:
                            [NSString stringWithFormat:@"Index %i is out "
                                                       @"of RangeBound %@ ",
                                      _index,
                                      NSStringFromRange(_range)]]raise];
}

static inline BOOL isInRange(NSRange *_range, int _int) {
  if ((_range->location <= _int) && ((_range->location + _range->length) >= _int))
    return YES;
  return NO;
}



static inline BOOL __compareDictArray(NSShortAttributedString *_attrString,
                        NSDictionary *_dict,
                        int *pos) {
  int i = 0;
  int countI = _attrString->lastAttributeEntry;
  for (i = 0; i <= countI; i++) {
    if ([_attrString->attributeDicts[i] isEqualToDictionary:_dict]) {
      *pos = i;
      return YES;
    }
  }
  return NO;
}

static inline NSArray *__allDictsForString(NSShortAttributedString *_string) {
  NSMutableArray *array = nil;
  int i, countI;
  countI = _string->lastAttributeEntry;
  array = [[NSMutableArray alloc] initWithCapacity:10];
  for (i = 0; i <= countI; i++) {
    [array addObject:_string->attributeDicts[i]];
  }
  return AUTORELEASE(array);
}

inline void __removeUnusedCombinations(NSShortAttributedString *_string) {
  int  i         = 0;
  int  last      = 0;
  int  strLen    = 0;
  char *attr     = NULL;
  char *tmpTable = NULL;
  
  strLen   = [_string->string length];
  last     = _string->lastAttributeEntry;  
  tmpTable = NGMalloc((sizeof(char) * last) + 1);
  for (i = 0; i <= last; i++)
    tmpTable[i] = 0;
  
  attr = _string->attributes;  
  for (i = 0; i < strLen; i++) 
    tmpTable[(int)attr[i]] = 1;

  for (i = 0; i <= last; i++)
    if (tmpTable[i] == 0) {
      int k                         = 0;
      NSDictionary *oldDict  = nil;
      NSDictionary *lastDict = nil;

      oldDict  = _string->attributeDicts[i];
      lastDict = _string->attributeDicts[last];
      _string->attributeDicts[last] = nil;
      _string->lastAttributeEntry--;
      RELEASE(oldDict);
      oldDict = nil;
      _string->attributeDicts[i] = lastDict;
      lastDict = nil;
      for (k = 0; k <= strLen; k++)
        if (attr[k] == last)
          attr[k] = i;

      tmpTable[i--] = tmpTable[last--];
    }
  NGFree(tmpTable);
  tmpTable = NULL;
}

// Create ShortAttributedString
NSShortAttributedString *NSCreateShortAStringFromString(NSString *_string,
                                                        NSDictionary *_attr,
                                                        void *_zone) {
  NSShortAttributedString *attrString  = NULL;
  int                     strLen       = 0;
  int                     i            = 0;
  NSDictionary            *dict        = nil;

  if (_string == nil)
    return NULL;
  
  _zone = NULL;
  RETAIN(_string);
  if (_attr) 
    dict = [_attr copy];
  else
    dict = [[NSDictionary alloc] init];

  attrString                     = NGMalloc(sizeof(NSShortAttributedString));
  strLen                         = [_string length];
  attrString->zone               = _zone;
  attrString->attributeDicts     = NGMalloc(sizeof(id)   * 255);
  attrString->attributeDicts[0]  = dict;  
  attrString->string             = _string;
  attrString->attributes         = NGMalloc(sizeof(char) * strLen);
  attrString->lastAttributeEntry = 0;
  for (i = 0; i < strLen; i++)
    attrString->attributes[i] = 0;

  return attrString;
}


NSShortAttributedString *NSCreateShortAStringWithLongAString(
                                                 NSLongAttributedString *_longString) {

  NSShortAttributedString *shortString       = NULL;
  NSLongASListElement     *element           = NULL;
  id                      *attributeDicts    = NULL;
  char                    *attributes        = NULL;
  int                     lastAttributeEntry = 0;
  int                     elementLength      = 0;

  attributes     = NGMalloc(sizeof(char) * [_longString->string length]);
  attributeDicts = NGMalloc(sizeof(id) *256);
  element    = _longString->attributes;
  lastAttributeEntry = -1;
  while (element) {
    int          dictPointer = 0;
    BOOL         status      = NO;
    register int i           = 0;
      
    for (i = 0; i <= lastAttributeEntry; i++) 
      if ([attributeDicts[i] isEqualToDictionary:element->dict]) {
        dictPointer = i;
        status = YES;
        break;
      }
    if (!status) {
      if (dictPointer + 1 == 255) 
        raiseToManyAttributeCombinationsException();

      dictPointer = ++lastAttributeEntry;
      attributeDicts[dictPointer] = [element->dict copy];
    }
    else
      status = NO;
    
    elementLength = element->range->location + element->range->length;
    for (i = element->range->location; i < elementLength; i++)
      attributes[i] = dictPointer;

    element = element->next;
  }
  shortString                     = NGMalloc(sizeof(NSShortAttributedString));
  shortString->zone               = NULL;
  shortString->attributes         = attributes;
  shortString->attributeDicts     = attributeDicts;
  shortString->lastAttributeEntry = lastAttributeEntry;
#if !NSAttributedString_COPYMODE
    shortString->string         = RETAIN(_longString->string);
#else
    shortString->string         = [_longString->string copy];
#endif
  return shortString;  
}

NSShortAttributedString *NSCreateShortAStringWithShortAString(
                                          NSShortAttributedString *_string,
                                          void                    *_zone) {
  NSShortAttributedString *attrString = NULL; 
  int stringLen = 0;
  int i         = 0;
 
  if (_string == NULL) 
    return NULL; 
   
  _zone = NULL; 
   stringLen = [_string->string length];   
 
  attrString                     = NGMalloc(sizeof(NSShortAttributedString)); 
  attrString->zone               = _zone; 
  attrString->attributeDicts     = NGMalloc(sizeof(id)   * 256); 
  attrString->attributes         = NGMalloc(sizeof(char) * stringLen); 
  attrString->lastAttributeEntry = _string->lastAttributeEntry; 
  attrString->string             = [_string->string copy]; 
  for (i = 0; i <= _string->lastAttributeEntry; i++)
    attrString->attributeDicts[i] = [_string->attributeDicts[i] copy];
  for (i = 0; i < stringLen; i++) {
    attrString->attributes[i] = _string->attributes[i];
  }
  return attrString;
}


// Free ShortAttributedString
void NSFreeShortAttributedString(NSShortAttributedString *_string) {
  int i = 0, cntI = 0;
  
  RELEASE(_string->string);
  _string->string = nil;
  _string->zone = nil;  
  cntI = _string->lastAttributeEntry;
  for (i = 0; i <= cntI; i++) {
    RELEASE(_string->attributeDicts[i]);
    _string->attributeDicts[i] = nil;
  }
  NGFree(_string->attributeDicts);
  NGFree(_string->attributes);
  _string->attributes = NULL;  
  NGFree(_string);
  _string = NULL;
}

// Retrieving attribute information


NSDictionary *NSShortAllAttributesAtIndex(
                                 NSShortAttributedString *_string,
                                 unsigned int            _index,
                                 NSRange                 *_longestEffectiveRange_,
                                 NSRange                 *_rangeLimit) {
  char    attrEntry = 0;
  int     i = 0, j = 0, from = 0, to = 0;
  NSRange *range    = NULL;
  BOOL    status    = YES;
  int     strLen    = 0;

  strLen = [_string->string length];
  if (!(_index < strLen))
    raiseIndexOutOfBoundException(_index, strLen - 1);
  attrEntry = _string->attributes[_index];
  if (_longestEffectiveRange_) {
    if (_rangeLimit) {
      if (!isInRange(_rangeLimit, _index))
        raiseIndexOutOfRangeBoundException(_index, *_rangeLimit);
      
      from = _rangeLimit->location;
      to   = _rangeLimit->location + _rangeLimit->length;
    }
    else {
      from = 0;
      to   = strLen;
    }
    i         = _index;
    range     = _longestEffectiveRange_;
    status    = YES;
    do {
      if (i == from) {
        status = NO;
        range->location = i;
      }
      else
        if (_string->attributes[--i] != attrEntry) {
          status          = NO;
          range->location = ++i;
        }
    } while (status);
    status = YES;
    j      = _index;
    do {
      if (j == to) {
        status = NO;
        range->length = j - i;
      }
      else
        if (_string->attributes[++j] != attrEntry) {
          status        = NO;
          range->length = j - i;
        }
    } while (status);
  }
  return _string->attributeDicts[(int)attrEntry];  
}

id NSShortAttributeAtIndex(
                           NSShortAttributedString *_string,
                           NSString                *_attributeName,
                           unsigned int            _index,
                           NSRange                 *_longestEffectiveRange_,
                           NSRange                 *_rangeLimit) {
  char    attrEntry          = 0;
  int     i = 0, j = 0, from = 0, to = 0;
  NSRange *range             = NULL;
  BOOL    status             = YES;
  int     strLen             = 0;
  id      searchedAttribute  = nil;

  strLen = [_string->string length];
  if (!(_index < strLen))
    raiseIndexOutOfBoundException(_index, strLen - 1);
  attrEntry = _string->attributes[_index];
  if (_longestEffectiveRange_) {
    if (_rangeLimit) {
      if (!isInRange(_rangeLimit, _index))
        raiseIndexOutOfRangeBoundException(_index, *_rangeLimit);
      
      from = _rangeLimit->location;
      to   = _rangeLimit->location + _rangeLimit->length;
    }
    else {
      from = 0;
      to   = strLen;
    }
    i         = _index;
    range     = _longestEffectiveRange_;
    status    = YES;
    searchedAttribute = [_string->attributeDicts[(int)_string->attributes[_index]]
                                objectForKey:_attributeName];
    if (!searchedAttribute) {
      NSRange *range = NULL;
      range = NGMalloc(sizeof(NSRange));
      range->location = _index;
      range->length   = 0;
      _longestEffectiveRange_ = range;
      return nil;
    }
    do {
      if (i == from) {
        status = NO;
        range->location = i;
      }
      else
        if (_string->attributes[--i] != attrEntry) {
          if ([searchedAttribute isEqual:[_string->attributeDicts[(int)_string->
                                                                 attributes[i]]
                                                 objectForKey:_attributeName]]) {
            attrEntry = _string->attributes[i];
          }
          else {
            status          = NO;
            range->location = ++i;
          }
        }
    } while (status);
    status = YES;
    j      = _index;
    attrEntry = _string->attributes[_index];
    do {
      if (j == to) {
        status = NO;
        range->length = j - i;
      }
      else
        if (_string->attributes[++j] != attrEntry) {
          if ([searchedAttribute isEqual:[_string->attributeDicts[(int)_string->
                                                                 attributes[j]]
                                                 objectForKey:_attributeName]]) {
            attrEntry = _string->attributes[j];
          }
          else {
            status        = NO;
            range->length = j - i;
          }
        }
    } while (status);
  }
  return searchedAttribute;  
}

// Comparing strings

BOOL NSCompareShortAttributedStrings(NSShortAttributedString *_string,
                                     NSShortAttributedString *_otherString) {
  int  lastEntry            = 0;
  int  countFoundedEntries  = 0;
  int  i = 0, j = 0, countI = 0;
  char *foundedEntries      = NULL;
  char *tmpTable            = NULL;
  BOOL isEqual              = NO;
  
  if (!([_string->string isEqualToString:_otherString->string]))
    return NO;

  __removeUnusedCombinations(_string);
  __removeUnusedCombinations(_otherString);  
  if (_string->lastAttributeEntry != _otherString->lastAttributeEntry)
    return NO;
  
  lastEntry      = _string->lastAttributeEntry;
  foundedEntries = NGMalloc(sizeof(char) * lastEntry + 1);
  tmpTable       = NGMalloc(sizeof(char) * lastEntry + 1);
  for (i = 0; i <= lastEntry; i++)
    foundedEntries[i] = 0;
  
  for (i = 0; i <= lastEntry; i++)
    for (j = 0; j <=lastEntry; j++) 
      if (foundedEntries[j] == 0) 
        if ([_string->attributeDicts[i] isEqualToDictionary:
                    _otherString->attributeDicts[j]]) {
          foundedEntries[j] = 1;
          tmpTable[i]       = j;
          countFoundedEntries++;
          break;
        }
  if (countFoundedEntries != lastEntry + 1)
    return NO;
  
  isEqual = YES;
  countI  = [_string->string length];
  for (i = 0; i < countI; i++) {
    if (tmpTable[(int)_string->attributes[i]] != _otherString->attributes[i]) {
      isEqual = NO;
      break;
    }
  }
  return isEqual;
}

// Substring
NSShortAttributedString *NSShortAttributedSubstring(
                                                    NSShortAttributedString *_string,
                                                    NSRange                 _range) {
  NSShortAttributedString *newAttrString = NULL;
  NSString                *newString     = nil;
  int  i              = 0;
  int  rangeLength    = 0;
  int  countAttr      = 0;
  int  countNew       = 0;  
  id   *newAttrDict   = NULL;
  char *newAttributes = NULL;
  char *tmpAttr       = NULL;
  BOOL status         = NO;

  rangeLength   = _range.length + _range.location;
  if ([_string->string length] <= rangeLength - 1)
    raiseRangeOutOfBoundException([_string->string length], _range);
  
  newAttrDict   = NGMalloc(sizeof(id)   * 256);
  newAttributes = NGMalloc(sizeof(char) *  _range.length);
  tmpAttr       = NGMalloc(sizeof(char) * _range.length);  
  newString     = [_string->string substringWithRange:_range];
  RETAIN(newString);
  countAttr = 0;
  countNew  = 0;
  for (i = _range.location; i < rangeLength; i++) {
    int j;

    NSCAssert(countAttr < 256, @"count attr ..");
    status = NO;    
    for (j = 0; j < countAttr; j++) {
      NSCAssert(j < countAttr, @"j 1 ...");
      if (tmpAttr[j] == _string->attributes[i]) {
        newAttributes[countNew++] = j;        
        status = YES;
        break;
      }
      NSCAssert(j < countAttr, @"j 1 ...");
    }
    NSCAssert(j <= countAttr, @"j 2 ...");
    
    if (!status) {
      tmpAttr[countAttr] = _string->attributes[i];

#if NSAttributedString_COPYMODE
      newAttrDict[countAttr] = [_string->attributeDicts[(int)(_string->
                                                              attributes[i])]
                                       copy];
#else
      newAttrDict[countAttr]    = _string->attributeDicts[(int)_string->
                                                         attributes[i]];
      RETAIN(newAttrDict[countAttr]);             
#endif
      
      printf("i : %d\n", i);
      printf("countAttr :%d\n", countAttr);
      printf("_string->attributes[i] : %d\n",_string->attributes[i]);      
      printf("%s\n",[[NSString stringWithFormat:@"%@",newAttrDict[countAttr]] cString]);

      newAttributes[countNew++] = countAttr++;
    }
  }
  NGFree(tmpAttr);
  tmpAttr = NULL;
  
  newAttrString = NGMalloc(sizeof(NSShortAttributedString));
  newAttrString->attributes         = newAttributes;
  newAttrString->attributeDicts     = newAttrDict;
  newAttrString->string             = newString;
  newAttrString->lastAttributeEntry = countAttr - 1;
  newAttrString->zone               = _string->zone;
  return newAttrString;
}


// Changing
void NSShortReplaceWithString(NSShortAttributedString *_string,
                              NSRange                 _range,
                              NSString                *_nsString) {
  
  NSMutableString *mutString = nil;
  NSString        *string    = nil;
  int  i = 0, j = 0;
  int  attrRange              = 0;
  int  oldAttrLength = 0, newAttribute = 0;
  int  newLength = 0, backLength = 0, nsStringLength = 0;
  char *newAttributes        = NULL;  
  char *oldAttributes        = NULL;

  if (_string == NULL) {
    [[[NSAttributedStringException alloc] initWithReason:
                                            @"try to insert nil-object"]
                                   raise];
  }

  if ((_range.location +_range.length) && ([_string->string length] <=
                                           (_range.location + _range.length - 1)))
    raiseRangeOutOfBoundException([_string->string length], _range);

  
  if (_range.location == 0)
    newAttribute = _string->attributes[0];
  else
    if (_range.length == 0) 
      newAttribute = _string->attributes[_range.location];
    else
      newAttribute = _string->attributes[_range.location - 1];
  

  attrRange     = MAX([_nsString length],_range.length);
  string        = _string->string;
  mutString     = [string mutableCopy];
  oldAttrLength = [string length];
  oldAttributes = _string->attributes;  
  RELEASE(string);
  [mutString replaceCharactersInRange:_range withString:_nsString];
  string = [mutString copy];
  RELEASE(mutString);

  newLength      = [string length];
  nsStringLength = [_nsString length];
  newAttributes  = NGMalloc(sizeof(char) * [string length]);
  for (i = 0; i < _range.location; i++)
    newAttributes[i] = oldAttributes[i];

  j = oldAttrLength - 1;
  backLength = MIN(_range.length, [_nsString length]);
  backLength = _range.location + backLength - 1;
  for (i = newLength - 1; i > backLength; i--)
    newAttributes[i] = oldAttributes[j--];

  j = _range.location;
  for (i = 0; i < nsStringLength; i++)
    newAttributes[j++] = newAttribute;

  NGFree(_string->attributes);
  _string->string     = string;
  _string->attributes = newAttributes;
  oldAttributes       = NULL;
  newAttributes       = NULL;
  string              = nil;
  mutString           = nil;
}

void NSShortReplaceWithAttributedString(NSShortAttributedString *_string,
                                        NSRange                 _range,
                                        NSShortAttributedString *_otherString) {
  NSMutableString *mutString = nil;
  NSString        *string    = nil;
  int  i = 0, j = 0;
  int  attrRange              = 0;
  int  oldAttrLength = 0, newLength = 0, backLength = 0;
  int  otherStringLength = 0;
  char *oldAttributes    = NULL;
  char *otherAttributes  = NULL;
  char *newAttributes    = NULL;

  if ([_string->string length] <= (signed)(_range.location + _range.length - 1))
    raiseRangeOutOfBoundException([_string->string length], _range);

  if (!_string) {
    THROW([[NSAttributedStringException alloc] initWithReason:
                                                 @"try to insert nil-Object"]);
  }

  attrRange       = MAX([_otherString->string length],_range.length);
  string          = _string->string;
  mutString       = [string mutableCopy];
  oldAttrLength   = [string length];
  oldAttributes   = _string->attributes;
  otherAttributes = _otherString->attributes;

  RELEASE(string);
  [mutString replaceCharactersInRange:_range withString:_otherString->string];
  string = [mutString copy];
  RELEASE(mutString);

  newLength         = [string length];
  otherStringLength = [_otherString->string length];
  newAttributes     = NGMalloc(sizeof(char) * [string length]);

  for (i = 0; i < _range.location; i++)
    newAttributes[i] = oldAttributes[i];

  j = oldAttrLength - 1;
  backLength = newLength - oldAttrLength + _range.location - 1;
  for (i = newLength - 1; i >= backLength; i--)
    newAttributes[i] = oldAttributes[j--];

  j = _range.location;
  for (i = 0; i < otherStringLength; i++) {
    NSDictionary *mDict = nil;
    int pos;

    mDict = _otherString->attributeDicts[(int)otherAttributes[i]];
    if (__compareDictArray(_string, mDict, &pos))
      newAttributes[j++] = pos;
    else {
      if (_string->lastAttributeEntry == 255) {
        __removeUnusedCombinations(_string);
        if (_string->lastAttributeEntry == 255)
          raiseToManyAttributeCombinationsException();

      }
      _string->attributeDicts[++_string->lastAttributeEntry] = RETAIN(mDict);
      newAttributes[j++] = _string->lastAttributeEntry;
    }
  }

  NGFree(_string->attributes);
  _string->string     = string;
  _string->attributes = newAttributes;
  oldAttributes       = NULL;
  newAttributes       = NULL;
  string              = NULL;
  mutString           = NULL;
}

void NSShortSetAttributes(NSShortAttributedString *_string,
                          NSRange                 _range,
                          NSDictionary            *_attributes) {
  int i = 0, pos = 0, last = 0;
  int rangeLength = 0;
  
  last        = _string->lastAttributeEntry;
  rangeLength = _range.location + _range.length;
  if ([_string->string length] <= rangeLength - 1)
    raiseRangeOutOfBoundException([_string->string length], _range);
  
  if (__compareDictArray(_string, _attributes, &pos)) 
    for (i = _range.location; i < rangeLength; i++)
      _string->attributes[i] = pos;
  else {
    int last = 0;
    
    RETAIN(_attributes);
    if (_string->lastAttributeEntry == 255)
      __removeUnusedCombinations(_string);

    last = ++_string->lastAttributeEntry;
    if (last > 255) 
      raiseToManyAttributeCombinationsException();

    _string->attributeDicts[last] = _attributes;
    for (i = _range.location; i < rangeLength; i++)
      _string->attributes[i] = last;
  }
}
  

  
void NSShortAddAttributes(NSShortAttributedString *_string,
                          NSRange                 _range,
                          NSDictionary            *_attributes) {

  NSRange range;
  int     i = 0, j = 0, pos = 0, countI = 0;

  countI = _range.location + _range.length;
  if ([_string->string length] <= countI - 1)
    raiseRangeOutOfBoundException([_string->string length], _range);
  
  for (i = _range.location; i < countI;) {
    int                  rangeLength = 0;
    NSMutableDictionary  *mDict      = nil;
    NSDictionary         *dict       = nil;

    mDict = [NSShortAllAttributesAtIndex(_string, i, &range, &_range) mutableCopy];

    if (range.location >= _range.location)
      rangeLength = range.location + range.length;
    else
      rangeLength = _range.location + range.length;

    if (rangeLength >= countI)
      rangeLength = countI;

    [mDict addEntriesFromDictionary:_attributes];
    dict = [mDict copy];
    RELEASE(mDict);
    if (__compareDictArray(_string, dict, &pos)) {
      for (j = range.location; j < rangeLength; j++)
        _string->attributes[j] = pos;
      RETAIN(dict);
    }
    else {
      int last = 0;
      if (_string->lastAttributeEntry == 255)
        __removeUnusedCombinations(_string);

      last = ++_string->lastAttributeEntry;
      if (last > 255) 
        raiseToManyAttributeCombinationsException();

      _string->attributeDicts[last] = dict;
      for (j = range.location; j < rangeLength; j++)
        _string->attributes[j] = last;
    }
    i = range.location + range.length;
  }
}

void NSShortRemoveAttribute(NSShortAttributedString *_string,
                            NSRange                 _range,
                            NSString                *_name) {
  int  i = 0, j = 0, cntI = 0;
  int  dictNumber    = 0;
  int  countDicts    = 0;
  id   *changedDicts = NULL;
  char *dictNumbers  = NULL;
  BOOL status        = NO;

  cntI         = _range.location + _range.length;
  dictNumbers  = NGMalloc(sizeof(char) * 256);
  changedDicts = NGMalloc(sizeof(id)   * 256);
  countDicts   = 0;
  
  for (i = _range.location; i < cntI; i++) {
    dictNumber  = _string->attributes[i];
    status = NO;    
    for (j = 0; j < countDicts; j++)
      if (dictNumbers[j] == dictNumber) {
        status = YES;
        break;
      }
    if (!status) {
      dictNumbers[countDicts]  = dictNumber;
      changedDicts[countDicts] = [_string->attributeDicts[dictNumber] mutableCopy];
      [changedDicts[countDicts++] removeObjectForKey:_name];
    }
  }
  
  for (i = 0; i < countDicts; i++) {
    int pos = 0;
    
    if (!(__compareDictArray(_string, changedDicts[i], &pos))) {
      if (_string->lastAttributeEntry == 255) {
        __removeUnusedCombinations(_string);
        if (_string->lastAttributeEntry == 255)
          raiseToManyAttributeCombinationsException();
      }
      pos  = ++_string->lastAttributeEntry;
      _string->attributeDicts[_string->lastAttributeEntry] = [changedDicts[i] copy];
      RETAIN(_string->attributeDicts[_string->lastAttributeEntry]);
      RELEASE(changedDicts[i]);
    }
    for (j = _range.location; j < cntI; j++)
      if (_string->attributes[j] == dictNumbers[i])
        _string->attributes[j] = pos;
  }
  __removeUnusedCombinations(_string);
}

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
