/*
   NSString+Formatting.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+Formatting.m,v 1.8 1999/03/16 13:36:36 helge Exp $

#import "common.h"
#import "NSString+Formatting.h"

// format
//   %|[justification]|[fieldwidth]|[precision]|formatChar
//   %[-]{digit}*[.{digit}*]conv

static const char *formatChars =
  "diouxX" // integers
  "feEgG"  // double
  "c"      // int arg -> uchar
  "s"      // string
  "@"      // object
  "$"      // string object
;
static const char *formatAttrs = "#0- +,";
static const char *formatConv  = "hlLqZ";

typedef enum {
  DEFAULT_TYPE = 0,
  SHORT_TYPE,
  LONG_TYPE,
  VERY_LONG_TYPE
} NGFormatType;

// implementation

static inline NSString *wideString(const char *str, int width, char fillChar,
                                   BOOL isRightAligned) {
  if (width > 0) {
    int len = strlen(str);

    if (len < width) {
      unsigned char *tmp = objc_atomic_malloc(width + 1);
      register int cnt;

      if (isRightAligned) { // right aligned
        for (cnt = 0; cnt < (width - len); cnt++)
          tmp[cnt] = fillChar;
        strncpy(tmp + (width - len), str, len);
        tmp[width] = '\0';
      }
      else { // left aligned
        for (cnt = len; cnt < width; cnt++)
          tmp[cnt] = fillChar;
        strncpy(tmp, str, len);
        tmp[cnt] = '\0';
      }

      return [NSString stringWithCStringNoCopy:tmp
                       freeWhenDone:YES];
    }
  }
  return [NSString stringWithCString:str];
}

static inline NSString *xsSimpleFormatObject(char *fmt, id _value, NSString *_ds) {
  /*
    formats an object value in an simple format
    (this is only the '%{mods}{type}' part of a real
    format string !)
    The method to get the value of the object is
    determined by the {type} of the Format,
    eg: i becomes intValue, @ becomes description ..

    the decimal point in float-strings can be replaced by the _ds
    String. To do this, the result of the Format is scanned for
    '.' and _ds is expanded for this.
  */
  int  fmtLen  = strlen(fmt);
  char fmtChar = fmt[fmtLen - 1];

  NGFormatType typeMode = DEFAULT_TYPE;
  int  width      = -1;
  int  prec       = -1;
  BOOL alignRight = YES;
  char *pos = fmt + 1;
  char *tmp;

  while (index(formatAttrs, *pos)) {
    if (*pos == '-')
      alignRight = NO;
    pos++;
  }
  
  // width
  tmp = pos;
  while (isdigit((int)*pos)) pos++;
  if (tmp != pos) {
    char old = *pos;
    *pos = '\0';
    width = atoi(tmp);
    *pos = old;
  }

  // prec
  if (*pos == '.') {
    pos++;
    tmp = pos;
    while (isdigit((int)*pos)) pos++;
    if (tmp != pos) {
      char old = *pos;
      *pos = '\0';
      prec = atoi(tmp);
      *pos = old;
    }
  }

  // conversion
  if (index(formatConv, *pos)) {
    switch (*pos) {
      case 'h': typeMode = SHORT_TYPE;     break;
      case 'l': typeMode = LONG_TYPE;      break;
      case 'L':
      case 'q': typeMode = VERY_LONG_TYPE; break;
    }
    pos++;
  }

  if (index("diouxX", fmtChar)) {
    char buf[128];
    int  len = -1;

    switch (typeMode) {
      case SHORT_TYPE: {
        unsigned short int i = [(NSNumber *)_value unsignedShortValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      case LONG_TYPE: {
        unsigned long int i = [(NSNumber *)_value unsignedLongValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      case VERY_LONG_TYPE: {
        long long i = [(NSNumber *)_value unsignedLongLongValue];
        len = sprintf(buf, fmt, i);
        break;
      }
      default: {
        unsigned int i = [(NSNumber *)_value unsignedIntValue];
        len = sprintf(buf, fmt, i);
        break;
      }
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (fmtChar == 'c') {
    char buf[64];
    int  i = [(NSNumber *)_value unsignedCharValue];
    int  len;

    len = sprintf(buf, fmt, (unsigned char)i);

    return (len > 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (index("feEgG", fmtChar)) {
    char buf[256];
    int  len;

    if (typeMode == VERY_LONG_TYPE) {
      long double i = [(NSNumber *)_value doubleValue];
      len = sprintf(buf, fmt, i);
    }
    else {
      double i = [(NSNumber *)_value doubleValue];
      len = sprintf(buf, fmt, i);
    }

    if (len >= 0) {
      if (_ds == nil)
        return [NSString stringWithCString:buf length:len];
      else {
        NSMutableString *result =
          [NSMutableString stringWithCapacity:len + [_ds cStringLength]];
        unsigned int cnt;
        char         *ptr = buf;
        unsigned int bufCount = 0;

        for (cnt = 0; cnt < len; cnt++) {
          if (buf[cnt] == '.') {
            if (bufCount > 0)
              [result appendString:[NSString stringWithCString:ptr length:bufCount]];

            [result appendString:_ds];
            
            ptr = &(buf[cnt + 1]);
            bufCount = 0;
          }
          else {
            bufCount++;
          }
        }
        
        if (bufCount > 0)
          [result appendString:[NSString stringWithCString:ptr length:bufCount]];
        return result;
      }
    }
    else return nil;
  }
  else {
    switch (fmtChar) {
      case 's': {
        char *str = (char *)[(NSString *)_value cString];

        if (str == NULL) str = "<null>";
        return wideString(str, width, ' ', alignRight);
      }

      case '@': {
        id       obj   = _value;
        NSString *dstr = obj ? [obj description] : @"<nil>";
        char     *cstr = (char *)[dstr cString];

        if (cstr == NULL) cstr = "<null>";
        return wideString(cstr, width, ' ', alignRight);
      }

      case '$': {
        id       obj   = _value;
        NSString *dstr = obj ? [obj stringValue] : @"<nil>";
        char     *cstr = (char *)[dstr cString];

        if (cstr == NULL) cstr = "<null>";

        return wideString(cstr, width, ' ', alignRight);
      }
        
      default:
        fprintf(stderr, "unknown printf format used: '%s'\n", fmt);
        break;
    }
  }
  return nil;
}

static inline NSString *handleFormat(char *fmt, int fmtLen, va_list *ap) {
  char fmtChar    = fmt[fmtLen - 1];
  char typeMode   = DEFAULT_TYPE;
  int  width      = -1;
  int  prec       = -1;
  BOOL alignRight = YES;
  char *pos = fmt + 1;
  char *tmp;

  while (index(formatAttrs, *pos)) {
    if (*pos == '-')
      alignRight = NO;
    pos++;
  }
  
  // width
  tmp = pos;
  while (isdigit((int)*pos)) pos++;
  if (tmp != pos) {
    char old = *pos;
    *pos = '\0';
    width = atoi(tmp);
    *pos = old;
  }

  // prec
  if (*pos == '-') {
    pos++;
    tmp = pos;
    while (isdigit((int)*pos)) pos++;
    if (tmp != pos) {
      char old = *pos;
      *pos = '\0';
      prec = atoi(tmp);
      *pos = old;
    }
  }

  // conversion
  if (index(formatConv, *pos)) {
    switch (*pos) {
      case 'h': typeMode = SHORT_TYPE;     break;
      case 'l': typeMode = LONG_TYPE;      break;
      case 'L':
      case 'q': typeMode = VERY_LONG_TYPE; break;
    }
    pos++;
  }

  //  printf("  width=%i prec=%i\n", width, prec); fflush(stdout);
    
  if (index("diouxX", fmtChar)) {
    char buf[128];
    int  len = -1;

    switch (typeMode) {
      case SHORT_TYPE: {
        unsigned short int i = va_arg(*ap, unsigned short int);
        len = sprintf(buf, fmt, i);
        break;
      }
      case LONG_TYPE: {
        unsigned long int i = va_arg(*ap, unsigned long int);
        len = sprintf(buf, fmt, i);
        break;
      }
      case VERY_LONG_TYPE: {
        long long i = va_arg(*ap, long long);
        len = sprintf(buf, fmt, i);
        break;
      }
      default: {
        unsigned int i = va_arg(*ap, unsigned int);
        len = sprintf(buf, fmt, i);
        break;
      }
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else if (fmtChar == 'c') {
    char buf[64];
    int  i = va_arg(*ap, int);
    //    int  i = va_arg(*ap, unsigned char);
    int  len;

    if (i != 0) {
      len = sprintf(buf, fmt, (unsigned char)i);

      // xsprintf("got format %s char %i made %s len %i\n",
      //         fmt, i, buf, len);
    
      return (len > 0)
        ? [NSString stringWithCString:buf length:len] : nil;
    }
    else
      return @"<'\\0'-char>";
  }
  else if (index("feEgG", fmtChar)) {
    char buf[256];
    int  len;

    if (typeMode == VERY_LONG_TYPE) {
      long double i = va_arg(*ap, long double);
      len = sprintf(buf, fmt, i);
    }
    else {
      double i = va_arg(*ap, double);
      len = sprintf(buf, fmt, i);
    }

    return (len >= 0) ? [NSString stringWithCString:buf length:len] : nil;
  }
  else {
    id result = nil;
    
    switch (fmtChar) {
      case 's':
      case '@':
      case '$': {
        const char *cstr = NULL;

        if (fmtChar == 's') {
          cstr = va_arg(*ap, char *);
        }
        else {
          id obj = va_arg(*ap, id);
          if (obj == nil)
            cstr = "<nil>";
          else {
            if (fmtChar == '@')
              cstr = [[obj description] cString];
            else
              cstr = [[obj stringValue] cString];
          }
        }

        if (cstr == NULL) cstr = "<null>";

        result = wideString(cstr, width, ' ', alignRight);
        break;
      }

      default:
        fprintf(stderr, "unknown printf format used: '%s'\n", fmt);
        break;
    }
    return result;
  }
  return nil;
}

static inline NSString *_stringWithCFormat(const char *_format, va_list _ap) {
  const char *firstPercent = index(_format, '%');

  //fprintf(stderr, "OWN: format='%s'\n", _format);
  //fflush(stderr);

  // first check whether there are any '%' in the format ..
  if (firstPercent == NULL) {
    // no formatting contained in _format
    return [NSString stringWithCString:_format];
  }
  else {
    NSMutableString *result = [NSMutableString stringWithCapacity:256];

    if ((firstPercent - _format) > 0) {
      [result appendString:[NSString stringWithCString:_format
                                     length:(firstPercent - _format)]];
      _format = firstPercent;
    }

    while (*_format != '\0') { // until end of format string
      if (*_format == '%') { // found formatting character
        _format++; // skip '%'

        if (*_format == '%') { // was a quoted '%'
          [result appendString:@"%"];
          _format++;
        }
        else { // check format
          char extFmt[16];
          char *pos  = extFmt;

          extFmt[0] = '%';
          pos++;
        
          while ((*_format != '\0') &&
                 (index(formatChars, *_format) == NULL)) {
            *pos = *_format;
            _format++;
            pos++;
          }
          *pos = *_format;
          _format++;
          pos++;
          *pos = '\0';

          // printf("handling ext format '%s'\n", extFmt); fflush(stdout);
        
          [result appendString:handleFormat(extFmt, strlen(extFmt), &_ap)];
        }
      }
      else { // normal char
        const char *start = _format; // remember start
        _format++; // skip found char

        // further increase format until '\0' or '%'
        while ((*_format != '\0') && (*_format != '%'))
          _format++;
        
        [result appendString:[NSString stringWithCString:start
                                       length:(_format - start)]];
      }
    }
    return result;
  }
}

@implementation NSString(XSFormatting)

+ (id)stringWithCFormat:(const char *)_format arguments:(va_list)_ap {
  return [self stringWithString:_stringWithCFormat(_format, _ap)];
}

+ (id)stringWithFormat:(NSString *)_format arguments:(va_list)_ap {
  return [self stringWithCFormat:[_format cString] arguments:_ap];
}

+ (id)stringWithCFormat:(const char *)_format, ... {
  id      result = nil;
  va_list ap;
  
  va_start(ap, _format);
  result = [self stringWithCFormat:_format arguments:ap];
  va_end(ap);
  return result;
}

+ (id)stringWithFormat:(NSString *)_format, ... {
  id      result = nil;
  va_list ap;
  
  va_start(ap, _format);
  result = [self stringWithCFormat:[_format cString] arguments:ap];
  va_end(ap);
  return result;
}

- (id)initWithFormat:(NSString *)_format arguments:(va_list)_ap {
  return [self initWithString:_stringWithCFormat([_format cString], _ap)];
}

@end

@implementation NSMutableString(XSFormatting)

- (void)appendFormat:(NSString *)_format arguments:(va_list)_ap {
  [self appendString:_stringWithCFormat([_format cString], _ap)];
}
- (void)appendFormat:(NSString *)_format, ... {
  va_list ap;
  va_start(ap, _format);
  [self appendString:_stringWithCFormat([_format cString], ap)];
  va_end(ap);
}

@end

// var args wrappers

int xs_sprintf(char *str, const char *format, ...) {
  va_list ap;
  int result;
  va_start(ap, format);
  result = xs_vsprintf(str, format, ap);
  va_end(ap);
  return result;
}
int xs_snprintf(char *str, size_t size, const char *format, ...) {
  va_list ap;
  int result;
  va_start(ap, format);
  result = xs_vsnprintf(str, size, format, ap);
  va_end(ap);
  return result;
}

// linking

void __link_NSString_Formatting(void) {
  __link_NSString_Formatting();
}
