/* 
   NSString+misc.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+misc.m,v 1.4 1998/12/30 08:56:07 helge Exp $

#import "common.h"
#import "NGMemoryAllocation.h"
#import "NSString+misc.h"

@implementation NSString(misc)

- (NSString *)stringByApplyingCEscaping {
  const char   *cstr = [self cString];
  char         *buffer = NGMallocAtomic([self cStringLength] * 2 + 1);
  register int pos = 0;
    
  while (*cstr) {
    switch (*cstr) {
      case '\n':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 'n';
        break;
      case '\r':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 'r';
        break;
      case '\t':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 't';
        break;

      default:
        buffer[pos] = *cstr;
        break;
    }
    cstr++;
    pos++;
  }
  buffer[pos] = '\0';

  return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
}

@end

#ifdef LIB_FOUNDATION_LIBRARY
#if STRING_DEBUG

static inline void Free(void*) __attribute__((unused));

#if LIB_FOUNDATION_BOEHM_GC
static inline void Free(void* p) {}
#else
static inline void Free(void* p) {
  return [[NSZone zoneFromPointer:p] freePointer:p];
}
#endif

@implementation NSInline8BitString(StringDebugging)

- initWithCString:(const char*)byteString length:(unsigned int)length
{
    NSAssert(strlen(byteString) >= length,
             @"length of byte string must be greater or equal then given length");
    
    if (cLength != -1) {
	[[[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]
                                       raise];
    }
    cLength = length;
    memcpy(cString, byteString, length);
    cString[length] = 0;
    return self;
}

@end

@implementation NSNonOwned8BitString(StringDebugging)

- initWithCString:(char*)byteString length:(unsigned int)length copy:(BOOL)flag
{
    NSAssert(strlen(byteString) >= length,
             @"length of byte string must be greater or equal then given length");
    
    if (cLength || cString)
	[[[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]
                                       raise];
    if (flag) {
	[[[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s with flag set to YES to %@ to instance", 
	    sel_get_name(_cmd), NSStringFromClass([self class])]
                                        raise];
    }
    
    cLength = length;
    cString = byteString;
    return self;
}

@end

@implementation NSOwned8BitString(StringDebugging)

- initWithCString:(char*)byteString length:(unsigned int)length copy:(BOOL)flag
{
    NSAssert(strlen(byteString) >= length,
             @"length of byte string must be greater or equal then given length");
    
    if (cLength || cString)
	[[[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]
                                       raise];

    cLength = length;
    if (flag) {
	cString = NSZoneMallocAtomic([self zone], sizeof(char)*(length+1));
	memcpy(cString, byteString, length);
	cString[cLength] = 0;
    }
    else
	cString = byteString;
    return self;
}

@end

@implementation NSOwnedOpen8BitString(StringDebugging)

- initWithCString:(char*)byteString length:(unsigned int)length copy:(BOOL)flag
{
    NSAssert(strlen(byteString) >= length,
             @"length of byte string must be greater or equal then given length");
    
    if (cLength || cString)
	[[[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]
                                       raise];

    cLength = length;
    if (flag) {
	cString = NSZoneMallocAtomic([self zone], sizeof(char)*(length));
	memcpy(cString, byteString, length);
    }
    else
	cString = byteString;
    return self;
}

@end

@implementation NSMutableSimple8BitString(StringDebugging)

- initWithCString:(char*)byteString length:(unsigned int)length copy:(BOOL)flag
{
    NSAssert(strlen(byteString) >= length,
             @"length of byte string must be greater or equal then given length");
    
    if (flag) {
	if (cCapacity < length) {
	    Free(cString);
	    cString = NSZoneMallocAtomic([self zone], sizeof(char)*length);
	    cCapacity = length;
	}
	cLength = length;
	memcpy(cString, byteString, length);
    }
    else {
	Free(cString);
	cString = byteString;
	cLength = cCapacity = length;
    }
    return self;
}

@end

#endif // STRING_DEBUG
#endif // lf

// linking

void __link_NSString_misc(void) {
  __link_NSString_misc();
}
