/* 
   NGActiveSocket.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGActiveSocket.m,v 1.48 1999/03/26 12:42:43 helge Exp $

#include "config.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#  include <sys/select.h>
#endif
#ifdef HAVE_SYS_FILIO_H
#  include <sys/filio.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
#endif
#ifdef HAVE_TIME_H
#  include <time.h>
#endif
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#endif

#if defined(WIN32)
#  include <winsock.h>
#  define ioctl ioctlsocket
#endif

#include "common.h"
#include <NGStreams/NGDescriptorFunctions.h>
#include "NGActiveSocket.h"
#include "NGSocketExceptions.h"
#include "NGSocket+private.h"

#if !defined(POLLRDNORM)
#  define POLLRDNORM POLLIN
#endif

@interface NGActiveSocket(PrivateMethods)

- (id)_initWithDescriptor:(int)_fd
  localAddress:(id<NGSocketAddress>)_local
  remoteAddress:(id<NGSocketAddress>)_remote;

@end

@implementation NGActiveSocket

#if !defined(WIN32)

+ (BOOL)socketPair:(id<NGSocket>[2])_pair inDomain:(id<NGSocketDomain>)_domain {
  int fds[2];

  _pair[0] = nil;
  _pair[1] = nil;

  if (socketpair([_domain socketDomain], SOCK_STREAM, [_domain protocol],
                 fds) == 0) {
    NGActiveSocket *s1 = nil;
    NGActiveSocket *s2 = nil;
    
    s1 = [[self alloc] _initWithDomain:_domain descriptor:fds[0]];
    s2 = [[self alloc] _initWithDomain:_domain descriptor:fds[1]];
    AUTORELEASE(s1);
    AUTORELEASE(s2);

    if ((s1 != nil) && (s2 != nil)) {
      s1->mode           = NGStreamMode_readWrite;
      s1->receiveTimeout = 0.0;
      s1->sendTimeout    = 0.0;
      s2->mode           = NGStreamMode_readWrite;
      s2->receiveTimeout = 0.0;
      s2->sendTimeout    = 0.0;

      _pair[0] = s1;
      _pair[1] = s2;

      return YES;
    }
    else
      return NO;
  }
  else {
    int      e       = errno;
    NSString *reason = nil;

    switch (e) {
      case EACCES:
        reason = @"Not allowed to create socket of this type";
        break;
      case ENOMEM:
        reason = @"Could not create socket: Insufficient user memory available";
        break;
      case ENOSR:
        reason = @"Could not create socket: Insufficient resources available";
        break;
      case EPROTONOSUPPORT:
        reason = @"The protocol is not supported by the address family or "
                 @"implementation";
        break;
      case EPROTOTYPE:
        reason = @"The socket type is not supported by the protocol";
        break;
      case EMFILE:
        reason = @"Could not create socket: descriptor table is full";
        break;
      case EOPNOTSUPP:
        reason = @"The specified protocol does not permit creation of socket "
                 @"pairs";
        break;

      default:
        reason = [NSString stringWithFormat:@"Could not create socketpair: %s",
                             strerror(e)];
        break;
    }
    THROW([[NGCouldNotCreateSocketException alloc]
              initWithReason:reason domain:_domain]);
    return NO;
  }
}

#endif

+ (id)socketConnectedToAddress:(id<NGSocketAddress>)_address {
  volatile id sock = [[self alloc] initWithDomain:[_address domain]];

  if (sock != nil) {
    AUTORELEASE(sock);
    [sock connectToAddress:_address];
  }
  return sock;
}

- (id)initWithDomain:(id<NGSocketDomain>)_domain {
  // designated initializer
  if ((self = [super initWithDomain:_domain])) {
    self->mode           = NGStreamMode_readWrite;
    self->receiveTimeout = 0.0;
    self->sendTimeout    = 0.0;
  }
  return self;
}

- (id)_initWithDescriptor:(int)_fd
  localAddress:(id<NGSocketAddress>)_local
  remoteAddress:(id<NGSocketAddress>)_remote {

  if ((self = [self _initWithDomain:[_local domain] descriptor:_fd])) {
    ASSIGN(self->localAddress,  _local);
    ASSIGN(self->remoteAddress, _remote);
    self->mode = NGStreamMode_readWrite;

#if !defined(WIN32)
    NGAddDescriptorFlag(self->fd, O_NONBLOCK);
#endif
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->remoteAddress); self->remoteAddress = nil;
  [super dealloc];
}
#endif

// operations

- (void)primaryConnectToAddress:(id<NGSocketAddress>)_address {
  // throws
  //   NGCouldNotConnectException  if the the connect() call fails
  
  if (connect(fd,
              (struct sockaddr *)[_address internalAddressRepresentation],
              [_address addressRepresentationSize]) != 0) {
    NSString *reason = nil;
    int errorCode = errno;

    switch (errorCode) {
      case EACCES:
        reason = @"search permission denied for element in path";
        break;
#if defined(WIN32)
      case WSAEADDRINUSE:
        reason = @"address already in use";
        break;
      case WSAEADDRNOTAVAIL:
        reason = @"address is not available on remote machine";
        break;
      case WSAEAFNOSUPPORT:
        reason = @"addresses in the specified family cannot be used with the socket";
        break;
      case WSAEALREADY:
        reason = @"a previous non-blocking attempt has not yet been completed";
        break;
      case WSAEBADF:
        reason = @"descriptor is invalid";
        break;
      case WSAECONNREFUSED:
        reason = @"connection refused";
        break;
      case WSAEINTR:
        reason = @"connect was interrupted";
        break;
      case WSAEINVAL:
        reason = @"the address length is invalid";
        break;
      case WSAEISCONN:
        reason = @"socket is already connected";
        break;
      case WSAENETUNREACH:
        reason = @"network is unreachable";
        break;
      case WSAETIMEDOUT:
        reason = @"timeout occured";
        break;
#else
      case EADDRINUSE:
        reason = @"address already in use";
        break;
      case EADDRNOTAVAIL:
        reason = @"address is not available on remote machine";
        break;
      case EAFNOSUPPORT:
        reason = @"addresses in the specified family cannot be used with the socket";
        break;
      case EALREADY:
        reason = @"a previous non-blocking attempt has not yet been completed";
        break;
      case EBADF:
        reason = @"descriptor is invalid";
        break;
      case ECONNREFUSED:
        reason = @"connection refused";
        break;
      case EINTR:
        reason = @"connect was interrupted";
        break;
      case EINVAL:
        reason = @"the address length is invalid";
        break;
      case EIO:
        reason = @"an IO error occured";
        break;
      case EISCONN:
        reason = @"socket is already connected";
        break;
      case ENETUNREACH:
        reason = @"network is unreachable";
        break;
      case ETIMEDOUT:
        reason = @"timeout occured";
        break;
#endif

      default:
        reason = [NSString stringWithCString:strerror(errorCode)];
        break;
    }

    reason = [NSString stringWithFormat:@"Could not connect to address %@: %@",
                         _address, reason];
    
    THROW([[NGCouldNotConnectException alloc]
              initWithReason:reason socket:self address:_address]);
  }

  // connect was successful

  ASSIGN(self->remoteAddress, _address);

#if !defined(WIN32)
  // mark socket as non-blocking
  NGAddDescriptorFlag(self->fd, O_NONBLOCK);
  NSAssert((NGGetDescriptorFlags(self->fd) & O_NONBLOCK),
           @"could not enable non-blocking mode ..");
#endif
}

- (void)connectToAddress:(id<NGSocketAddress>)_address {
  // throws
  //   NGSocketAlreadyConnectedException  if the socket is already connected
  //   NGInvalidSocketDomainException     if the remote domain != local domain
  //   NGCouldNotCreateSocketException    if the socket creation failed
  
  if ([self isConnected]) {
    THROW([[NGSocketAlreadyConnectedException alloc]
              initWithReason:@"Could not connected: socket is already connected"
              socket:self address:self->remoteAddress]);
  }

  // check whether the remote address is in the same domain like the bound one
  if (flags.isBound) {
    if (![[localAddress domain] isEqual:[_address domain]]) {
      THROW([[NGInvalidSocketDomainException alloc]
                initWithReason:@"local and remote socket domains are different"
                socket:self domain:[_address domain]]);
    }
  }

  // connect, remote-address is non-nil if this returns
  [self primaryConnectToAddress:_address];

  // if the socket wasn't bound before (normal case), bind it now
  if (!flags.isBound)
    [self kernelBoundAddress];
}

- (void)shutdown {
  if (self->fd != NGInvalidSocketDescriptor) {
    if (self->mode != NGStreamMode_undefined) {
      if (shutdown(self->fd, SHUT_RDWR) == 0)
        self->mode = NGStreamMode_undefined;
      //else
      //  NSLog(@"ERROR: shutdown of socket %@ failed: %s",
      // self, strerror(errno));
    }

#if defined(WIN32)
    if (closesocket(self->fd) == 0) {
#else
    if (close(self->fd) == 0) {
#endif
      //NSLog(@"closed socket ..");
      self->fd = NGInvalidSocketDescriptor;
    }
    else {
      NSLog(@"ERROR: close of socket %@ (fd=%i) alive=%s failed: %s",
            self, self->fd, [self isAlive] ? "YES" : "NO", strerror(errno));
    }
    
    RELEASE(self->remoteAddress); self->remoteAddress = nil;
  }
}

- (void)shutdownSendChannel {
  if (NGCanWriteInStreamMode(self->mode)) {
    shutdown(self->fd, SHUT_WR);
    
    if (self->mode == NGStreamMode_readWrite)
      self->mode = NGStreamMode_readOnly;
    else {
      self->mode = NGStreamMode_undefined;
#if defined(WIN32)
      closesocket(self->fd);
#else
      close(self->fd);
#endif
      self->fd = NGInvalidSocketDescriptor;
    }
  }
}
- (void)shutdownReceiveChannel {
  if (NGCanReadInStreamMode(self->mode)) {
    shutdown(self->fd, SHUT_RD);
    
    if (self->mode == NGStreamMode_readWrite)
      self->mode = NGStreamMode_writeOnly;
    else {
      self->mode = NGStreamMode_undefined;
#if defined(WIN32)
      closesocket(self->fd);
#else
      close(self->fd);
#endif
      self->fd = NGInvalidSocketDescriptor;
    }
  }
}

// ******************** accessors ******************

- (id<NGSocketAddress>)remoteAddress {
  return self->remoteAddress;
}

- (BOOL)isConnected {
  return (self->remoteAddress != nil);
}

- (int)socketType {
  return SOCK_STREAM;
}

- (void)setSendTimeout:(NSTimeInterval)_timeout {
  self->sendTimeout = _timeout;
}
- (NSTimeInterval)sendTimeout {
  return self->sendTimeout;
}

- (void)setReceiveTimeout:(NSTimeInterval)_timeout {
  self->receiveTimeout = _timeout;
}
- (NSTimeInterval)receiveTimeout {
  return self->receiveTimeout;
}

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  short events = 0;

  if ((![self isConnected]) || (fd == NGInvalidSocketDescriptor))
    return NO;

  if (NGCanReadInStreamMode(_mode))  events |= POLLRDNORM;
  if (NGCanWriteInStreamMode(_mode)) events |= POLLWRNORM;

  // timeout of 0 means return immediatly
  return (NGPollDescriptor([self fileDescriptor], events, 0) == 1 ? NO : YES);
}

- (int)waitForMode:(NGStreamMode)_mode timeout:(NSTimeInterval)_timeout {
  short events = 0;

  if (NGCanReadInStreamMode(_mode))  events |= POLLRDNORM;
  if (NGCanWriteInStreamMode(_mode)) events |= POLLWRNORM;

  // timeout of 0 means return immediatly
  return NGPollDescriptor([self fileDescriptor], events,
                          (int)(_timeout * 1000.0));
}

- (unsigned)numberOfAvailableBytesForReading {
  int len;

  // need to check whether socket is connected
  if (self->remoteAddress == nil) {
    THROW([[NGSocketNotConnectedException alloc]
              initWithReason:@"socket is not connected" socket:self]);
  }
  
  if (!NGCanReadInStreamMode(self->mode))
    THROW([[NGWriteOnlyStreamException alloc] initWithStream:self]);

#if !defined(WIN32)
  while (ioctl(self->fd, FIONREAD, &len) == -1) {
    if (errno == EINTR) continue;
    
    THROW([[NGSocketException alloc]
              initWithReason:@"could not get number of available bytes"
              socket:self]);
  }
#else
  // PeekNamedPipe()
  len = 0;
#endif
  return len;
}

- (BOOL)isAlive {
  if (self->fd == NGInvalidSocketDescriptor)
    return NO;

  // poll socket for input
  {
    struct timeval to;
    fd_set readMask;

    while (YES) {
      FD_ZERO(&readMask);
      FD_SET(self->fd, &readMask);
      to.tv_sec = to.tv_usec = 0;
      
      if (select(self->fd + 1, &readMask, NULL, NULL, &to) >= 0)
        break;

      switch (errno) {
        case EINTR:
          continue;
        case EBADF:
          goto notAlive;
        default:
          NSLog(@"socket select() failed: %s", strerror(errno));
          goto notAlive;
      }
    }

    // no input is pending, connection is alive
    if (!FD_ISSET(self->fd, &readMask)) 
      return YES;
  }

  // input is pending: If select() indicates pending input, but ioctl()
  // indicates zero bytes of pending input, the connection is broken
  {
#if defined(WIN32)
    u_long len;
#else
    int len;
#endif
    while (ioctl(self->fd, FIONREAD, &len) == -1) {
      if (errno == EINTR) continue;
      goto notAlive;
    }
    if (len > 0) return YES;
  }
  
 notAlive:
  // valid descriptor, but not alive .. so we close the socket
#if defined(WIN32)
  closesocket(self->fd);
#else
  close(self->fd);
#endif
  self->fd = NGInvalidSocketDescriptor;
  RELEASE(self->remoteAddress); self->remoteAddress = nil;
  return NO;
}
 
// ******************** NGStream ********************

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  // throws
  //   NGStreamReadErrorException    when the read call failed
  //   NGSocketNotConnectedException when the socket is not connected
  //   NGEndOfStreamException        when the end of the stream is reached
  //   NGWriteOnlyStreamException    when the receive channel was shutdown

  if (self->fd == NGInvalidSocketDescriptor) {
    THROW([[NGSocketException alloc]
              initWithReason:@"socket is not open" socket:self]);
  }

  // need to check whether socket is connected
  if (self->remoteAddress == nil) {
    THROW([[NGSocketNotConnectedException alloc]
              initWithReason:@"socket is not connected" socket:self]);
  }

  if (!NGCanReadInStreamMode(self->mode))
    THROW([[NGWriteOnlyStreamException alloc] initWithStream:self]);

  {
#if defined(WIN32)
    int readResult;

    readResult = recv(self->fd, _buf, _len, 0);
    if (readResult == 0) {
      THROW([[NGSocketShutdownException alloc] initWithStream:self]);
      abort();
    }
    else if (readResult < 0) {
      int errorCode = WSAGetLastError();

      switch (errorCode) {
        case WSAECONNRESET:
          THROW([[NGSocketConnectionResetException alloc] initWithStream:self]);
          break;
        case WSAETIMEDOUT:
          THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
          break;

        case WSAEWOULDBLOCK:
          NSLog(@"WARNING: descriptor would block ..");
          
        default:
          THROW([[NGStreamReadErrorException alloc]
                    initWithStream:self errorCode:errorCode]);
          break;
      }
    }
#else
    int readResult = NGDescriptorRecv(self->fd, _buf, _len, 0,
                                      (self->receiveTimeout == 0.0)
                                      ? -1 // block until data
                                      : (int)(self->receiveTimeout * 1000.0));
    if (readResult == 0)
      THROW([[NGSocketShutdownException alloc] initWithStream:self]);
    else if (readResult == -2)
      THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
    else if (readResult < 0) {
      int errorCode = errno;

      switch (errorCode) {
        case ECONNRESET:
          THROW([[NGSocketConnectionResetException alloc] initWithStream:self]);
          break;
        case ETIMEDOUT:
          THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
          break;

        case EWOULDBLOCK:
          NSLog(@"WARNING: descriptor would block ..");
          
        default:
          THROW([[NGStreamReadErrorException alloc]
                    initWithStream:self errorCode:errorCode]);
          break;
      }
    }
#endif  
    return readResult;
  }
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  // throws
  //   NGStreamWriteErrorException   when the write call failed
  //   NGSocketNotConnectedException when the socket is not connected
  //   NGReadOnlyStreamException     when the send channel was shutdown

  if (self->fd == NGInvalidSocketDescriptor) {
    THROW([[NGSocketException alloc]
              initWithReason:@"socket is not open" socket:self]);
  }

  // need to check whether socket is connected
  if (self->remoteAddress == nil) {
    THROW([[NGSocketNotConnectedException alloc]
              initWithReason:@"socket is not connected" socket:self]);
  }

  if (!NGCanWriteInStreamMode(self->mode))
    THROW([[NGReadOnlyStreamException alloc] initWithStream:self]);

  {
#if defined(WIN32)
    int writeResult;

    writeResult = send(self->fd, _buf, _len, 0);
    if (writeResult == 0)
      THROW([[NGSocketShutdownException alloc] initWithStream:self]);
    else if (writeResult < 0) {
      int errorCode = WSAGetLastError();
      
      switch (errorCode) {
        case WSAECONNRESET:
          THROW([[NGSocketConnectionResetException alloc] initWithStream:self]);
          break;
        case WSAETIMEDOUT:
          THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
          break;

        case WSAEWOULDBLOCK:
          NSLog(@"WARNING: descriptor would block ..");

        default:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self errorCode:errno]);
          break;
      }
    }
#else
    int writeResult = NGDescriptorSend(self->fd, _buf, _len, 0,
                                       (self->sendTimeout == 0.0)
                                       ? -1 // block until data
                                       : (int)(self->sendTimeout * 1000.0));

    if (writeResult == 0)
      THROW([[NGSocketShutdownException alloc] initWithStream:self]);
    else if (writeResult == -2)
      THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
    else if (writeResult < 0) {
      int errorCode = errno;
      
      switch (errorCode) {
        case ECONNRESET:
          THROW([[NGSocketConnectionResetException alloc] initWithStream:self]);
          break;
        case ETIMEDOUT:
          THROW([[NGSocketTimedOutException alloc] initWithStream:self]);
          break;

        case EPIPE:
          THROW([[NGSocketShutdownException alloc] initWithStream:self]);
          break;

        case EWOULDBLOCK:
          NSLog(@"WARNING: descriptor would block ..");

        default:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self errorCode:errno]);
          break;
      }
    }
#endif
    return writeResult;
  }
}

- (void)flush {
}
- (void)close {
  [self shutdown];
}

- (NGStreamMode)mode {
  return self->mode;
}

// methods method which write exactly _len bytes

- (void)safeReadBytes:(void *)_buf count:(unsigned)_len {
  int  toBeRead = _len;
  int  readResult;
  void *pos;
  unsigned (*readBytes)(id, SEL, void *, unsigned)
    = (void *)[self methodForSelector:@selector(readBytes:count:)];

  *(&pos) = _buf;

  NS_DURING {
    while (YES) {
      readResult = readBytes(self, @selector(readBytes:count:), pos, toBeRead);

      NSAssert(readResult != 0, @"ERROR: readBytes may not return '0' ..");

      if (readResult == toBeRead) {
        // all bytes were read successfully, return
        break;
      }

      if (readResult < 1) {
        [NSException raise:NSInternalInconsistencyException
                     format:@"readBytes:count: returned a value < 1"];
      }

      toBeRead -= readResult;
      pos += readResult;
    }
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      [[[[localException class] alloc]
                          initWithStream:self
                          readCount:(_len - toBeRead)
                          safeCount:_len
                          data:[NSData dataWithBytes:_buf
                                       length:(_len - toBeRead)]]
                          raise];
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;
}

- (void)safeWriteBytes:(const void *)_buf count:(unsigned)_len {
  int  toBeWritten = _len;
  int  writeResult;
  void *pos = (void *)_buf;
  int  (*writeBytes)(id,SEL,const void*,unsigned)
    = (void *)[self methodForSelector:@selector(writeBytes:count:)];

  while (YES) {
    writeResult =
      (int)writeBytes(self, @selector(writeBytes:count:), pos, toBeWritten);

    if (writeResult == toBeWritten) {
      // all bytes were written successfully, return
      break;
    }

    if (writeResult < 1) {
      [NSException raise:NSInternalInconsistencyException
                   format:@"writeBytes:count: returned a value < 1 in stream %@",
                     self];
    }

    toBeWritten -= writeResult;
    pos += writeResult;
  }
}

- (void)mark {
}
- (void)rewind {
  THROW([[NGStreamException alloc] initWithStream:self
                                   reason:@"stream doesn't support a mark"]);
}
- (BOOL)markSupported {
  return NO;
}

// convenience methods

- (int)readByte { // java semantics (-1 returned on EOF)
  int result;

  NS_DURING {
    unsigned char c;
    [self readBytes:&c count:sizeof(unsigned char)];
    result = c;
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]])
      result = -1;
    else
      [localException raise];
  }
  NS_ENDHANDLER;
  return result;
}


// description

- (NSString *)modeDescription {
  NSString *result = @"<unknown>";
  
  switch ([self mode]) {
    case NGStreamMode_undefined: result = @"<closed>"; break;
    case NGStreamMode_readOnly:  result = @"r";        break;
    case NGStreamMode_writeOnly: result = @"w";        break;
    case NGStreamMode_readWrite: result = @"rw";       break;
    default:
      THROW([[NGUnknownStreamModeException alloc] initWithStream:self]);
      break;
  }
  return result;
}

- (NSString *)description {
  NSMutableString *d = [[NSMutableString allocWithZone:[self zone]] init];

  [d appendFormat:@"<%@[0x%08X]: mode=%@ address=%@",
       NSStringFromClass([self class]),
       (unsigned)self,
       [self modeDescription], [self localAddress]];

  if ([self isConnected])
    [d appendFormat:@" connectedTo=%@", [self remoteAddress]];

  if ([self sendTimeout] != 0.0) 
    [d appendFormat:@" send-timeout=%4.3fs", [self sendTimeout]];
  if ([self receiveTimeout] != 0.0) 
    [d appendFormat:@" receive-timeout=%4.3fs", [self receiveTimeout]];

  [d appendString:@">"];
  return AUTORELEASE(d);
}

@end

@implementation NGBufferedStream(FastSocketForwarders)

- (BOOL)isConnected {
  return [(id)self->source isConnected];
}
- (int)fileDescriptor {
  return [(NSFileHandle *)self->source fileDescriptor];
}

@end
