/* 
   NGInternetSocketAddress.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGInternetSocketAddress.m,v 1.24 1999/03/26 12:42:43 helge Exp $

#include "config.h"
#if defined(HAVE_NETINET_IN_H)
#  include <netinet/in.h>
#endif

#include "common.h"
#include "NGSocketExceptions.h"
#include "NGInternetSocketAddress.h"
#include "NGInternetSocketDomain.h"

@implementation NGInternetSocketAddress

#if LIB_FOUNDATION
extern NSRecursiveLock *libFoundationLock = nil;
#define systemLock libFoundationLock
#else
static NSRecursiveLock *systemLock = nil;
#endif

#if !LIB_FOUNDATION
+ (void)initialize {
  [NGSocket initialize];
  
  [[NSNotificationCenter defaultCenter]
                         addObserver:self selector:@selector(taskNowMultiThreaded:)
                         name:NSWillBecomeMultiThreadedNotification
                         object:nil];
}

+ (void)taskNowMultiThreaded:(NSNotification *)_notification {
  if (systemLock == nil)
    systemLock = [NSRecursiveLock new];
}
#endif

static inline NSString *_nameOfLocalhost(void) {
  NSString *hostName = nil;
  
  [systemLock lock];
  {
    char buffer[1024];
    gethostname(buffer, sizeof(buffer));
    hostName = [[NSString alloc] initWithCString:buffer];
  }
  [systemLock unlock];

  return AUTORELEASE(hostName);
}

- (void)_resolveHostName {
  // throws
  //   NGCouldNotResolveHostNameException  when a DNS lookup fails
  
  // fill up the INET address
#if defined(WIN32)
  u_long *ia = &(address.sin_addr.s_addr);
#else
  unsigned int *ia = &(address.sin_addr.s_addr);
#endif

  if ([self isWildcardAddress]) {
    *ia = htonl(INADDR_ANY); // wildcard (0)
    isResolved = YES;
  }
  else {
    // try to interpret hostname as INET dotted address (eg 122.133.44.87)
    *ia = inet_addr([[self hostName] cString]);

    if (*ia != -1) { // succeeded
      isResolved = YES;
    }
    else { // failed, try to interpret hostname as DNS hostname
      BOOL didFail   = NO;
      int  errorCode = 0;
      int  addrType  = AF_INET;
#if defined(HAVE_GETHOSTBYNAME_R) && !defined(linux)
      char buffer[4096];
      struct hostent hostEntity;
#else
      struct hostent *hostEntity; // only valid during lock
#endif

#if defined(HAVE_GETHOSTBYNAME_R) && !defined(linux)
      if (gethostbyname_r([[self hostName] cString], &hostEntity,
                          buffer, sizeof(buffer), &errorCode) == NULL) {
        didFail = YES;
      }
      else {
        addrType = hostEntity.h_addrtype;

        if (addrType == AF_INET)
          *ia = ((struct in_addr *)(hostEntity.h_addr_list[0]))->s_addr;
        else
          didFail = YES; // invalid domain (eg AF_INET6)
      }
#else
      [systemLock lock];
      {
        if ((hostEntity = gethostbyname([[self hostName] cString])) == NULL) {
          didFail = YES;
#if defined(WIN32)
          errorCode = WSAGetLastError();
#else
          errorCode = h_errno;
#endif
        }
        else {
          addrType = hostEntity->h_addrtype;

          if (addrType == AF_INET)
            *ia = ((struct in_addr *)(hostEntity->h_addr_list[0]))->s_addr;
          else
            didFail = YES; // invalid domain (eg AF_INET6)
        }
      }
      [systemLock unlock];
#endif

      if (didFail) { // could not resolve hostname
        // did not find host
        NSString *reason = nil;

        if (addrType != AF_INET) {
          // invalid domain (eg AF_INET6)
          reason = @"resolved address is in invalid domain";
        }
        else {
          switch (errorCode) {
            case HOST_NOT_FOUND: reason = @"host not found"; break;
            case TRY_AGAIN:      reason = @"try again";      break;
            case NO_RECOVERY:    reason = @"no recovery";    break;
            case NO_DATA:        reason = @"no address available"; break;
            default:
              reason = [NSString stringWithFormat:@"error code %i", errorCode];
              break;
          }
        }
        THROW([[NGCouldNotResolveHostNameException alloc]
                  initWithHostName:[self hostName] reason:reason]);
      }

      isResolved = YES;
    }
  }
}

// constructors

+ (id)addressWithPort:(int)_port onHost:(id)_host {
  return AUTORELEASE([[self alloc] initWithPort:_port onHost:_host]);
}
+ (id)addressWithPort:(int)_port {
  return AUTORELEASE([[self alloc] initWithPort:_port]);
}

+ (id)addressWithService:(NSString *)_sname onHost:(id)_host
  protocol:(NSString *)_protocol {
  
  return AUTORELEASE([[self alloc] initWithService:_sname
                                   onHost:_host
                                   protocol:_protocol]);
}
+ (id)addressWithService:(NSString *)_sname protocol:(NSString *)_protocol {
  return AUTORELEASE([[self alloc] initWithService:_sname protocol:_protocol]);
}

+ (id)wildcardAddress {
  return AUTORELEASE([[self alloc] initWithPort:0 onHost:@"*"]);
}
+ (id)wildcardAddressWithPort:(int)_port {
  return AUTORELEASE([[self alloc] initWithPort:_port onHost:@"*"]);
}

- (id)initWithPort:(int)_port onHost:(id)_host { // designated initializer
  if ((self = [super init])) {
    if (_host != nil) {
      if ([_host isKindOfClass:[NSHost class]])
        _host = [(NSHost *)_host address];
      else if ([_host isEqualToString:@"*"])
        self->hostName = nil; // wildcard
      else
        self->hostName = [_host copyWithZone:[self zone]];
    }
    // else wildcard

    address.sin_family = [[self domain] socketDomain];
    address.sin_port   = htons((short)(_port & 0xffff));

    isResolved = NO;
  }
  return self;
}

- (id)initWithService:(NSString *)_serviceName onHost:(id)_host
  protocol:(NSString *)_protocol {
  
  NSException *exc = nil;
  int port = -1;
#if defined(HAVE_GETSERVBYNAME_R)
  char   buffer[2048];
  struct servent entry;
#else
  struct servent *entry;
#endif

#if defined(HAVE_GETSERVBYNAME_R)
  if (getservbyname_r((char *)[_serviceName cString], [_protocol cString],
                      &entry, buffer, sizeof(buffer)) == NULL) {
    exc = [[NGDidNotFindServiceException alloc] initWithServiceName:_serviceName];
  }
  else
    port = entry.s_port;
#else
  [systemLock lock];
  {
    entry = getservbyname((char *)[_serviceName cString], [_protocol cString]);
    if (entry == NULL)
      exc = [[NGDidNotFindServiceException alloc] initWithServiceName:_serviceName];
    else
      port = entry->s_port;
  }
  [systemLock unlock];
#endif

  if (exc) {
    AUTORELEASE(self);
    THROW(exc);
    return nil;
  }
  return [self initWithPort:port onHost:_host];
}

- (id)initWithPort:(int)_port {
  return [self initWithPort:_port onHost:_nameOfLocalhost()];
}

- (id)initWithService:(NSString *)_serviceName protocol:(NSString *)_protocol {
  return [self initWithService:_serviceName
               onHost:_nameOfLocalhost()
               protocol:_protocol];
}

- (id)initWithDomain:(id)_domain
  internalRepresentation:(void *)_representation
  size:(int)_length {
  
  struct hostent     *hostEntity = NULL; // only valid during lock
  struct sockaddr_in *sockAddr = _representation;
  NSString           *newHost  = nil;
  int                errorCode = 0;

  NSAssert(_length == sizeof(struct sockaddr_in), @"invalid socket address length");

  if (sockAddr->sin_addr.s_addr != 0) { // not a wildcard address
#if !defined(HAVE_GETHOSTBYADDR_R)
    [systemLock lock];
#endif
    {
      BOOL done = NO;

      while (!done) {
#if defined(HAVE_GETHOSTBYADDR_R) && !defined(linux)
        struct hostent hostEntityBuffer;
        char buffer[8200];

        hostEntity = gethostbyaddr_r((char *)&(sockAddr->sin_addr.s_addr),
                                     4,
                                     [_domain socketDomain],
                                     &hostEntityBuffer,
                                     buffer, 8200,
                                     &errorCode);
#else
        hostEntity = gethostbyaddr((char *)&(sockAddr->sin_addr.s_addr),
                                   4,
                                   [_domain socketDomain]);
#  if defined(WIN32)
        errorCode = WSAGetLastError();
#  else
        errorCode = h_errno;
#  endif
#endif

        if (hostEntity == NULL) {
          done = YES;
        
          switch (errorCode) {
            case HOST_NOT_FOUND:
              NSLog(@"host not found ..");
              break;

            case TRY_AGAIN:
              NSLog(@"try again");
              done = NO;
              break;
            
            case NO_RECOVERY:
              NSLog(@"no recovery");
              break;
            
            case NO_DATA:
              NSLog(@"no data");
              break;
            
            default:
              NSLog(@"unknown error (length was %i): h_errno=%i errno=%s",
                    _length, errorCode, strerror(errno));
              break;
          }

          newHost = [NSString stringWithCString:inet_ntoa(sockAddr->sin_addr)];
        }
        else {
          newHost = [NSString stringWithCString:hostEntity->h_name];
          done = YES;
        }
      }
    }
#if !defined(HAVE_GETHOSTBYADDR_R)
    [systemLock unlock];
#endif

    if (hostEntity == NULL) {
      // throw could not get address ..
      NSLog(@"could not get DNS name of address %@ in domain %@: %i",
            newHost, _domain, errorCode);
    }
  }
  else {
    // wildcard address
    newHost = nil;
  }

  return [self initWithPort:sockAddr->sin_port onHost:newHost];
}


#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->hostName); self->hostName = nil;
  [super dealloc];
}
#endif

// accessors

- (NSString *)hostName {
  return AUTORELEASE([self->hostName copy]);
}
- (NSString *)address {
#if defined(WIN32)
  u_long *ia;
  ia = (u_long *)&(address.sin_addr.s_addr);
#else
  unsigned int ia;
  ia = (unsigned int)&(address.sin_addr.s_addr);
#endif

  if (self->hostName == nil) // wildcard
    return nil;
  
  if (!isResolved)
    [self _resolveHostName];

  {
    char     *ptr = NULL;
    NSString *str = nil;

    [systemLock lock];
    {
      ptr = inet_ntoa(*((struct in_addr *)ia));
      str = [NSString stringWithCString:ptr];
    }
    [systemLock unlock];

    return str;
  }
}

- (int)port {
  // how to do ?
  return ntohs(address.sin_port);
}

- (BOOL)isWildcardAddress {
  return (self->hostName == nil);
}

// NGSocketAddress protocol

- (void *)internalAddressRepresentation {
  // throws
  //   NGCouldNotResolveHostNameException  when a DNS lookup fails
  
  if (!isResolved)
    [self _resolveHostName];
  
  return &address;
}

- (int)addressRepresentationSize {
  return [[self domain] addressRepresentationSize];
}
- (id)domain {
  return [NGInternetSocketDomain domain];
}

// comparing

- (BOOL)isEqualToAddress:(NGInternetSocketAddress *)_otherAddress {
  if (![[_otherAddress hostName] isEqualToString:[self hostName]])
    return NO;
  if ([_otherAddress port] != [self port])
    return NO;
  return YES;
}

- (BOOL)isEqual:(id)_object {
  if (_object == self) return YES;
  if ([_object class] != [self class]) return NO;
  return [self isEqualToAddress:_object];
}

// NSCopying

- (id)copyWithZone:(NSZone *)_zone {
  // socket addresses are immutable, therefore just retain self
  return RETAIN(self);
}

// NSCoding

- (void)encodeWithCoder:(NSCoder *)_encoder {
  int aPort = [self port];

  [_encoder encodeValueOfObjCType:@encode(int) at:&aPort];
  [_encoder encodeObject:[self hostName]];
}
- (id)initWithCoder:(NSCoder *)_decoder {
  int aPort;
  id  aHost;

  [_decoder decodeValueOfObjCType:@encode(int) at:&aPort];
  aHost = [_decoder decodeObject];

  return [self initWithPort:aPort onHost:aHost];
}

// description

- (NSString *)stringValue {
  NSString *name = [self hostName];
  if (name == nil) name = @"*";
  return [NSString stringWithFormat:@"%@:%i", name, [self port]];
}

- (NSString *)description {
  NSString *name = [self hostName];

  if (name == nil) name = @"*";
  
  return [NSString stringWithFormat:@"<InetSocketAddress: %@:%i>",
                     name, [self port]];
}

@end

@implementation NGActiveSocket(NGInternetActiveSocket)

+ (id)socketConnectedToPort:(int)_port onHost:(id)_host {
  // this method calls +socketConnectedToAddress: with an NGInternetSocketAddress
  
  return [self socketConnectedToAddress:
                 [NGInternetSocketAddress addressWithPort:_port onHost:_host]];
}

- (void)connectToPort:(int)_port onHost:(id)_host {
  // this method calls -connectToAddress: with an NGInternetSocketAddress

  return [self connectToAddress:
                 [NGInternetSocketAddress addressWithPort:_port onHost:_host]];
}

@end
