/* 
   NGPassiveSocket.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGPassiveSocket.m,v 1.12 1999/01/27 17:02:07 helge Exp $

#include "common.h"
#include "NGPassiveSocket.h"
#include "NGSocketExceptions.h"
#include "NGActiveSocket.h"
#include "NGSocket+private.h"

@interface NGActiveSocket(privateMethods)

- (id)_initWithDescriptor:(int)_fd
  localAddress:(id<NGSocketAddress>)_local
  remoteAddress:(id<NGSocketAddress>)_remote;

@end

@implementation NGPassiveSocket

+ (id)socketBoundToAddress:(id<NGSocketAddress>)_address {
  volatile id sock = [[self alloc] initWithDomain:[_address domain]];

  if (sock != nil) {
    AUTORELEASE(sock);
    [sock bindToAddress:_address];
  }
  return sock;
}

- (id)initWithDomain:(id<NGSocketDomain>)_domain { // designated initializer
  if ((self = [super initWithDomain:_domain])) {
    backlogSize = -1; // -1 means 'not listening'
    
    if ([NSThread isMultiThreaded])
      acceptLock = [[NSLock allocWithZone:[self zone]] init];
    else {
      acceptLock = nil;
      [[NSNotificationCenter defaultCenter]
                             addObserver:self
                             selector:@selector(taskNowMultiThreaded:)
                             name:NSWillBecomeMultiThreadedNotification
                             object:nil];
    }

    if (self->fd != NGInvalidSocketDescriptor) {
      int i_yes = 1;
      
      if (setsockopt(self->fd, SOL_SOCKET, SO_REUSEADDR,
                     (void *)&i_yes, sizeof(int)) != 0) {
        NSLog(@"WARNING: could not set SO_REUSEADDR option for socket %@: %s",
              self, strerror(errno));
      }
    }
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  [[NSNotificationCenter defaultCenter]
                         removeObserver:self
                         name:NSWillBecomeMultiThreadedNotification
                         object:nil];
  
  RELEASE(acceptLock); acceptLock = nil;
  [super dealloc];
}
#endif

- (void)taskNowMultiThreaded:(NSNotification *)_notification {
  if (acceptLock == nil)
    acceptLock = [[NSLock allocWithZone:[self zone]] init];
}

// accessors

- (BOOL)isListening {
  return (backlogSize != -1);
}

- (id<NGSocketAddress>)localAddress {
  return localAddress;
}

- (int)socketType {
  return SOCK_STREAM;
}

// operations

- (void)listenWithBacklog:(int)_backlogSize {
  // throws
  //   NGSocketIsAlreadyListeningException  when the socket is in the listen state
  //   NGCouldNotListenException            when the listen call failed
  
  if ([self isListening]) {
    THROW([[NGSocketIsAlreadyListeningException alloc]
              initWithReason:@"already called listen" socket:self]);
  }

  if (listen([self fileDescriptor], _backlogSize) != 0) {
    NSString *reason = nil;
#if defined(WIN32)
    int errorCode = WSAGetLastError();

    switch (errorCode) {
      case WSAEBADF:
        reason = @"not a valid socket descriptor";
        break;
      case WSAENOTSOCK:
        reason = @"descriptor is not a socket descriptor";
        break;
      case WSAEOPNOTSUPP:
        reason = @"socket does not support listen";
        break;

      default:
        reason = [NSString stringWithCString:strerror(errorCode)];
        break;
    }
#else
    int errorCode = errno;

    switch (errorCode) {
      case EBADF:
        reason = @"not a valid socket descriptor";
        break;
      case ENOTSOCK:
        reason = @"descriptor is not a socket descriptor";
        break;
      case EOPNOTSUPP:
        reason = @"socket does not support listen";
        break;

      default:
        reason = [NSString stringWithCString:strerror(errorCode)];
        break;
    }
#endif

    reason = [NSString stringWithFormat:@"Could not listen: %@", reason];

    THROW([[NGCouldNotListenException alloc]
              initWithReason:reason socket:self]);
  }

  // set backlog size (and mark socket as 'listening')
  self->backlogSize = _backlogSize;
}

- (id<NGActiveSocket>)accept {
  // throws
  //   NGCouldNotAcceptException  when the socket is not listening
  //   NGCouldNotAcceptException  when the accept call failed

  id<NGActiveSocket> socket;
  *(&socket) = nil;
  
  if (![self isListening]) {
    THROW([[NGCouldNotAcceptException alloc]
              initWithReason:@"socket is not listening" socket:self]);
  }
  
  SYNCHRONIZED(self->acceptLock) {
    id<NGSocketAddress> local  = nil;
    id<NGSocketAddress> remote = nil;
    int  len   = [[self domain] addressRepresentationSize];
    char data[len];
    int  newFd = NGInvalidSocketDescriptor;

    if ((newFd = accept(fd, (void *)&(data[0]), &len)) == -1) {
      // call failed
      NSString *reason = nil;
#if defined(WIN32)
      int errorCode = WSAGetLastError();

      switch (errorCode) {
        case WSAEINTR:
          reason = @"interrupted by signal";
          break;
        case WSAEMFILE:
          reason = @"descriptor table is full";
          break;
          
        default:
          reason = [NSString stringWithCString:strerror(errorCode)];
          break;
      }
#else
      int errorCode = errno;

      switch (errorCode) {
        case EINTR:
          reason = @"interrupted by signal";
          break;
        case EMFILE:
          reason = @"descriptor table is full";
          break;
        case ENOSR:
          reason = @"insufficient STREAMS resources available";
          break;
        case EPROTO:
          reason = @"a protocol error has occured";
          break;
          
        default:
          reason = [NSString stringWithCString:strerror(errorCode)];
          break;
      }
#endif

      reason = [NSString stringWithFormat:@"Could not accept: %@", reason];

      THROW([[NGCouldNotAcceptException alloc]
                initWithReason:reason socket:self]);
    }

    // produce remote socket address object
    remote = [[self domain] addressWithRepresentation:(void *)&(data[0]) size:len];
    
    // getsockname if wildcard-IP-bind to get local IP address assigned to the conn
    if (getsockname(newFd, (void *)&data, &len) != 0) { // function is MT-safe
      THROW([[NGSocketException alloc]
                initWithReason:@"could not get local socket name" socket:self]);
    }
    local = [[self domain] addressWithRepresentation:(void *)data size:len];

    socket = [[NGActiveSocket allocWithZone:[self zone]]
                              _initWithDescriptor:newFd
                              localAddress:local
                              remoteAddress:remote];
    AUTORELEASE(socket);
  }
  END_SYNCHRONIZED;
  return socket;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:@"<PassiveSocket: address=%@>",
                     [self localAddress]];
}

@end
