/* 
   NGBufferedStream.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGBufferedStream.m,v 1.10 1999/01/29 13:36:07 helge Exp $

#include "common.h"
#include "NGBufferedStream.h"

#define NEWLINE_CHAR '\n'

@implementation NGBufferedStream

static const unsigned DEFAULT_BUFFER_SIZE = 512;

// returns the number of bytes which where read from the buffer
#define numberOfConsumedReadBufferBytes(self) \
  ((self->readBufferSize == 0) ? 0 : (self->readBufferPos - self->readBuffer))

// returns the number of bytes which can be read from buffer (without source access)
#define numberOfAvailableReadBufferBytes(self) \
  (self->readBufferFillSize - numberOfConsumedReadBufferBytes(self))

// look whether all bytes in the buffer where consumed, if so, reset the buffer
#define checkReadBufferFillState(self) \
  if (numberOfAvailableReadBufferBytes(self) == 0) { \
    self->readBufferPos = self->readBuffer; \
    self->readBufferFillSize = 0;  \
  }

// ******************** constructors ********************

+ (id)filterWithSource:(id<NGStream>)_source bufferSize:(unsigned)_size {
  return AUTORELEASE([[self alloc] initWithSource:_source bufferSize:_size]);
}

- (id)initWithSource:(id<NGStream>)_source bufferSize:(unsigned)_size {
  if ((self = [super initWithSource:_source])) {
    self->readBuffer  = NSZoneMallocAtomic(NULL, _size);
    self->writeBuffer = NSZoneMallocAtomic(NULL, _size);

    self->readBufferPos       = self->readBuffer;
    self->readBufferSize      = _size;
    self->readBufferFillSize  = 0; // no bytes are read from source
    self->writeBufferFillSize = 0;
    self->writeBufferSize     = _size;
    self->flags._flushOnNewline = 1;
  }
  return self;
}

- (id)initWithInputSource:(id<NGInputStream>)_source bufferSize:(unsigned)_size {
  if ((self = [super initWithInputSource:_source])) {
    self->readBuffer            = NSZoneMallocAtomic(NULL, _size);
    self->readBufferPos         = self->readBuffer;
    self->readBufferSize        = _size;
    self->readBufferFillSize    = 0; // no bytes are read from source
    self->flags._flushOnNewline = 1;
  }
  return self;
}
- (id)initWithOutputSource:(id<NGOutputStream>)_src bufferSize:(unsigned)_size {
  if ((self = [super initWithOutputSource:_src])) {
    self->writeBuffer           = NSZoneMallocAtomic(NULL, _size);
    self->writeBufferFillSize   = 0;
    self->writeBufferSize       = _size;
    self->flags._flushOnNewline = 1;
  }
  return self;
}

- (id)initWithSource:(id<NGStream>)_source {
  return [self initWithSource:_source bufferSize:DEFAULT_BUFFER_SIZE];
}
- (id)initWithInputSource:(id<NGInputStream>)_source {
  return [self initWithInputSource:_source bufferSize:DEFAULT_BUFFER_SIZE];
}
- (id)initWithOutputSource:(id<NGOutputStream>)_source {
  return [self initWithOutputSource:_source bufferSize:DEFAULT_BUFFER_SIZE];
}

- (void)gcFinalize {
  [self flush];
}

#if !(LIB_FOUNDATION_BOEHM_GC)
- (void)dealloc {
  [self gcFinalize];
  
  if (self->readBuffer) {
    NSZoneFree(NULL, self->readBuffer);
    self->readBuffer    = NULL;
    self->readBufferPos = NULL;
  }
  self->readBufferFillSize = 0;
  self->readBufferSize     = 0;

  if (self->writeBuffer) {
    NSZoneFree(NULL, self->writeBuffer);
    self->writeBuffer = NULL;
  }
  self->writeBufferFillSize = 0;
  self->writeBufferSize     = 0;
  [super dealloc];
}
#endif

// accessors

- (void)setReadBufferSize:(unsigned)_size {
  [self flush];

  if (_size == self->readBufferSize)
    return;

  if (_size == 0) {
    if (self->readBuffer != NULL) {
      NSZoneFree(NULL, self->readBuffer);
      self->readBuffer = NULL;
    }
    self->readBufferSize = _size;
    self->readBufferPos  = NULL;
  }
  else {
    if (self->readBuffer != NULL)
      self->readBuffer = NSZoneRealloc(NULL, self->readBuffer, _size);
    else
      self->readBuffer = NSZoneMallocAtomic(NULL, _size);

    self->readBufferSize     = _size;
    self->readBufferPos      = self->readBuffer;
    self->readBufferFillSize = 0; // no bytes a read from source
  }
}
- (unsigned)readBufferSize {
  return self->readBufferSize;
}

- (void)setWriteBufferSize:(unsigned)_size {
  [self flush];

  if (_size == self->writeBufferSize)
    return;

  self->writeBuffer = NSZoneRealloc(NULL, self->writeBuffer, _size);
  self->writeBufferSize = _size;
}
- (unsigned)writeBufferSize {
  return self->writeBufferSize;
}

// blocking ..

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  BOOL canRead, canWrite;

  if (self->readBufferSize == 0)
    canRead = NO;
  else
    canRead = (numberOfAvailableReadBufferBytes(self) > 0);

  if (self->writeBufferSize == 0)
    canWrite = NO;
  else
    canWrite = (self->writeBufferFillSize > 0);

  if ((_mode == NGStreamMode_readWrite) && canRead && canWrite)
    return NO;
  if ((_mode == NGStreamMode_readOnly) && canRead) {
    //NSLog(@"%@: %i bytes available ..", self,
    //      numberOfAvailableReadBufferBytes(self));
    return NO;
  }
  if ((_mode == NGStreamMode_writeOnly) && canWrite)
    return NO;

  if ([self->source respondsToSelector:@selector(wouldBlockInMode:)])
    return [(id)self->source wouldBlockInMode:_mode];
  else
    return YES;
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  register unsigned availBytes = numberOfAvailableReadBufferBytes(self);
  
  if (self->readBufferSize == 0) { // no read buffering is done (buffersize==0)
    return (readBytes != NULL)
      ? readBytes(source, _cmd, _buf, _len)
      : [source readBytes:_buf count:_len];
  }
    
  if (availBytes >= _len) {
    // there are enough bytes in the buffer to fulfill the request
    if (_len == 1) {
      *(unsigned char *)_buf = *(unsigned char *)self->readBufferPos;
      self->readBufferPos++;
    }
    else {
      memcpy(_buf, self->readBufferPos, _len);
      self->readBufferPos += _len;          // update read position (consumed-size)
    }
    checkReadBufferFillState(self); // check whether all bytes where consumed
    return _len;
  }
  else if (availBytes > 0) {
    // there are some bytes in the buffer, these are returned
    
    memcpy(_buf, self->readBufferPos, availBytes); // copy all bytes from buffer
    self->readBufferPos      = self->readBuffer; // reset position
    self->readBufferFillSize = 0;          // no bytes available in buffer anymore
    return availBytes;
  }
  else if (_len > self->readBufferSize) {
    // requested _len is bigger than the buffersize, so we can bypass the
    // buffer (which is empty, as guaranteed by the previous 'ifs'

    NSAssert(self->readBufferPos == self->readBuffer, @"read buffer position is not reset");
    NSAssert(self->readBufferFillSize == 0,     @"there are bytes in the buffer");

    availBytes = (readBytes != NULL)
      ? (unsigned)readBytes(source, _cmd, _buf, _len)
      : [source readBytes:_buf count:_len];

    NSAssert(availBytes != 0, @"readBytes:count: may never return zero !");

    return availBytes; // return the number of bytes which could be read
  }
  else {
    // no bytes are available and the requested _len is smaller than the possible
    // buffer size, we have to read the next block of input from the source

    NSAssert(self->readBufferPos == self->readBuffer, @"read buffer position is not reset");
    NSAssert(self->readBufferFillSize == 0,     @"there are bytes in the buffer");

    self->readBufferFillSize = (readBytes != NULL)
      ? (unsigned)readBytes(source, _cmd, self->readBuffer, self->readBufferSize)
      : [source readBytes:self->readBuffer count:self->readBufferSize];

    NSAssert(self->readBufferFillSize != 0, @"readBytes:count: may never return zero !");

    // no comes a section which is roughly the same like the first to conditionals
    // in this method
    if (self->readBufferFillSize >= _len) {
      // there are enough bytes in the buffer to fulfill the request
    
      memcpy(_buf, self->readBufferPos, _len);
      self->readBufferPos += _len;          // update read position (consumed-size)
      checkReadBufferFillState(self); // check whether all bytes where consumed
      return _len;
    }
    else { // (readBufferFillSize > 0) (this is ensured by the above assert)
      // there are some bytes in the buffer, these are returned

      availBytes = self->readBufferFillSize;
      memcpy(_buf, self->readBufferPos, self->readBufferFillSize); // copy all bytes from buffer
      self->readBufferPos      = self->readBuffer; // reset position
      self->readBufferFillSize = 0;          // no bytes available in buffer anymore
      return availBytes;
    }
  }
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  register unsigned tmp       = 0;
  register unsigned remaining = _len;
  register void     *track    = (void *)_buf;

  while (remaining > 0) {
    // how much bytes available in buffer ?
    tmp = self->writeBufferSize - self->writeBufferFillSize; 
    tmp = (tmp > remaining) ? remaining : tmp;
  
    memcpy((self->writeBuffer + self->writeBufferFillSize), track, tmp);
    track += tmp;
    remaining -= tmp;
    self->writeBufferFillSize += tmp;

    if (self->writeBufferFillSize == self->writeBufferSize) {
      [source safeWriteBytes:self->writeBuffer count:self->writeBufferFillSize];
      self->writeBufferFillSize = 0;
    }
  }

  if (self->flags._flushOnNewline == 1) {
    // scan buffer for newlines, if one is found, flush buffer
    
    for (tmp = 0; tmp < _len; tmp++) {
      if (tmp == NEWLINE_CHAR) {
        [self flush];
        break;
      }
    }
  }
  
  // clean up for GC
  tmp       = 0;    
  track     = NULL; // clean up for GC
  remaining = 0;
  
  return _len;
}

- (void)close {
  [self flush];
  
  if (self->readBuffer) {
    NSZoneFree(NULL, self->readBuffer);
    self->readBuffer = NULL;
    self->readBufferPos = NULL;
  }
  self->readBufferFillSize = 0;
  self->readBufferSize = 0;

  if (self->writeBuffer) {
    NSZoneFree(NULL, self->writeBuffer);
    self->writeBuffer = NULL;
  }
  self->writeBufferFillSize = 0;
  self->writeBufferSize = 0;
  
  [super close];
}

- (void)flush {
  if (self->writeBufferFillSize > 0) {
    [source safeWriteBytes:self->writeBuffer count:self->writeBufferFillSize];
    self->writeBufferFillSize = 0;
  }
}

@end

@implementation NGStream(NGBufferedStreamExtensions)

- (NGBufferedStream *)bufferedStream {
  return [NGBufferedStream filterWithSource:self];
}

@end

@implementation NGBufferedStream(NGBufferedStreamExtensions)

- (NGBufferedStream *)bufferedStream {
  return self;
}

@end
