/* 
   NGCTextStream.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGCTextStream.m,v 1.13 1999/01/27 16:41:13 helge Exp $

#import <NGExtensions/NGCharBuffers.h>
#import <NGExtensions/NGTool.h>
#include "common.h"
#include "NGCTextStream.h"
#include "NGStreamExceptions.h"
#include "NGFileStream.h"

static NSString *NGNewLineString = @"\n";

@interface _NGCTextStreamLineEnumerator : NSEnumerator
{
  NGCTextStream *stream;
}
- (id)initWithTextStream:(NGCTextStream *)_stream;
- (id)nextObject;
@end

id<NGExtendedTextInputStream>  NGTextIn  = nil;
id<NGExtendedTextOutputStream> NGTextOut = nil;
id<NGExtendedTextOutputStream> NGTextErr = nil;

@implementation NGCTextStream

// stdio

void NGInitTextStdio(void) {
  static BOOL isInitialized = NO;
  if (!isInitialized) {
    isInitialized = YES;

    NGInitStdio();
    
    NGTextIn  = [[NGCTextStream alloc]
                    initWithSource:(id<NGStream>)NGIn];
    NGTextOut = [[NGCTextStream alloc]
                    initWithSource:(id<NGStream>)NGOut];
    NGTextErr = [[NGCTextStream alloc]
                    initWithSource:(id<NGStream>)NGErr];
  }
}

+ (void)_flushForExit:(NSNotification *)_notification {
  [NGTextIn  flush]; RELEASE(NGTextIn);  NGTextIn  = nil;
  [NGTextOut flush]; RELEASE(NGTextOut); NGTextOut = nil;
  [NGTextErr flush]; RELEASE(NGTextErr); NGTextErr = nil;
}

+ (void)initialize {
  BOOL isInitialized = NO;
  if (!isInitialized) {
    isInitialized = YES;

    [[NSNotificationCenter defaultCenter]
                           addObserver:self
                           selector:@selector(_flushForExit:)
                           name:NGProcessWillExitNotificationName
                           object:nil];
    [[NSNotificationCenter defaultCenter]
                           addObserver:self
                           selector:@selector(_flushForExit:)
                           name:NGProcessWillAbortNotificationName
                           object:nil];
  }
}

// system text stream

+ (id)textStreamWithInputSource:(id<NGInputStream>)_source {
  return AUTORELEASE([[self alloc] initWithInputSource:_source]);
}
+ (id)textStreamWithOutputSource:(id<NGOutputStream>)_source {
  return AUTORELEASE([[self alloc] initWithOutputSource:_source]);
}
+ (id)textStreamWithSource:(id<NGStream>)_stream {
  return AUTORELEASE([[self alloc] initWithSource:_stream]);
}

- (id)initWithSource:(id<NGStream>)_stream {
  if ((self = [super init])) {
    source = RETAIN(_stream);
    
    if ([source isKindOfClass:[NSObject class]]) {
      readBytes   = (NGIOReadMethodType)
        [(NSObject *)source methodForSelector:@selector(readBytes:count:)];
      writeBytes  = (NGIOWriteMethodType)
        [(NSObject *)source methodForSelector:@selector(writeBytes:count:)];
      flushBuffer = (void (*)(id,SEL))
        [(NSObject *)source methodForSelector:@selector(flush)];
    }
  }
  return self;
}
- (id)initWithInputSource:(id<NGInputStream>)_source {
  return [self initWithSource:(id)_source];
}
- (id)initWithOutputSource:(id<NGOutputStream>)_source {
  return [self initWithSource:(id)_source];
}

- (void)dealloc {
  RELEASE(source); source = nil;
  readBytes   = NULL;
  writeBytes  = NULL;
  flushBuffer = NULL;
  [super dealloc];
}

// accessors

- (id<NGStream>)source {
  return source;
}

// operations

- (void)close {
  [source close];
}

// NGTextInputStream

- (unichar)readCharacter {
  return [self readChar];
}

- (char)readChar {
  unsigned char c;

  if (readBytes)
    readBytes(source, @selector(readBytes:count:), &c, sizeof(unsigned char));
  else
    [source readBytes:&c count:sizeof(unsigned char)];

  return c;
}

- (NSString *)readLineAsString {
  NGCharBuffer8   buffer = NULL;
  unsigned char   c;

  buffer = NGCharBuffer8_new(128);

  NS_DURING {
    if (readBytes) {
      do {
        readBytes(source, @selector(readBytes:count:),
                  &c, sizeof(unsigned char));
        if (c == '\r')
          readBytes(source, @selector(readBytes:count:),
                    &c, sizeof(unsigned char));

        if ((c != '\n') && (c != 0)) {
          NSAssert1(c != 0, @"tried to add '0' character to buffer '%s' ..",
                    buffer->buffer);
          NGCharBuffer8_addChar(buffer, c);
          //NSLog(@"added %i, length now %i (buffer '%s')",
          //      (int)c, (int)buffer->length, buffer->buffer);
        }
      }
      while ((c != '\n') && (c != 0));
    }
    else {
      do {
        [source readBytes:&c count:sizeof(unsigned char)];
        if (c == '\r')
          [source readBytes:&c count:sizeof(unsigned char)];
      
        if ((c != '\n') && (c != 0))
          NGCharBuffer8_addChar(buffer, c);
      }
      while ((c != '\n') && (c != 0));
    }
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      if (buffer->length == 0) {
        NGCharBuffer8_dealloc(buffer);
        buffer = NULL;
      }
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;

  return buffer ? NGCharBuffer8_makeStringAndDealloc(buffer) : nil;
}

- (NSEnumerator *)lineEnumerator {
  return AUTORELEASE([[_NGCTextStreamLineEnumerator alloc]
                          initWithTextStream:self]);
}

// NGTextOutputStream

- (void)writeCharacter:(unichar)_character {
  if (_character > ((sizeof(unsigned char) * 256) - 1)) {
    // character is not in range of maximum system encoding
    [NSException raise:@"NGCTextStreamEncodingException"
                 format:
                   @"called writeCharacter: with character code (0x%X) exceeding"
                   @" the maximum system character code (0x%X)",
                   _character, ((sizeof(unsigned char) * 256) - 1)];
  }
  else {
    unsigned char c = _character;

    if (writeBytes)
      writeBytes(source, @selector(writeBytes:count:), &c, sizeof(unsigned char));
    else
      [source writeBytes:&c count:sizeof(unsigned char)];
  }
}

- (void)writeString:(NSString *)_string {
  const char *str = [_string cString];
  int strLen      = [_string cStringLength];

  if (writeBytes)
    writeBytes(source, @selector(writeBytes:count:), str, strLen);
  else
    [source writeBytes:str count:strLen];
}

- (void)flush {
  if (flushBuffer)
    flushBuffer(source, @selector(flush));
  else
    [source flush];
}

- (void)writeNewline {
  [self writeString:NGNewLineString];
  [self flush];
}

@end

@implementation _NGCTextStreamLineEnumerator

- (id)initWithTextStream:(NGCTextStream *)_stream {
  stream = RETAIN(_stream);
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(stream); stream = nil;
  [super dealloc];
}
#endif

- (id)nextObject {
  id result = nil;
  
  NS_DURING {
    result = [stream readLineAsString];
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]])
      result = nil;
    else
      [localException raise];
  }
  NS_ENDHANDLER;
  
  return result;
}

@end
