/* 
   NGDataStream.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDataStream.m,v 1.3 1999/01/27 16:11:15 helge Exp $

#include "common.h"
#include "NGDataStream.h"
#include "NGStreamExceptions.h"

@implementation NGDataStream

+ (id)dataStream {
  return [self streamWithData:[NSMutableData dataWithCapacity:1024]];
}
+ (id)dataStreamWithCapacity:(int)_capacity {
  return [self streamWithData:[NSMutableData dataWithCapacity:_capacity]];
}

+ (id)streamWithData:(NSData *)_data {
  return AUTORELEASE([[self alloc] initWithData:_data]);
}
- (id)initWithData:(NSData *)_data {
  if ((self = [super init])) {
    data     = RETAIN(_data);
    position = 0;

    streamMode = [data isKindOfClass:[NSMutableData class]]
      ? NGStreamMode_readWrite
      : NGStreamMode_readOnly;
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(data); data = nil;
  [super dealloc];
}
#endif

// accessors

- (NSData *)data {
  return data;
}

- (unsigned)availableBytes {
  // returns number of available bytes
  unsigned currentLength = [data length];

  return (currentLength == position)
    ? 0
    : (currentLength - position);
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  // throws
  //   NGStreamNotOpenException   when the stream is not open
  //   NGEndOfStreamException     when the end of the stream is reached
  unsigned currentLength = [data length];

  if (data == nil) {
    THROW([[NGStreamNotOpenException alloc] initWithStream:self reason:
            @"tried to read from a data stream which was closed"]);
  }

  if (currentLength == position)
    THROW([[NGEndOfStreamException alloc] initWithStream:self]);

  {
    NSRange range;
    range.location = position;

    if ((position + _len) > currentLength)
      range.length = currentLength - position;
    else
      range.length = _len;

    [data getBytes:_buf range:range];

    position += range.length;
    return range.length;
  }
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  // throws
  //   NGReadOnlyStreamException when the stream is not writeable
  //   NGStreamNotOpenException  when the stream is not open
  
  if (data == nil) {
    THROW([[NGStreamNotOpenException alloc] initWithStream:self reason:
            @"tried to write to a data stream which was closed"]);
  }
  if (!NGCanWriteInStreamMode(streamMode))
    THROW([[NGReadOnlyStreamException alloc] initWithStream:self]);

  [(NSMutableData *)data appendBytes:_buf length:_len];

  return _len;
}

- (void)close {
  RELEASE(data); data = nil;
  position   = 0;
  streamMode = NGStreamMode_undefined;
}

- (NGStreamMode)mode {
  return streamMode;
}
- (BOOL)isRootStream {
  return YES;
}

// NGPositionableStream

- (void)moveToLocation:(unsigned)_location {
  position = _location;
}
- (void)moveByOffset:(int)_delta {
  position += _delta;
}

// blocking ..

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  return NO;
}

@end
