/*
   NGDescriptorFunctions.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDescriptorFunctions.m,v 1.15 1999/01/29 12:38:29 helge Exp $

#if !defined(WIN32)

// similiar functions for Windows can be found in NGSocket.[hm]

#include "config.h"
#include "common.h"
#include "NGDescriptorFunctions.h"
#include "NGStreamExceptions.h"

#ifdef HAVE_POLL
#  ifdef HAVE_POLL_H
#    include <poll.h>
#  endif
#  ifdef HAVE_SYS_POLL_H
#    include <sys/poll.h>
#  endif
#  ifndef POLLRDNORM
#    define POLLRDNORM POLLIN /* needed on Linux */
#  endif
#else
#  ifdef HAVE_SELECT_H
#    include <select.h>
#  endif
#endif

#ifdef HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
#endif

// ******************** Poll *********************

int NGPollDescriptor(int _fd, short _events, int _timeout) {
#ifdef HAVE_POLL
  struct pollfd pfd;
  int           result;

  pfd.fd      = _fd;
  pfd.events  = _events;
  pfd.revents = 0;

  do {
    result = poll(&pfd, 1, _timeout);

    if (result < 0) { // error
      int e = errno;

      if ((e != EAGAIN) && (e != EINTR)) 
        // only retry of interrupted or repeatable
        break;
    }
  }
  while (result < 0);

  return (result < 0) ? -1 : result;
#else
  struct timeval timeout;
  fd_set rSet;
  fd_set wSet;
  fd_set eSet;
  int    result;
  FD_ZERO(&rSet);
  FD_ZERO(&wSet);
  FD_ZERO(&eSet);

  if (_events & POLLIN)  FD_SET(_fd, &rSet);
  if (_events & POLLOUT) FD_SET(_fd, &wSet);
  if (_events & POLLERR) FD_SET(_fd, &eSet);

  timeout.tv_sec  = _timeout / 1000;
  timeout.tv_usec = _timeout * 1000 - timeout.tv_sec * 1000000;

  do {
    result = select(FD_SETSIZE, &rSet, &wSet, &eSet, &timeout);
    if (result == -1) { // error
      int e = errno;
      if ((e != EAGAIN) && (e != EINTR)) 
        // only retry of interrupted or repeatable
        break;
    }
  }
  while (result == -1);

  return (result < 0) ? -1 : result;
#endif
}

// ******************** Flags ********************

int NGGetDescriptorFlags(int _fd) {
  int val;

  val = fcntl(_fd, F_GETFL, 0);
  if (val < 0)
    THROW([[NGIOException alloc]
              initWithReason:@"could not get descriptor flags"]);
  return val;
}
void NGSetDescriptorFlags(int _fd, int _flags) {
  if (fcntl(_fd, F_SETFL, _flags) == -1)
    THROW([[NGIOException alloc]
              initWithReason:@"could not set descriptor flags"]);
}

void NGAddDescriptorFlag (int _fd, int _flag) {
  int val = NGGetDescriptorFlags(_fd);
  NGSetDescriptorFlags(_fd, val | _flag);
}

// ******************** NonBlocking IO ************

int NGDescriptorRecv(int _fd, char *_buf, int _len, int _flags, int _timeout) {
  int errorCode;
  int result;

  result = recv(_fd, _buf, _len, _flags);
  if (result == 0) return 0; // EOF

  errorCode = errno;

  if ((result == -1) && (errorCode == EWOULDBLOCK)) { // retry
    struct pollfd pfd;
    pfd.fd      = _fd;
    pfd.events  = POLLRDNORM;
    pfd.revents = 0;

    do {
      if ((result = poll(&pfd, 1, _timeout)) < 0) {
        errorCode = errno;

        // retry if interrupted
        if ((errorCode != EINTR) && (errorCode != EAGAIN)) 
          break;
      }
    }
    while (result < 0);

    if (result == 1) { // data waiting, try to read
      result = recv(_fd, _buf, _len, _flags);
      if (result == 0)
        return 0; // EOF
      else if (result == -1) {
        errorCode = errno;

        if (errorCode == EWOULDBLOCK)
          NSLog(@"WARNING: would block although descriptor was polled ..");
      }
    }
    else if (result == 0) {
#if 0
      NSLog(@"nonblock: recv on %i timed out after %i milliseconds ..",
             _fd, _timeout);
#endif
      result = -2;
    }
    else
      result = -1;
  }

  return result;
}

int NGDescriptorSend(int _fd, const char *_buf, int _len, int _flags, int _timeout) {
  int errorCode;
  int result;

  result = send(_fd, _buf, _len, _flags);
  if (result == 0) return 0; // EOF

  errorCode = errno;

  if ((result == -1) && (errorCode == EWOULDBLOCK)) { // retry
    struct pollfd pfd;
    pfd.fd      = _fd;
    pfd.events  = POLLWRNORM;
    pfd.revents = 0;

    do {
      if ((result = poll(&pfd, 1, _timeout)) < 0) {
        errorCode = errno;

        if (errorCode != EINTR) // retry only if interrupted
          break;
      }
    }
    while (result < 0);

    if (result == 1) { // data waiting, try to read
      result = send(_fd, _buf, _len, _flags);
      if (result == 0) return 0; // EOF
    }
    else if (result == 0) {
      NSLog(@"nonblock: send on %i timed out after %i milliseconds ..",
             _fd, _timeout);
      result = -2;
    }
    else
      result = -1;
  }

  return result;
}

#endif // WIN32
