/* 
   NGFileStream.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGFileStream.m,v 1.28 1999/01/29 13:19:21 helge Exp $

#import  <NGExtensions/NGTool.h>
#include "config.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include "common.h"
#include "NGFileStream.h"
#include "NGBufferedStream.h"
#include "NGConcreteStreamFileHandle.h"
#include "NGFileUrl.h"
#include "NGLockingStream.h"
#include "NGStreamExceptions.h"
#include "NGUrl.h"
#include "NGUrlExceptions.h"
#include "NGDescriptorFunctions.h"

#if !defined(POLLRDNORM)
#  define POLLRDNORM POLLIN
#endif

NSString *NGFileReadOnly    = @"r";
NSString *NGFileWriteOnly   = @"w";
NSString *NGFileReadWrite   = @"rw";
NSString *NGFileAppend      = @"a";
NSString *NGFileReadAppend  = @"ra";

static const int NGInvalidUnixDescriptor = -1;
static const int NGFileCreationMask      = 0666; // rw-rw-rw-

@interface _NGConcreteFileStreamFileHandle : NGConcreteStreamFileHandle
@end

id<NGInputStream>  NGIn  = nil;
id<NGOutputStream> NGOut = nil;
id<NGOutputStream> NGErr = nil;

@implementation NGFileStream

// stdio stream

#if defined(WIN32)
- (id)__initWithInConsole {
  if ((self = [self init])) {
    self->systemPath = @"CONIN$";
    self->streamMode = NGStreamMode_readWrite;
    self->fh = GetStdHandle(STD_INPUT_HANDLE);
    /*
    self->fh = CreateFile("CONIN$", GENERIC_READ, FILE_SHARE_READ,
                          NULL,
                          OPEN_EXISTING,
                          0,
                          NULL);
     */
  }
  return self;
}
- (id)__initWithOutConsole {
  if ((self = [self init])) {
    DWORD written;
    self->systemPath = @"CONOUT$";
    self->streamMode = NGStreamMode_readWrite;
    self->fh         = GetStdHandle(STD_OUTPUT_HANDLE);
    /*
    self->fh = CreateFile("CONOUT$", GENERIC_WRITE, FILE_SHARE_WRITE,
                          NULL,
                          OPEN_EXISTING,
                          0,
                          NULL);
     */
    if (!WriteFile(self->fh, "hallo\r\n", 7, &written, NULL))
      abort();
    FlushFileBuffers(self->fh);
  }
  return self;
}
#else
- (id)__initWithDescriptor:(int)_fd mode:(NGStreamMode)_mode {
  if ((self = [self init])) {
    self->fd         = _fd;
    self->streamMode = _mode;
  }
  return self;
}
#endif

void NGInitStdio(void) {
  static BOOL isInitialized = NO;
  if (!isInitialized) {
    NGFileStream *ti = nil, *to = nil, *te = nil;
    
    isInitialized = YES;

#if defined(WIN32)
    ti = [[NGFileStream alloc] __initWithInConsole];
    to = [[NGFileStream alloc] __initWithOutConsole];
    te = RETAIN(to);
#else
    ti = [[NGFileStream alloc] __initWithDescriptor:0 mode:NGStreamMode_readOnly];
    to = [[NGFileStream alloc] __initWithDescriptor:1 mode:NGStreamMode_writeOnly];
    te = [[NGFileStream alloc] __initWithDescriptor:2 mode:NGStreamMode_writeOnly];
#endif

    NGIn  = [[NGBufferedStream alloc] initWithSource:ti];
    NGOut = [[NGBufferedStream alloc] initWithSource:to];
    NGErr = [[NGBufferedStream alloc] initWithSource:te];

    RELEASE(ti); ti = nil;
    RELEASE(to); to = nil;
    RELEASE(te); te = nil;
  }
}

+ (void)_makeThreadSafe:(NSNotification *)_notification {
  NGLockingStream *li = nil, *lo = nil, *le = nil;
  
  if ([NGIn isKindOfClass:[NGLockingStream class]])
    return;

  li = [[NGLockingStream alloc] initWithSource:NGIn];  RELEASE(NGIn);  NGIn  = li;
  lo = [[NGLockingStream alloc] initWithSource:NGOut]; RELEASE(NGOut); NGOut = lo;
  le = [[NGLockingStream alloc] initWithSource:NGErr]; RELEASE(NGErr); NGErr = le;
}

+ (void)_flushForExit:(NSNotification *)_notification {
  [NGIn  flush];
  [NGOut flush];
  [NGErr flush];
}

+ (void)initialize {
  BOOL isInitialized = NO;
  if (!isInitialized) {
    isInitialized = YES;

    if ([NSThread isMultiThreaded])
      [self _makeThreadSafe:nil];
    else {
      [[NSNotificationCenter defaultCenter]
                             addObserver:self
                             selector:@selector(_makeThreadSafe:)
                             name:NSWillBecomeMultiThreadedNotification
                             object:nil];
    }

    [[NSNotificationCenter defaultCenter]
                           addObserver:self
                           selector:@selector(_flushForExit:)
                           name:NGProcessWillExitNotificationName
                           object:nil];
    [[NSNotificationCenter defaultCenter]
                           addObserver:self
                           selector:@selector(_flushForExit:)
                           name:NGProcessWillAbortNotificationName
                           object:nil];
  }
}

// normal file stream

+ (id)streamForResource:(id<NGUrl>)_url {
  if (![_url isKindOfClass:[NGFileUrl class]]) {
    [[[NGInvalidUrlTypeException alloc] initWithFormat:
         @"file url required for file stream, got %@", _url] raise];
  }
  return AUTORELEASE([[self alloc] initWithFileUrl:_url]);
}

- (id)init {
  if ((self = [super init])) {
    self->streamMode = NGStreamMode_undefined;
    self->systemPath = nil;
    self->markDelta  = -1;
    self->handle     = nil;
#if defined(WIN32)
    self->fh         = INVALID_HANDLE_VALUE;
#else
    self->fd         = NGInvalidUnixDescriptor;
#endif
  }
  return self;
}

- (id)initWithFileUrl:(NGFileUrl *)_url {
  if ((self = [self init])) {
    self->systemPath = [[_url fileSystemPath] copyWithZone:[self zone]];
  }
  return self;
}
- (id)initWithFileHandle:(NSFileHandle *)_handle {
  if ((self = [self init])) {
#if defined(WIN32)
    self->fh = [_handle nativeHandle];
#else
    self->fd = [_handle fileDescriptor];
#endif
  }
  return self;
}

- (void)gcFinalize {
  if ([self isOpen]) {
    NSLog(@"NGFileStream(gcFinalize): closing %@", self);
    [self close];
  }
}
- (void)dealloc {
  [self gcFinalize];
  self->streamMode = NGStreamMode_undefined;
  RELEASE(self->systemPath); self->systemPath = nil;
  self->handle = nil;
  [super dealloc];
}

// opening

- (void)openInMode:(NSString *)_mode {
  // throws
  //   NGUnknownStreamModeException  when _mode is invalid
  //   NGCouldNotOpenStreamException when the file could not be opened
#if defined(WIN32)
  DWORD openFlags;
  DWORD shareMode;

  if (self->fh != INVALID_HANDLE_VALUE)
    [self close]; // if stream is open, close and reopen

  if ([_mode isEqualToString:NGFileReadOnly]) {
    self->streamMode = NGStreamMode_readOnly;
    openFlags = GENERIC_READ;
    shareMode = FILE_SHARE_READ;
  }
  else if ([_mode isEqualToString:NGFileWriteOnly]) {
    self->streamMode = NGStreamMode_writeOnly;
    openFlags = GENERIC_WRITE;
    shareMode = FILE_SHARE_WRITE;
  }
  else if ([_mode isEqualToString:NGFileReadWrite]) {
    self->streamMode = NGStreamMode_readWrite;
    openFlags = GENERIC_READ | GENERIC_WRITE;
    shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE;
  }
  else {
    THROW([[NGUnknownStreamModeException alloc]
              initWithStream:self mode:_mode]);
    return;
  }

  self->fh = CreateFile([self->systemPath fileSystemRepresentation],
                        openFlags, shareMode, NULL,
                        OPEN_ALWAYS, // same as the Unix O_CREAT flag
                        0,           // security flags ?
                        NULL);

  if (self->fh == INVALID_HANDLE_VALUE) {
    THROW([[NGCouldNotOpenStreamException alloc] initWithStream:self]);
  }
#else
  int openFlags; // flags passed to open() call

  if (self->fd != NGInvalidUnixDescriptor)
    [self close]; // if stream is open, close and reopen

  if ([_mode isEqualToString:NGFileReadOnly]) {
    self->streamMode = NGStreamMode_readOnly;
    openFlags = O_RDONLY;
  }
  else if ([_mode isEqualToString:NGFileWriteOnly]) {
    self->streamMode = NGStreamMode_writeOnly;
    openFlags = O_WRONLY | O_CREAT;
  }
  else if ([_mode isEqualToString:NGFileReadWrite]) {
    self->streamMode = NGStreamMode_readWrite;
    openFlags = O_RDWR | O_CREAT;
  }
  else if ([_mode isEqualToString:NGFileAppend]) {
    self->streamMode = NGStreamMode_writeOnly;
    openFlags = O_WRONLY | O_CREAT | O_APPEND;
  }
  else if ([_mode isEqualToString:NGFileReadAppend]) {
    self->streamMode = NGStreamMode_readWrite;
    openFlags = O_RDWR | O_CREAT | O_APPEND;
  }
  else {
    THROW([[NGUnknownStreamModeException alloc]
              initWithStream:self mode:_mode]);
    return;
  }

  self->fd = open([self->systemPath fileSystemRepresentation],
                  openFlags,
                  NGFileCreationMask);

  if (self->fd == -1) {
    self->fd = NGInvalidUnixDescriptor;

    THROW([[NGCouldNotOpenStreamException alloc] initWithStream:self]);
  }
#endif

  self->markDelta = -1; // not marked
}

- (BOOL)isOpen {
#if defined(WIN32)
  return (self->fh != INVALID_HANDLE_VALUE) ? YES : NO;
#else
  return (self->fd != NGInvalidUnixDescriptor) ? YES : NO;
#endif
}

// Foundation file handles

- (void)resetFileHandle { // called by NSFileHandle on dealloc
  self->handle = nil;
}
- (NSFileHandle *)fileHandle {
  if (self->handle == nil)
    self->handle = [[_NGConcreteFileStreamFileHandle allocWithZone:[self zone]]
                                                     initWithStream:self];
  return AUTORELEASE(self->handle);
}

#if defined(WIN32)
- (HANDLE)windowsFileHandle {
  return self->fh;
}
#endif

- (int)fileDescriptor {
#if defined(WIN32)
  return (int)[self fileHandle];
#else
  return self->fd;
#endif
}

// primitives

static void _checkOpen(NGFileStream *self, NSString *_reason) {
#if defined(WIN32)
  if (self->fh == INVALID_HANDLE_VALUE)
#else
  if (self->fd == NGInvalidUnixDescriptor)
#endif
    THROW([[NGStreamNotOpenException alloc] initWithStream:self reason:_reason]);
}

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  // throws
  //   NGWriteOnlyStreamException  when the stream is not readable
  //   NGStreamNotOpenException    when the stream is not open
  //   NGEndOfStreamException      when the end of the stream is reached
  //   NGStreamReadErrorException  when the read call failed

  _checkOpen(self, @"tried to read from a file stream which is closed");

  if (!NGCanReadInStreamMode(streamMode))
    [NGWriteOnlyStreamException raiseWithStream:self];

  {
#if defined(WIN32)
    DWORD readResult = 0;

    if (ReadFile(self->fh, _buf, _len, &readResult, NULL) == FALSE) {
      DWORD lastErr = GetLastError();

      if (lastErr == ERROR_HANDLE_EOF)
        [NGEndOfStreamException raiseWithStream:self];
      else
        [NGStreamReadErrorException raiseWithStream:self errorCode:lastErr];

      NSLog(@"invalid program state, aborting ..");
      [[NGTool tool] coreDump];
    }
    if (readResult == 0)
      [NGEndOfStreamException raiseWithStream:self];
#else
    size_t readResult;
    int    retryCount = 0;

    do {
      readResult = read(self->fd, _buf, _len);
      
      if (readResult == 0)
        [NGEndOfStreamException raiseWithStream:self];
      else if (readResult == -1) {
        int errCode = errno;

        if (errCode == EINTR)
          // system call was interrupted
          retryCount++;
        else
          [NGStreamReadErrorException raiseWithStream:self errorCode:errCode];
      }
    }
    while ((readResult <= 0) && (retryCount < 10));

    if (retryCount >= 10)
      [NGStreamReadErrorException raiseWithStream:self errorCode:EINTR];
#endif
    
    NSAssert(readResult > 0, @"invalid read method state");

    // adjust mark
    if (self->markDelta != -1)
      self->markDelta += readResult; // increase delta
    
    return readResult;
  }
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  // throws
  //   NGReadOnlyStreamException   when the stream is not writeable
  //   NGStreamNotOpenException    when the stream is not open
  //   NGStreamWriteErrorException when the write call failed
  
  _checkOpen(self, @"tried to write to a file stream which is closed");
  
  if (!NGCanWriteInStreamMode(streamMode))
    [NGReadOnlyStreamException raiseWithStream:self];

  {
#if defined(WIN32)
    DWORD writeResult = 0;

    if (WriteFile(self->fh, _buf, _len, &writeResult, NULL) == FALSE) {
      DWORD errorCode = GetLastError();

      switch (errorCode) {
        case ERROR_INVALID_HANDLE:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self reason:@"incorrect file handle"]);
          break;
        case ERROR_WRITE_PROTECT:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self reason:@"disk write protected"]);
          break;
        case ERROR_NOT_READY:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self reason:@"the drive is not ready"]);
          break;
        case ERROR_HANDLE_EOF:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self reason:@"reached end of file"]);
          break;
        case ERROR_DISK_FULL:
          THROW([[NGStreamWriteErrorException alloc]
                    initWithStream:self reason:@"disk is full"]);
          break;
        
        default:
          [NGStreamWriteErrorException raiseWithStream:self
                                       errorCode:GetLastError()];
      }
      
      NSLog(@"invalid program state, aborting");
      abort();
    }
#else
    size_t writeResult;
    int    retryCount = 0;

    do {
      writeResult = write(self->fd, _buf, _len);

      if (writeResult == -1) {
        int errCode = errno;

        if (errCode == EINTR)
          // system call was interrupted
          retryCount++;
        else
          [NGStreamWriteErrorException raiseWithStream:self errorCode:errno];
      }
    }
    while ((writeResult == -1) && (retryCount < 10));

    if (retryCount >= 10)
      [NGStreamWriteErrorException raiseWithStream:self errorCode:EINTR];
#endif
    
    return writeResult;
  }
}

- (void)close {
#if defined(WIN32)
  if (self->fh == INVALID_HANDLE_VALUE) {
    NSLog(@"tried to close already closed stream %@", self);
    return;
  }

  if (CloseHandle(self->fh) == FALSE)
    THROW([[NGCouldNotCloseStreamException alloc] initWithStream:self]);

  self->fh = INVALID_HANDLE_VALUE;
#else
  if (self->fd == NGInvalidUnixDescriptor) {
    NSLog(@"tried to close already closed stream %@", self);
    return;
  }

  if (close(self->fd) != 0)
    [NGCouldNotCloseStreamException raiseWithStream:self];
  
  self->fd = NGInvalidUnixDescriptor;
#endif
  self->markDelta = -1;
}

- (NGStreamMode)mode {
  return self->streamMode;
}
- (BOOL)isRootStream {
  return YES;
}

#if defined(WIN32)
- (void)flush {
  if (self->fh != INVALID_HANDLE_VALUE)
    FlushFileBuffers(self->fh);
}
#endif

// blocking

#if defined(WIN32)
- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  NSLog(@"%@ not supported in Windows environment !",
        NSStringFromSelector(_cmd));
  return YES;
}
#else
- (BOOL)wouldBlockInMode:(NGStreamMode)_mode {
  short events = 0;

  if (self->fd == NGInvalidUnixDescriptor)
    return NO;

  if (NGCanReadInStreamMode(_mode))  events |= POLLRDNORM;
  if (NGCanWriteInStreamMode(_mode)) events |= POLLWRNORM;

  // timeout of 0 means return immediatly
  return (NGPollDescriptor(self->fd, events, 0) == 1 ? NO : YES);
}
#endif

// marking

- (void)mark {
  self->markDelta = 0;
}
- (void)rewind {
  [self moveByOffset:-(self->markDelta)];
  self->markDelta = -1;
}
- (BOOL)markSupported {
  return YES;
}

// NGPositionableStream

- (void)moveToLocation:(unsigned)_location {
  self->markDelta = -1;

#if defined(WIN32)
  if (SetFilePointer(self->fh, _location, NULL, FILE_BEGIN) == -1)
    THROW([[NGStreamSeekErrorException alloc]
              initWithStream:self errorCode:GetLastError()]);
#else
  if (lseek(self->fd, _location, SEEK_SET) == -1)
    THROW([[NGStreamSeekErrorException alloc] initWithStream:self errorCode:errno]);
#endif
}
- (void)moveByOffset:(int)_delta {
  self->markDelta += _delta;
  
#if defined(WIN32)
  if (SetFilePointer(self->fh, _delta, NULL, FILE_CURRENT) == -1)
    THROW([[NGStreamSeekErrorException alloc]
              initWithStream:self errorCode:GetLastError()]);
#else
  if (lseek(self->fd, _delta, SEEK_CUR) == -1)
    THROW([[NGStreamSeekErrorException alloc] initWithStream:self errorCode:errno]);
#endif
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<%@[0x%08X] path=%@ mode=%@>",
                     NSStringFromClass([self class]), (unsigned)self,
                     self->systemPath ? self->systemPath : @"nil",
                     [self modeDescription]];
}

@end

@implementation _NGConcreteFileStreamFileHandle

// accessors

#if defined(WIN32)
- (HANDLE)fileHandle {
  return [(NGFileStream *)stream windowsFileHandle];
}
#endif

- (int)fileDescriptor {
  return [(NGFileStream *)stream fileDescriptor];
}

@end
