/* 
   NGFileUrl.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGFileUrl.m,v 1.29 1999/01/29 13:50:49 helge Exp $

#include "config.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include "common.h"
#include "NGFileUrl.h"
#include "NGUrlExceptions.h"
#include "NGStreamExceptions.h"

NSString *NGFileUrlProtocolName = @"file";

static inline NSString *_systemPathToUrlPath(NSString *_path) {
#if defined(WIN32)
  if (_path) {
    unsigned      len = [_path cStringLength];
    char          buffer[len + 16];
    unsigned int  pos = 0;
    unsigned char driveLetter = '\0';

    [_path getCString:buffer];

    // convert all backslashes to slashes
    for (pos = 0; pos < len; pos++) {
      if (buffer[pos] == '\\')
        buffer[pos] = '/';
    }
    
    if (len > 1) {
      if (buffer[1] == ':') {
        // with drive letter (eg C:\tmp\auto.bat)
        // => 'file:///C|/tmp/auto.bat'
        driveLetter = buffer[0];
        buffer[1] = '|'; // convert ':' => '|'
      }
    }

    if (driveLetter) {
      // path is 'C:\autoexec.bat"
      BOOL needsSlash = YES;

      if (len > 2) { // 'C:\'
        if (buffer[2] == '/')
          needsSlash = NO;
      }

      if (!needsSlash) {
        _path = [@"///" stringByAppendingString:
                          [NSString stringWithCString:buffer length:len]];
      }
      else {
        _path = [NSString stringWithFormat:@"///%c|/%s",
                            (int)driveLetter, (buffer + 2)];
      }
    }
    else
      _path = [NSString stringWithCString:buffer length:len];
  }
#endif
  return _path;
}

static inline NSString *_urlPathToSystemPath(NSString *_path) {
  if (_path) {
#if defined(WIN32)
    // Example: 'file:///H|/dev/LSOffice3'
    //          'file:////trex2.mdlink.de/home/helge/doof.txt'
    
    unsigned len = [_path cStringLength];
    char     path[len + 1];

    [_path getCString:path];

    if (len > 2) {
      if ((path[0] == '/') && (path[2] == '|')) {
        // absolute with logical drive, like '/G|' becomes just 'G:'
        char *orig = path;
        char *dest = path + 1;

        while ((*orig = *dest)) {
          orig++;
          dest++;
        }
      }
    }

    // need to check for Windows network host ('//mycomputer/Temp/file.txt')

    // now convert slashes to back slashes
    {
      char *trace = path;
      while (*trace) {
        if (*trace == '/')
          *trace = '\\';
        trace++;
      }
    }

    return [NSString stringWithCString:path];
#else
    NSString   *sysPath = nil;
    const char *tmp = NULL;
  
    sysPath = [_path stringByStandardizingPath];
    if (sysPath == nil) {
      //NSLog(@"WARNING: could not standardize path for file url path '%@'", _path);
      return AUTORELEASE([_path copy]);
    }
  
    tmp = [sysPath fileSystemRepresentation];
    if (tmp == NULL) {
      NSLog(@"WARNING: could not get filesys-representation for path '%@'", _path);
      return sysPath;
    }

    return [NSString stringWithCString:tmp];
#endif
  }
  return nil;
}

static int _compareUrlNames(id<NGUrl> _url1, id<NGUrl> _url2, id _userData) {
  return [[(id)_url1 resourceName]
                     isEqualToString:
                       [(id)_url2 resourceName]];
}

@interface NGRelativeFileUrl : NGFileUrl
{
  id<NGTreeUrl> parentUrl;
  NSString      *resourceName;
}

+ (id)urlWithString:(NSString *)_name relativeToUrl:(id<NGTreeUrl>)_url;
- (id)initWithString:(NSString *)_name relativeToUrl:(id<NGTreeUrl>)_url;

@end

@interface NGFileUrl(_PrivateMethods)
- (id)_initWithPath:(NSString *)_path andSystemPath:(NSString *)_sysPath;
@end

@implementation NGFileUrl

static NSFileManager *fileManager = nil;

+ (void)initialize {
  BOOL isInitialized = NO;
  if (!isInitialized) {
    isInitialized = YES;

    fileManager = RETAIN([NSFileManager defaultManager]);
  }
}

- (id)_initWithPath:(NSString *)_path andSystemPath:(NSString *)_sysPath {
  if ((self = [super init])) {
    self->path       = [_path    copyWithZone:[self zone]];
    self->systemPath = [_sysPath copyWithZone:[self zone]];
  }
  return self;
}

+ (id)fileUrlForCurrentDirectory {
  NSString *p = [fileManager currentDirectoryPath];
  
  return AUTORELEASE([[self alloc] _initWithPath:_systemPathToUrlPath(p)
                                   andSystemPath:p]);
}

+ (id)urlWithString:(NSString *)_string protocol:(NSString *)_protocolName {
  if (_string == nil) return nil;
  return AUTORELEASE([[self alloc] _initWithPath:_string
                                   andSystemPath:_urlPathToSystemPath(_string)]);
}

+ (id)fileUrlWithString:(NSString *)_string {
  if (_string == nil) return nil;
  return AUTORELEASE([[self alloc] _initWithPath:_string
                                   andSystemPath:_urlPathToSystemPath(_string)]);
}

+ (id)fileUrlWithPath:(NSString *)_path {
  id<NGUrl> url = nil;
  
  if (_path == nil) return nil;
  
  if (![_path isAbsolutePath]) {
    // produce returns retained url
    url = [[[NSProcessInfo processInfo] currentResource]
                           produceRelativeUrlForChildResource:_path];
  }
  else {
    url = [[self alloc] _initWithPath:_systemPathToUrlPath(_path)
                        andSystemPath:_path];
  }
  return AUTORELEASE(url);
}

- (id)initWithString:(NSString *)_string protocol:(NSString *)_protocolName {
  if (_string == nil) {
    AUTORELEASE(self);
    return nil;
  }
  return [self _initWithPath:_string
               andSystemPath:_urlPathToSystemPath(_string)];
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->path);
  RELEASE(self->systemPath);
  [super dealloc];
}
#endif

// accessors

- (NSString *)path {
  return self->path;
}
- (NSString *)fileSystemPath {
  return self->systemPath;
}

- (NSString *)resourceName {
  NSString *ppath = [self fileSystemPath];
  
  if ([ppath isEqualToString:@"/"]) // root directory, no parent
    return @"/";
  else
    return [ppath lastPathComponent];
}

- (id<NGUrl>)parentResource {
  NSString *ppath = [self fileSystemPath];

  if ([ppath isEqualToString:@"/"]) // root directory, no parent
    return nil;
  else {
    ppath = [ppath stringByDeletingLastPathComponent];
    return [NGFileUrl fileUrlWithPath:ppath];
  }
}

- (id<NGUrl>)childResourceWithName:(NSString *)_name {
  if ([self isLeafResource]) {
    THROW([[NGInvalidUrlException alloc] initWithReason:
              @"requested child resource of a leaf resource"]);
    return nil;
  }

  // produce returns retained url
  return AUTORELEASE([self produceRelativeUrlForChildResource:_name]);
}
- (id<NGUrl>)childResourceWithName:(NSString *)_name extension:(NSString *)_ext {
  return [self childResourceWithName:[_name stringByAppendingPathExtension:_ext]];
}

// operations

- (BOOL)doesResourceExist {
  return [fileManager fileExistsAtPath:[self fileSystemPath]];
}

- (BOOL)isAccessibleInMode:(NGStreamMode)_mode {
  // throws
  //   NGUnknownStreamModeException
  BOOL result = NO;
  
  switch (_mode) {
    case NGStreamMode_readOnly:
      result = [fileManager isReadableFileAtPath:[self fileSystemPath]];
      break;
    case NGStreamMode_writeOnly:
      result = [fileManager isWritableFileAtPath:[self fileSystemPath]];
      break;
    case NGStreamMode_readWrite:
      result = [fileManager isReadableFileAtPath:[self fileSystemPath]];
      if (result) result = [fileManager isWritableFileAtPath:[self fileSystemPath]];
      break;

    default:
      THROW([[NGUnknownStreamModeException alloc] init]);
      break;
  }
  return result;
}

// stat info

- (NSDictionary *)attributes {
  return [fileManager fileAttributesAtPath:[self fileSystemPath]
                      traverseLink:YES];
}

- (unsigned long long)size {       // returns the size of the resource in bytes
  return [[[fileManager fileAttributesAtPath:[self fileSystemPath]
                        traverseLink:YES]
                        objectForKey:NSFileSize]
                        unsignedLongValue];
}

- (NSDate *)lastModificationDate { // time of last modification (write)
  return [[fileManager fileSystemAttributesAtPath:[self fileSystemPath]]
                       objectForKey:NSFileModificationDate];
}

// comparing

- (unsigned)hash {
  return [self->path hash];
}
- (BOOL)isEqual:(id)_object {
  if (![_object isKindOfClass:[NGFileUrl class]])
    return NO;
  else
    return [((NGFileUrl *)_object)->path isEqualToString:self->path];
}

// NGUrl protocol

- (NSString *)protocol {
  return NGFileUrlProtocolName;
}

- (NSString *)stringValue {
  return [NSString stringWithFormat:@"%@:%@", [self protocol], self->path];
}

// NGTreeUrl

- (NSString *)pathRelativeToParent:(NGFileUrl *)_parent {
  if (_parent == nil)
    return [self path];
  else if ([_parent isEqual:self])
    return @"";
  else {
    NSMutableArray  *components = [NSMutableArray arrayWithCapacity:16];
    NSMutableString *relPath    = nil;
    NGFileUrl       *parent     = nil;
    BOOL isFirst = YES;
    int i, count;

    [components addObject:[self resourceName]];
    parent = [self parentResource];
    while (![parent isEqual:_parent] && (parent != nil)) {
      [components addObject:[parent resourceName]];
      parent = [parent parentResource];
    }

    if (parent == nil) { // did not found  parent !
      // should throw a better exception
      THROW([[NGInvalidUrlException alloc] init]);
    }

    count = [components count];
    relPath = [NSMutableString stringWithCapacity:count * 8];
    for (i = (count - 1); i >= 0; i--) {
      if (isFirst) isFirst = NO;
      else [relPath appendString:@"/"];
      
      [relPath appendString:[components objectAtIndex:i]];
    }
    return relPath;
  }
}

- (BOOL)isLeafResource {
  BOOL isDirectory;
  BOOL doesExist;

  doesExist = [fileManager fileExistsAtPath:[self fileSystemPath]
                           isDirectory:&isDirectory];

  if (!doesExist)   return YES;
  if (!isDirectory) return YES;
  return NO;
}

- (BOOL)isCollection {
  return ![self isLeafResource];
}

- (NSEnumerator *)childResources {
  return [self childResourcesWithFilter:NULL
               sortUsingFunction:_compareUrlNames
               context:nil];
}

- (NSEnumerator *)childResourcesWithFilter:(NGChildNameFilterFunction)_filter
  sortUsingFunction:(NGCompareUrlFunction)_compare
  context:(id)_userData {
  
  if ([self isLeafResource]) {
    THROW([[NGInvalidUrlException alloc] initWithReason:
            @"requested child resources of a leaf resource"]);
    return nil;
  }
  else {
    NSMutableArray *children = nil;
    NSArray        *contents = nil;
    int            i, count;
  
    contents = [fileManager directoryContentsAtPath:[self fileSystemPath]];
    count    = [contents count];
    children = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:count];

    for (i = 0; i < count; i++) {
      NSString  *childName = [contents objectAtIndex:i];
      BOOL      shouldAdd  = NO;

      if (_filter == NULL)
        shouldAdd = YES;
      else if (_filter(self, childName, _userData))
        shouldAdd = YES;

      if (shouldAdd) {
        id<NGUrl> child = nil;
        // produce returns retained url
        child = [self produceRelativeUrlForChildResource:childName];
        [children addObject:child];
        RELEASE(child); child = nil;
      }
    }
    contents = nil;

    if ([children count] > 0) {
      if (_compare) {
        // sort results
        [children sortUsingFunction:(int (*)(id,id,void*))_compare
                  context:_userData];
      }
      
      AUTORELEASE(children);
      return [children objectEnumerator];
    }
    else {
      RELEASE(children); children = nil;
      return nil;
    }
  }
}
- (NSEnumerator *)childResourcesWithExtension:(NSString *)_extension {
  static BOOL _filter(id<NGTreeUrl> _parent, NSString *_name, id _context) {
    return [[_name pathExtension] isEqualToString:_extension];
  }

  return [self childResourcesWithFilter:_filter
               sortUsingFunction:_compareUrlNames
               context:nil];
}

#if HAVE_NGRegularExpression
- (NSEnumerator *)childResourcesWithPattern:(NSString *)_pattern {
  NGRegularExpression *pattern = nil;
  BOOL (*doesMatch)(id,SEL,NSString *_string);
  
  static BOOL _filter(id<NGTreeUrl> _parent, NSString *_name, id _context) {
    return doesMatch(_context, @selector(matchesString:), _name);
  }

  pattern   = [NGRegularExpression regularExpression:_pattern];
  doesMatch = (void *)[pattern methodForSelector:@selector(matchesString:)];

  return [self childResourcesWithFilter:_filter
               sortUsingFunction:_compareUrlNames
               context:pattern];
}
#endif

- (id<NGUrl>)produceRelativeUrlForChildResource:(NSString *)_name {
  if ([self isLeafResource]) {
    THROW([[NGInvalidUrlTypeException alloc]
              initWithReason:@"tried to produce child-url for leaf-resource .."]);
  }
  
  return [[NGRelativeFileUrl alloc]
                             initWithString:_name relativeToUrl:self];
}

- (id<NGTreeUrl>)createCollectionWithName:(NSString *)_name {
  // throws
  //   NGUrlCollectionAlreadyExists  if there is already a subresource with the name
  //   NGUrlCouldNotCreateCollection if the collection could not be created
  NGFileUrl *newDir  = nil;

  // produce returns retained url
  newDir = [self produceRelativeUrlForChildResource:_name];

  if ([newDir doesResourceExist]) {
    RELEASE(newDir); newDir = nil;
    
    THROW([[NGUrlCollectionAlreadyExistsException alloc]
              initWithReason:@"could not create collection, it already exists !"]);
  }

  //NSLog(@"path=%@ fspath=%@", [newDir path], [newDir fileSystemPath]);

  // create collection
  {
    BOOL didSucceed;

    didSucceed = [fileManager createDirectoryAtPath:[newDir fileSystemPath]
                              attributes:nil];
    if (!didSucceed) {
      RELEASE(newDir); newDir = nil;
      THROW([[NGUrlCouldNotCreateCollectionException alloc]
                initWithReason:@"could not create directory"]);
    }
  }

  return AUTORELEASE(newDir);
}

// operations

- (void)deleteResource {
  // throws
  //   NGUrlCouldNotDeleteResourceException  when the resource couldn't be deleted
  //                                         or isn't deletable

  if (![fileManager isDeletableFileAtPath:[self fileSystemPath]]) {
    NSLog(@"errno is %s", strerror(errno));
    
    THROW([[NGUrlCouldNotDeleteResourceException alloc]
              initWithReason:@"resource is not deletable"]);
  }

  if (![fileManager removeFileAtPath:[self fileSystemPath] handler:nil]) {
    THROW([[NGUrlCouldNotDeleteResourceException alloc]
              initWithReason:@"could not delete resource"]);
  }
}

// description

- (NSString *)description {
  return [self stringValue];
}

// NSCopying protocol

- (id)copyWithZone:(NSZone *)_zone {
  return [[self class] initWithString:self->path protocol:NGFileUrlProtocolName];
}

// NSCoding protocol

- (Class)classForCoder {
  return [NGFileUrl class];
}
- (void)encodeWithCoder:(NSCoder *)_encoder {
  [_encoder encodeObject:self->path];
}
- (id)initWithCoder:(NSCoder*)_decoder {
  NSString *urlPath = [_decoder decodeObject];
  return [self initWithString:urlPath protocol:NGFileUrlProtocolName];
}

@end

// NGRelativeFileUrl

@implementation NGRelativeFileUrl

+ (id)urlWithString:(NSString *)_name relativeToUrl:(id<NGTreeUrl>)_url {
  return AUTORELEASE([[self alloc] initWithString:_name relativeToUrl:_url]);
}

- (id)initWithString:(NSString *)_name relativeToUrl:(id<NGTreeUrl>)_url {
  NSMutableString *aPath    = nil;
  NSString        *aSysPath = nil;

  aPath    = [[(NGFileUrl *)_url path] mutableCopyWithZone:[self zone]];
  aSysPath = [(NGFileUrl *)_url fileSystemPath];

  [aPath appendString:@"/"];
  [aPath appendString:_name];
  aSysPath = [aSysPath stringByAppendingPathComponent:_name];

  if ((self = [super _initWithPath:aPath andSystemPath:aSysPath])) {
    self->parentUrl    = RETAIN(_url);
    self->resourceName = [_name copyWithZone:[self zone]];
  }
  RELEASE(aPath); aPath = nil; aSysPath = nil;
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->parentUrl);
  RELEASE(self->resourceName);
  [super dealloc];
}
#endif

// accessors

- (id<NGUrl>)parentResource {
  return self->parentUrl;
}

- (NSString *)resourceName {
  NSAssert(self->resourceName, @"no resource name set ..");
  return self->resourceName;
}

// NGUrl

- (NSString *)protocol {
  return [self->parentUrl protocol];
}

@end

// categories

@implementation NSProcessInfo(NGCurrentDirectory)

- (NGFileUrl *)currentResource {
  char buf[1025];

#if defined(WIN32)
  DWORD result;
  result = GetCurrentDirectory(sizeof(buf), buf);

  if (result < sizeof(buf))
    return [NGFileUrl fileUrlWithPath:[NSString stringWithCString:buf]];

  if (result > sizeof(buf))
    THROW([[NGIOException alloc] initWithReason:@"path buffer to small"]);

  THROW([[NGIOException alloc]
            initWithReason:@"Could not retrieve current directory"]);
#else

#ifdef HAVE_GETCWD
  if (getcwd(buf, sizeof(buf) - 1))
    return [NGFileUrl fileUrlWithPath:[NSString stringWithCString:buf]];
#else
  if (getwd(buf))
    return [NGFileUrl fileUrlWithPath:[NSString stringWithCString:buf]];
#endif

  switch (errno) {
    case EACCES:
      THROW([[NGIOAccessException alloc] initWithReason:
                @"a parent directory cannot be read to get it's name for "
                @"the current directory"]);
      break;

    default:
      THROW([[NGIOException alloc] init]);
      break;
  }
#endif
  return nil;
}

@end

@implementation NSFileManager(NGFileUrlExtensions)

- (NGFileUrl *)currentDirectoryResource {
  return [NGFileUrl fileUrlWithPath:[self currentDirectoryPath]];
}

@end

@implementation NSTask(NGFileUrlExtensions)

- (void)setCurrentDirectoryResource:(NGFileUrl *)_url {
  [self setCurrentDirectoryPath:[_url fileSystemPath]];
}
- (NGFileUrl *)currentDirectoryResource {
  return [NGFileUrl fileUrlWithPath:[self currentDirectoryPath]];
}

- (void)setLaunchUrl:(NGFileUrl *)_url {
  [self setLaunchPath:[_url fileSystemPath]];
}
- (NGFileUrl *)launchUrl {
  return [NGFileUrl fileUrlWithPath:[self launchPath]];
}

@end

@implementation NSBundle(NGFileUrlExtensions)

- (NGFileUrl *)resourceWithName:(NSString *)_name ofType:(NSString *)_type {
  return [NGFileUrl fileUrlWithPath:[self pathForResource:_name ofType:_type]];
}

- (NGFileUrl *)resourceWithName:(NSString *)_name ofType:(NSString *)_type
  inDirectory:(NGFileUrl *)_dirResource {

  return [NGFileUrl fileUrlWithPath:
                      [self pathForResource:_name ofType:_type
                            inDirectory:[_dirResource fileSystemPath]]];
}

- (NGFileUrl *)bundleResource {
  return [NGFileUrl fileUrlWithPath:[self bundlePath]];
}

@end
