/* 
   NGStream.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStream.m,v 1.8 1999/01/27 16:41:16 helge Exp $

#include "common.h"
#include "NGStreamExceptions.h"
#include "NGStream.h"
#include "NGFilterStream.h"

@implementation NGStream

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  [self subclassResponsibility:_cmd];
  return 0;
}
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  [self subclassResponsibility:_cmd];
  return 0;
}

- (void)flush {
}
- (void)close {
}

- (NGStreamMode)mode {
  [self subclassResponsibility:_cmd];
  return 0;
}
- (BOOL)isRootStream {
  [self subclassResponsibility:_cmd];
  return NO;
}

// methods method which write exactly _len bytes

- (void)safeReadBytes:(void *)_buf count:(unsigned)_len {
  NGSafeReadBytesFromStream(self, _buf, _len);
}

- (void)safeWriteBytes:(const void *)_buf count:(unsigned)_len {
  NGSafeWriteBytesToStream(self, _buf, _len);
}

// marking

- (void)mark {
  NSLog(@"WARNING: called mark on a stream which doesn't support marking !");
}
- (void)rewind {
  THROW([[NGStreamException alloc] initWithStream:self
                                   reason:@"marking not supported"]);
}
- (BOOL)markSupported {
  return NO;
}

// convenience methods

- (int)readByte {
  return NGReadByteFromStream(self);
}

// description

- (NSString *)modeDescription {
  NSString *result = @"unknown";
  
  switch ([self mode]) {
    case NGStreamMode_undefined: result = @"undefined"; break;
    case NGStreamMode_readOnly:  result = @"r";         break;
    case NGStreamMode_writeOnly: result = @"w";         break;
    case NGStreamMode_readWrite: result = @"rw";        break;
    default:
      THROW([[NGUnknownStreamModeException alloc] initWithStream:self]);
      break;
  }
  return result;
}

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<%@[0x%08X] mode=%@>",
                     NSStringFromClass([self class]), (unsigned)self,
                     [self modeDescription]];
}

@end

// concrete implementations as functions

int NGReadByteFromStream(id<NGInputStream> _stream) {
  int result;

  NS_DURING {
    unsigned char c;
    [_stream readBytes:&c count:sizeof(unsigned char)];
    result = c;
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]])
      result = -1;
    else
      [localException raise];
  }
  NS_ENDHANDLER;
  
  return result;
}

void NGSafeReadBytesFromStream(id<NGInputStream> _in, void *_buf, unsigned _len){
  int toBeRead = _len;
  int readResult;
  NGIOReadMethodType readBytes;

  readBytes = (NGIOReadMethodType)
    [(NSObject *)_in methodForSelector:@selector(readBytes:count:)];

  NS_DURING {
    void *pos = _buf;
    
    while (YES) {
      readResult = (unsigned)readBytes(_in, @selector(readBytes:count:),
                                       pos, toBeRead);

      if (readResult == toBeRead) {
        // all bytes were read successfully, return
        break;
      }

      if (readResult < 1) {
        [NSException raise:NSInternalInconsistencyException
                     format:@"readBytes:count: returned a value < 1"];
      }

      toBeRead -= readResult;
      pos += readResult;
    }
  }
  NS_HANDLER {
    if ([localException isKindOfClass:[NGEndOfStreamException class]]) {
      [[[NGEndOfStreamException alloc]
                                initWithStream:(id)_in
                                readCount:(_len - toBeRead)
                                safeCount:_len
                                data:[NSData dataWithBytes:_buf
                                             length:(_len - toBeRead)]]
                                raise];
    }
    else
      [localException raise];
  }
  NS_ENDHANDLER;
}

void NGSafeWriteBytesToStream(id<NGOutputStream> _o,const void *_b,unsigned _l) {
  int  toBeWritten = _l;
  int  writeResult;
  void *pos = (void *)_b;
  NGIOWriteMethodType writeBytes;

  writeBytes = (NGIOWriteMethodType)
    [(NSObject *)_o methodForSelector:@selector(writeBytes:count:)];

  while (YES) {
    writeResult =
      (int)writeBytes(_o, @selector(writeBytes:count:), pos, toBeWritten);

    if (writeResult == toBeWritten) {
      // all bytes were written successfully, return
      break;
    }

    if (writeResult < 1) {
      [NSException raise:NSInternalInconsistencyException
                   format:@"writeBytes:count: returned a value<1 in %@", _o];
    }

    toBeWritten -= writeResult;
    pos += writeResult;
  }
}
