/* 
   NGStreamExceptions.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStreamExceptions.m,v 1.10 1999/01/27 16:41:17 helge Exp $

#include "common.h"
#include "NGStreamExceptions.h"

@implementation NGIOException

- (id)init {
  self = [super initWithName:NSStringFromClass([self class])
                reason:@"an IO exception occured"
                userInfo:nil];
  return self;
}
- (id)initWithReason:(NSString *)_reason {
  self = [super initWithName:NSStringFromClass([self class])
                reason:_reason
                userInfo:nil];
  return self;
}

@end

@implementation NGStreamException

- (NSString *)defaultReason {
  return @"a stream exception occured";
}

- (id)init {
  return [self initWithStream:nil reason:[self defaultReason]];
}
- (id)initWithStream:(id<NGStream>)_stream {
  return [self initWithStream:_stream reason:[self defaultReason]];
}

- (id)initWithStream:(id<NGStream>)_stream format:(NSString *)_format,... {
  NSString *tmp = nil;
  va_list  ap;
  
  va_start(ap, _format);
  tmp = [NSString stringWithFormat:_format arguments:ap];
  va_end(ap);

  return [self initWithStream:_stream reason:tmp];
}

- (id)initWithStream:(id<NGStream>)_stream reason:(NSString *)_reason {
  NSDictionary *ui = nil;

  ui = [NSDictionary dictionaryWithObjectsAndKeys:
                       [NSNumber numberWithInt:errno], @"errno",
                       [NSString stringWithCString:strerror(errno)], @"error",
                       _stream, @"stream",
                       nil];
  
  self = [super initWithName:NSStringFromClass([self class])
                reason:_reason
                userInfo:ui];
  if (self) {
    stream = RETAIN(_stream);
  }
  return self;
}

+ (void)raiseWithStream:(id<NGStream>)_stream {
  THROW([[self alloc] initWithStream:_stream]);
}
+ (void)raiseWithStream:(id<NGStream>)_stream format:(NSString *)_format,... {
  NSString *tmp = nil;
  va_list  ap;
  
  va_start(ap, _format);
  tmp = [NSString stringWithFormat:_format arguments:ap];
  va_end(ap);

  THROW([[self alloc] initWithStream:_stream reason:tmp]);
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(stream); stream = nil;
  [super dealloc];
}
#endif

@end

// ******************** NGEndOfStreamException ********************

@implementation NGEndOfStreamException

- (id)initWithStream:(id<NGStream>)_stream {
  return [self initWithStream:_stream
               readCount:0
               safeCount:0
               data:nil];
}

- (id)initWithStream:(id<NGStream>)_stream
  readCount:(unsigned)_readCount
  safeCount:(unsigned)_safeCount
  data:(NSData *)_data {

  NSString *tmp;

  tmp = [NSString stringWithFormat:@"reached end of stream %@", _stream];

  if ((self = [super initWithStream:_stream reason:tmp])) {
    readCount = _readCount;
    safeCount = _safeCount;
    data      = RETAIN(_data);
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(data); data = nil;
  [super dealloc];
}
#endif

// accessors

- (NSData *)readBytes {
  return data;
}

@end

// ******************** open state exceptions *********************

@implementation NGCouldNotOpenStreamException

- (NSString *)defaultReason {
  return @"could not open stream";
}

@end

@implementation NGCouldNotCloseStreamException

- (NSString *)defaultReason {
  return @"could not close stream";
}

@end

@implementation NGStreamNotOpenException

- (NSString *)defaultReason {
  return @"stream is not open";
}

@end

// ******************** NGStreamErrors ****************************

@implementation NGStreamErrorException

- (id)initWithStream:(id<NGStream>)_stream errorCode:(int)_code {
  NSString *tmp = nil;

  tmp = [NSString stringWithFormat:@"stream error occured, errno=%i error=%s",
                    _code, strerror(_code)];
  if ((self = [self initWithStream:_stream reason:tmp])) {
    osErrorCode = _code;
  }
  tmp = nil;
  return self;
}

+ (void)raiseWithStream:(id<NGStream>)_stream errorCode:(int)_code {
  [[[self alloc] initWithStream:_stream errorCode:_code] raise];
}

- (int)operationSystemErrorCode {
  return osErrorCode;
}
- (NSString *)operatingSystemError {
  return [NSString stringWithCString:strerror(osErrorCode)];
}

@end

@implementation NGStreamReadErrorException

- (NSString *)defaultReason {
  return @"read error on stream";
}

@end

@implementation NGStreamWriteErrorException

- (NSString *)defaultReason {
  return @"write error on stream";
}

@end

@implementation NGStreamSeekErrorException

- (NSString *)defaultReason {
  return @"seek error on stream";
}

@end

// ******************** NGStreamModeExceptions ********************

@implementation NGStreamModeException

- (NSString *)defaultReason {
  return @"stream mode failure";
}

@end

@implementation NGUnknownStreamModeException

- (NSString *)defaultReason {
  return @"unknow stream mode";
}

- (id)initWithStream:(id<NGStream>)_stream mode:(NSString *)_streamMode {
  if ((self = [super initWithStream:_stream
                     format:@"unknown stream mode: %@", _streamMode])) {
    streamMode = [_streamMode copy];
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(streamMode); streamMode = nil;
  [super dealloc];
}
#endif

@end

@implementation NGReadOnlyStreamException

- (NSString *)defaultReason {
  return @"stream is read only";
}

@end

@implementation NGWriteOnlyStreamException

- (NSString *)defaultReason {
  return @"stream is write only";
}

@end

// ******************** NGIOAccessException ******************

@implementation NGIOAccessException
@end

@implementation NGIOSearchAccessException
@end
