/* 
   NGStreamPipe.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStreamPipe.m,v 1.9 1999/01/27 16:41:17 helge Exp $

#include "config.h"
#include "common.h"
#include "NGStreamPipe.h"
#include "NGFileStream.h"
#include "NGBufferedStream.h"

#if defined(WIN32)

@implementation NGStreamPipe
@end

#else

static const int NGInvalidUnixDescriptor = -1;

@interface _NGConcretePipeFileHandle : NSFileHandle
{
@public
  int *fd;
}

- (id)initWithDescriptor:(int *)_fd;

@end

@interface NGFileStream(PrivateMethods)
- (id)__initWithDescriptor:(int)_fd mode:(NGStreamMode)_mode;
@end

@implementation NGStreamPipe

+ (id)pipe {
  return AUTORELEASE([[self alloc] init]);
}

- (id)init {
  if (pipe(self->fildes) == -1) {
    NSLog (@"pipe() system call failed: %s", strerror (errno));
    AUTORELEASE(self);
    return nil;
  }
  return self;
}

- (void)gcFinalize {
  [self close];
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  [self gcFinalize];
  RELEASE(self->fhIn);
  RELEASE(self->fhOut);
  [super dealloc];
}
#endif

- (NSFileHandle *)fileHandleForReading {
  if (self->fhIn == nil) {
    self->fhIn = [[_NGConcretePipeFileHandle alloc]
                      initWithDescriptor:&(self->fildes[0])];
  }
  return self->fhIn;
}
- (NSFileHandle *)fileHandleForWriting {
  if (self->fhOut == nil) {
    self->fhOut = [[_NGConcretePipeFileHandle alloc]
                       initWithDescriptor:&(self->fildes[1])];
  }
  return self->fhOut;
}

- (id<NGByteSequenceStream>)streamForReading {
  return self;
}
- (id<NGOutputStream>)streamForWriting {
  return self;
}

// NGInputStream

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  size_t readResult;

  if (self->fildes[0] == NGInvalidUnixDescriptor) {
    THROW([[NGStreamReadErrorException alloc]
                                       initWithStream:self
                                       reason:@"read end of pipe is closed"]);
  }
  
  readResult = read(self->fildes[0], _buf, _len);

  if (readResult == 0)
    [NGEndOfStreamException raiseWithStream:self];
  else if (readResult == -1)
    [NGStreamReadErrorException raiseWithStream:self errorCode:errno];

  return readResult;
}
- (void)safeReadBytes:(void *)_buf count:(unsigned)_len {
  NGSafeReadBytesFromStream(self, _buf, _len);
}

// marks

- (void)mark {
  NSLog(@"WARNING: called mark on a stream which doesn't support marking !");
}
- (void)rewind {
  THROW([[NGStreamException alloc]
            initWithStream:self reason:@"marking not supported"]);
}
- (BOOL)markSupported {
  return NO;
}

// NGOutputStream

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  size_t writeResult;

  if (self->fildes[1] == NGInvalidUnixDescriptor) {
    THROW([[NGStreamWriteErrorException alloc]
              initWithStream:self reason:@"write end of pipe is closed"]);
  }
  
  writeResult = write(self->fildes[1], _buf, _len);

  if (writeResult == -1)
    [NGStreamWriteErrorException raiseWithStream:self errorCode:errno];
  return writeResult;
}
- (void)safeWriteBytes:(const void *)_buf count:(unsigned)_len {
  NGSafeWriteBytesToStream(self, _buf, _len);
}

- (void)flush {
}

// NGStream

- (void)close {
  if (self->fildes[0] != NGInvalidUnixDescriptor) close(self->fildes[0]);
  if (self->fildes[1] != NGInvalidUnixDescriptor) close(self->fildes[1]);
}

- (NGStreamMode)mode {
  NGStreamMode mode = NGStreamMode_undefined;

  if (self->fildes[0] != NGInvalidUnixDescriptor)
    mode |= NGStreamMode_readOnly;
  if (self->fildes[1] != NGInvalidUnixDescriptor)
    mode |= NGStreamMode_writeOnly;

  return mode;
}

// NGByteSequenceStream

- (int)readByte {
  return NGReadByteFromStream(self);
}

// Extensions

- (BOOL)isOpen {
  return (self->fildes[0] == NGInvalidUnixDescriptor) &&
         (self->fildes[1] == NGInvalidUnixDescriptor) ? NO : YES;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:@"<0x%08X[%@]: in=%i out=%i>",
                     self, NSStringFromClass([self class]),
                     self->fildes[0], self->fildes[1]];
}

@end

@implementation _NGConcretePipeFileHandle

- (id)initWithDescriptor:(int *)_fd {
  self->fd = _fd;
  return self;
}

- (int)fileDescriptor {
  return *(self->fd);
}

- (void)closeFile {
  close(*(self->fd));
  *(self->fd) = NGInvalidUnixDescriptor;
}

@end

#endif /* WIN32 */
