/* 
   NGUrl.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge@mdlink.de)

   This file is part of the MDlink Object Framework 2 (MOF2)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGUrl.m,v 1.22 1999/03/16 11:29:00 helge Exp $

#include <objc/objc-api.h>
#include "common.h"
#include "NGUrl.h"
#include "NGFileUrl.h"
#include "NGUrlExceptions.h"
#include "NGUrlChars.h"

static NSString *NGUrlProtocolSeparator = @":";

@interface NGDefaultUrlFactory : NSObject < NGUrlFactory >
@end

@implementation NGUrl

static NSMutableDictionary *protocolNameToFactory = nil;
static NSRecursiveLock     *lock                  = nil;

+ (void)initialize {
  static BOOL initialized = NO;

  if (!initialized) {
    protocolNameToFactory = [[NSMutableDictionary alloc] initWithCapacity:16];
    
    [[NSNotificationCenter defaultCenter]
                           addObserver:self
                           selector:@selector(taskNowMultiThreaded:)
                           name:NSWillBecomeMultiThreadedNotification
                           object:nil];

    [self registerFactory:AUTORELEASE([[NGDefaultUrlFactory alloc] init])
          forProtocolWithName:NGFileUrlProtocolName];
    
    initialized = YES;
  }
}

+ (void)taskNowMultiThreaded:(NSNotification *)_notification {
  lock = [[NSRecursiveLock alloc] init];
}

// URL factory

+ (id<NGUrl>)urlWithString:(NSString *)_url {
  id<NGUrlFactory> factory = nil;
  NSRange  range;
  NSString *protocol = nil;
  NSString *value    = nil;

  range = [_url rangeOfString:NGUrlProtocolSeparator];
  if ((range.length == 0) || (range.location == 0)) {
    THROW([[NGInvalidUrlException alloc]
              initWithFormat:@"no protocol specified in url '%@'", _url]);
  }
  protocol = [_url substringToIndex:range.location];
  value    = [_url substringFromIndex:(range.location + range.length)];
  
  [lock lock];
  {
    factory = [protocolNameToFactory objectForKey:protocol];
  }
  [lock unlock];

  if (factory == nil) {
    THROW([[NGInvalidUrlException alloc]
          initWithFormat:@"did not found factory for protocol %@", protocol]);

  }

  return [factory produceUrlForString:value protocol:protocol];
}

+ (void)registerFactory:(id<NGUrlFactory>)_factory
  forProtocolWithName:(NSString *)_protocol {
  
  [lock lock];
  {
    [protocolNameToFactory setObject:_factory forKey:_protocol];
  }
  [lock unlock];
}

// NGUrl protocol

- (NSString *)protocol {
  return [self subclassResponsibility:_cmd];
}

- (NSString *)stringValue {
  return [self subclassResponsibility:_cmd];
}

- (NSString *)resourceName {
  return [self subclassResponsibility:_cmd];
}

@end

@implementation NGDefaultUrlFactory

- (id<NGUrl>)produceUrlForString:(NSString *)_string
  protocol:(NSString *)_protocolName {

  if ([_protocolName isEqualToString:NGFileUrlProtocolName])
    return [NGFileUrl urlWithString:_string protocol:_protocolName];
  return nil;
}

@end

// support functions

static inline int _valueOfHexChar(char _c) {
  switch (_c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      return (_c - 48); // 0-9 (ascii-char)'0' - 48 => (int)0
      
    case 'A': case 'B': case 'C':
    case 'D': case 'E': case 'F':
      return (_c - 55); // A-F, A=10..F=15, 'A'=65..'F'=70
      
    case 'a': case 'b': case 'c':
    case 'd': case 'e': case 'f':
      return (_c - 87); // a-f, a=10..F=15, 'a'=97..'f'=102

    default:
      return -1;
  }
}
static inline BOOL _isHexDigit(char _c) {
  switch (_c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case 'A': case 'B': case 'C':
    case 'D': case 'E': case 'F':
    case 'a': case 'b': case 'c':
    case 'd': case 'e': case 'f':
      return YES;

    default:
      return NO;
  }
}

static inline BOOL isToBeEscaped(char _c) {
  return (isUrlAlphaNum(_c) || (_c == '_')) ? NO : YES;
}

BOOL NGContainsUrlEscapeCharacters(const char *_buffer) {
  while (*_buffer) {
    if (*_buffer == '%')
      return YES;
    _buffer++;
  }
  return NO;
}
BOOL NGContainsUrlInvalidCharacters(const char *_buffer) {
  while (*_buffer) {
    if (isToBeEscaped(*_buffer))
      return YES;
    _buffer++;
  }
  return NO;
}

void NGUnescapeUrlBuffer(const char *_source, char *_dest) {
  BOOL done = NO;

  while (!done && (*_source != '\0')) {
    char c = *_source;

    //if (c == '+') // '+' stands for a space
    //  *_dest = ' ';
    if (c == '%') {
      _source++; c = *_source;
      
      if (c == '\0') {
        *_dest = '%';
        done = YES;
      }
      else if (_isHexDigit(c)) { // hex-escaped char, like '%F3'
        int decChar = _valueOfHexChar(c);
        _source++;
        c = *_source;
        decChar = decChar * 16 + _valueOfHexChar(c);
        *_dest = (char)decChar;
      }
      else // escaped char, like '%%' -> '%'
        *_dest = c;
    }
    else // char passed through
      *_dest = c;

    _dest++;
    _source++;
  }
  *_dest = '\0';
}

NSString *NGUnescapeUrlString(NSString *_source) {
  const char *cstr = [_source cString];
  
  if (NGContainsUrlEscapeCharacters(cstr)) {
    char *buffer = NULL;

    buffer = objc_atomic_malloc([_source cStringLength] + 2);
    NGUnescapeUrlBuffer([_source cString], buffer);
    return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
  }
  else
    return AUTORELEASE([_source copy]);
}

void NGEscapeUrlBuffer(const char *_source, char *_dest) {
  while (*_source) {
    //if (*_source == ' ') { // a ' ' becomes a '+'
    //  *_dest = '+'; _dest++;
    //}
    if (!isToBeEscaped(*_source)) {
      *_dest = *_source;
      _dest++;
    } 
    else { // any other char is escaped ..
      *_dest = '%'; _dest++;
      sprintf(_dest, "%02X", (unsigned)*_source);
      _dest += 2;
    }
    _source++;
  }
  *_dest = '\0';
}
NSString *NGEscapeUrlString(NSString *_source) {
  const char *cstr = [_source cString];
  
  if (NGContainsUrlInvalidCharacters(cstr)) { // needs to be escaped ?
    char *buffer = NULL;

    buffer = objc_atomic_malloc([_source cStringLength] * 3 + 2);
    NGEscapeUrlBuffer(cstr, buffer);

    return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
  }
  else
    return AUTORELEASE([_source copy]);
}
