// $Id: NGGZipStream.m,v 1.3 1999/01/29 14:32:13 helge Exp $

#include "zutil.h"
#undef Assert
#import "common.h"
#include <NGStreams/NGStreamExceptions.h>
#include "NGGZipStream.h"
#include "NSData+gzip.h"

@implementation NGGZipStream

- (id)initWithSource:(id<NGStream>)_source level:(int)_level {
  if ((self = [super initWithSource:_source])) {
    z_stream *zout;
    
    NSAssert1((_level >= NGGZipMinimalCompression &&
               _level <= NGGZipMaximalCompression)
              || (_level == Z_DEFAULT_COMPRESSION),
              @"invalid compression level %i (0-9)", _level);

    self->outBufLen = 2048;
    self->outBuf    = NSZoneMallocAtomic([self zone], self->outBufLen);
    self->out       = NSZoneMallocAtomic([self zone], sizeof(z_stream));
    zout = self->out;
    zout->zalloc    = (alloc_func)NULL;
    zout->zfree     = (free_func)NULL;
    zout->opaque    = (voidpf)NULL;
    zout->next_out  = self->outBuf;
    zout->avail_out = self->outBufLen;
    zout->next_in   = Z_NULL;
    zout->avail_in  = 0;
    self->crc       = crc32(0L, Z_NULL, 0);

    if (deflateInit2(zout, _level, Z_DEFLATED, -MAX_WBITS,
                     DEF_MEM_LEVEL, 0) != Z_OK) {
      NSLog(@"Could not init deflate ..");
      AUTORELEASE(self);
      return nil;
    }
  }
  return self;
}

- (void)gcFinalize {
  [self close];
}

- (void)dealloc {
  [self gcFinalize];
  [super dealloc];
}

// headers

- (void)writeGZipHeader {
  // gzip header
  char buf[10] = {
    0x1f, 0x8b,    // magic
    Z_DEFLATED, 0, // flags
    0, 0, 0, 0,    // time
    0, OS_CODE     // flags
  };

  [self safeWriteBytes:buf count:10];
}
- (void)writeGZipTrailer {
  static inline void putLong(NGGZipStream *self, uLong x) {
    int n;
    for (n = 0; n < 4; n++) {
      unsigned char c = (int)(x & 0xff);
      [self safeWriteBytes:&c count:1];
      x >>= 8;
    }
  }
 
  putLong(self, self->crc);
  putLong(self, ((z_stream *)self->out)->total_in);
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len { // decoder
  [self notImplemented:_cmd];
  return -1;
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len { // encoder
  z_stream *zout = self->out;
  
  if (!self->headerIsWritten) [self writeGZipHeader];

  { // gz_write
    zout->next_in   = (void*)_buf;
    zout->avail_in  = _len;
    
    while (zout->avail_in > 0) {
      int errorCode;
      
      if (zout->avail_out == 0) {
        [self safeWriteBytes:self->outBuf count:self->outBufLen];
        zout->next_out  = self->outBuf; // reset buffer position
        zout->avail_out = self->outBufLen;
      }
      errorCode = deflate(self->out, Z_NO_FLUSH);
      if (errorCode != Z_OK) {
        if (zout->state) deflateEnd(self->out);
        [NGStreamException raiseWithStream:self
                           format:@"could not deflate chunk !"];
      }
    }
    self->crc = crc32(self->crc, _buf, _len);
  }
  return _len;
}
- (void)safeWriteBytes:(const void *)_buf count:(unsigned)_len { // encoder
  // gzip writes are safe
  [self writeBytes:_buf count:_len];
}

- (void)close {
  [self flush];
  [self writeGZipTrailer];
  if (((z_stream *)self->out)->state) deflateEnd(self->out);
  [super close];
}

- (void)flush {
  int      errorCode = Z_OK;
  z_stream *zout     = self->out;
  BOOL     done      = NO;
    
  zout->next_in  = NULL;
  zout->avail_in = 0; // should be zero already anyway
    
  while (1) {
    int len = self->outBufLen - zout->avail_out;

    if (len > 0) {
      [self safeWriteBytes:self->outBuf count:len];
      zout->next_out  = self->outBuf;
      zout->avail_out = self->outBufLen;
    }
    if (done)
      break;
    errorCode = deflate(zout, Z_FINISH);

    // deflate has finished flushing only when it hasn't used up
    // all the available space in the output buffer: 
    done = (zout->avail_out != 0 || errorCode == Z_STREAM_END);

    if (errorCode != Z_OK && errorCode != Z_STREAM_END)
      break;
  }
  if (errorCode != Z_STREAM_END) {
    if (zout->state) deflateEnd(zout);
    [NGStreamException raiseWithStream:self format:@"flush failed"];
  }

  [super flush];
}

@end
