// $Id: minigzip.m,v 1.4 1998/11/06 15:49:31 helge Exp $
/* minigzip.c -- simulate gzip using the zlib compression library
 * Copyright (C) 1995-1996 Jean-loup Gailly.
 * For conditions of distribution and use, see copyright notice in zlib.h 
 */

/*
 * minigzip is a minimal implementation of the gzip utility. This is
 * only an example of using zlib and isn't meant to replace the
 * full-featured gzip. No attempt is made to deal with file systems
 * limiting names to 14 or 8+3 characters, etc... Error checking is
 * very limited. So use minigzip only for testing; use gzip for the
 * real thing. On MSDOS, use only on file names without extension
 * or in pipe mode.
 */

/* $Id: minigzip.m,v 1.4 1998/11/06 15:49:31 helge Exp $ */

#import "common.h"

#define BUFLEN       4096
#define MAX_NAME_LEN 1024

static char *GZ_SUFFIX = ".gz";
char *prog; // stores argv[0]

void error           (const char *msg);
void gz_compress     (FILE   *in, gzFile out);
void gz_uncompress   (gzFile in, FILE   *out);
void file_compress   (char  *file);
void file_uncompress (char  *file);

// Display error message and exit
void error(const char *msg) {
  fprintf(stderr, "%s: %s\n", prog, msg);
  exit(1);
}

// Compress input to output then close both files.
void gz_compress(FILE *in, gzFile out) {
  char buf[BUFLEN];
  int len;
  int err;

  for (;;) {
    len = fread(buf, 1, sizeof(buf), in);
    if (ferror(in)) {
      perror("fread");
      exit(1);
    }
    if (len == 0) break;

    if (gzwrite(out, buf, (unsigned)len) != len) error(gzerror(out, &err));
  }
  fclose(in);
  if (gzclose(out) != Z_OK) error("failed gzclose");
}

// Uncompress input to output then close both files.
void gz_uncompress(gzFile in, FILE *out) {
  char buf[BUFLEN];
  int len;
  int err;

  for (;;) {
    len = gzread(in, buf, sizeof(buf));
    if (len < 0) error (gzerror(in, &err));
    if (len == 0) break;

    if ((int)fwrite(buf, 1, (unsigned)len, out) != len) {
      error("failed fwrite");
    }
  }
  if (fclose(out)) error("failed fclose");

  if (gzclose(in) != Z_OK) error("failed gzclose");
}


// Compress the given file: create a corresponding .gz file and remove the original.
void file_compress(char *file) {
  char outfile[MAX_NAME_LEN];
  FILE  *in;
  gzFile out;

  strcpy(outfile, file);
  strcat(outfile, GZ_SUFFIX);

  in = fopen(file, "rb");
  if (in == NULL) {
    perror(file);
    exit(1);
  }
  out = gzopen(outfile, "wb"); /* use "wb9" for maximal compression */
  if (out == NULL) {
    fprintf(stderr, "%s: can't gzopen %s\n", prog, outfile);
    exit(1);
  }
  gz_compress(in, out);

  unlink(file);
}

// Uncompress the given file and remove the original.
void file_uncompress(char *file) {
  char buf[MAX_NAME_LEN];
  char *infile, *outfile;
  FILE  *out;
  gzFile in;
  int len = strlen(file);

  strcpy(buf, file);

  if (len > strlen(GZ_SUFFIX) &&
      strcmp(file+len-strlen(GZ_SUFFIX), GZ_SUFFIX) == 0) {
    infile = file;
    outfile = buf;
    outfile[len-3] = '\0';
  }
  else {
    outfile = file;
    infile = buf;
    strcat(infile, GZ_SUFFIX);
  }
  in = gzopen(infile, "rb");
  if (in == NULL) {
    fprintf(stderr, "%s: can't gzopen %s\n", prog, infile);
    exit(1);
  }
  out = fopen(outfile, "wb");
  if (out == NULL) {
    perror(file);
    exit(1);
  }

  gz_uncompress(in, out);

  unlink(infile);
}

// Usage:  minigzip [-d] [files...]
int main(int argc, char **argv, char **env) {
  int    uncompr = 0;

  [NSProcessInfo initializeWithArguments:argv count:argc environment:env];

  prog = argv[0];
  argc--, argv++;

  if (argc > 0) {
    uncompr = (strcmp(*argv, "-d") == 0);
    if (uncompr) {
      argc--, argv++;
    }
  }
  if (argc == 0) {
#if defined(WIN32)
    setmode(fileno(stdin), O_BINARY);
    setmode(fileno(stdout), O_BINARY);
#endif

    if (uncompr) {
      gzFile file;
      file = gzdopen(fileno(stdin), "rb");
      if (file == NULL) error("can't gzdopen stdin");
      gz_uncompress(file, stdout);
    }
    else {
      gzFile file;
      file = gzdopen(fileno(stdout), "wb"); // "wb9" for maximal compression
      if (file == NULL) error("can't gzdopen stdout");
      gz_compress(stdin, file);
    }
  }
  else {
    do {
      NSData   *data = nil;
      NSString *path = nil;

      path = [NSString stringWithCString:*argv];
      data = [NSData dataWithContentsOfFile:path];

      if (data == nil) {
        NSLog(@"ERROR: could not load file %@", path);
        continue;
      }

      if (uncompr) {
        file_uncompress(*argv);
        path = [path stringByDeletingPathExtension];
      }
      else {
        //file_compress(*argv);
        unsigned before = [data length];

        data = [data gzip];
        if (data == nil) {
          NSLog(@"ERROR: could not compress file %@", path);
          continue;
        }

        NSLog(@"compressed %i bytes => %i bytes.", before, [data length]);

        path = [path stringByAppendingPathExtension:@"gz"];
        [data writeToFile:path atomically:YES];
      }
    } while (argv++, --argc);
  }
  return 0;
}

LINK_NGExtensions;
LINK_NGStreams;
LINK_NGZlib;
